/* This module provides formatted file output services.
**
** Copyright (c) 1996 by Claflin & Clayton Inc. All rights reserved.
** Copyright (c) 1995 by Claflin & Clayton Inc. All rights reserved.
** Copyright (c) 1994 by Michael Newman, All rights reserved.
** Copyright (c) 1993 by Michael Newman, All rights reserved.
**
** MAKEDOC:EMIT_TO_HEADER_ON
#define makedocEdit 4
** MAKEDOC:EMIT_TO_HEADER_OFF
**
** Modification History:
**  29Feb95 MJNewman 4: Fix bug allocating memory in parseCode.
**  16Oct95 MJNewman 3: Fix matching of ** lines in comments.
**  26Sep95 MJNewman 2: Report modification history by dates.
**  22Sep95 MJNewman 1: Report modification history.
**  26Aug95 MJNewman 1: Add edit number and report it in comments.
**  26Aug95 MJNewman:	Add feature to copy sections to XH files. 
**  11Dec94 MJNewman:	Make .XH files test to see if they are included already
**  11Sep94 MJNewman:	Change declares to basetype.
**  17Jul94 MJNewman:	Merge NAV and TDLINK/NETPHONE versions.
**   5Jun94 MJNewman:	Report only first triple question of multiple lines
**		which contain triple questions.
**  28May94 MJNewman:	Fix bug finding field when there is a blank field.
**   3Feb93 MJNewman:	Add ability to output subroutine names and headers.
**   9Nov93 MJNewman:	Update make_fld to getField with new algorithm.
**  28Apr93 MJNewman:	Phase updated for -c<outputfile> command
**  22Mar93 MJNewman:	Report totals only if > 1 file read.
**   6Mar93 MJNewman:	Report lines of code read.
**   4Mar93 MJNewman:	Better english in translating message. Add removal of
**		control-Z character.
**  24Feb93 MJNewman:	Use AOSVS to ID DG environment.
**   9Feb93 MJNewman:	Rework header accumulation to be all at once allowing
**		complete removal of argument list for pre ANSI C.
**  31Jan93 MJNewman:	Handle indented preprocessor lines.
**  29Jan93 MJNewman:	Added translation to pre ANSI C.
**  29Jan93 MJNewman:	Fiddled names to remain unique in 8 characters.
**   5Jan93 MJNewman:	Created from GASSER file services.
*/

#include <stdio.h>
#include <string.h>

/* on the DG platform stdlib is missing, the stuff needed is in stdio */
#ifndef AOSVS
#include <stdlib.h>	/* exit() */
#endif

#include <ctype.h>
#include <math.h>
#include "basetype.h"

LOCALVAR BOOLEAN emitDocFile = FALSE;
LOCALVAR FILE *docFile = NULL;
LOCALVAR char doc_filename[STRING_LENGTH+1+10]; /* allow room to append type */
LOCALVAR char root_filename[STRING_LENGTH+1];
LOCALVAR FILE *code = NULL;
LOCALVAR FILE *entrys = NULL;
LOCALVAR int line_number = 0;
LOCALVAR int total_lines = 0;
LOCALVAR int files_read = 0;
LOCALVAR char *input_filename = NULL;
LOCALVAR char *documentationType;
LOCALVAR BOOLEAN preAnsiFlag = FALSE;
LOCALVAR BOOLEAN eliminateControlZsFlag = FALSE;
LOCALVAR unsigned int error_count;
LOCALVAR BOOLEAN emitGlobalEntrys = FALSE;
LOCALVAR BOOLEAN emitLocalEntrys = FALSE;
LOCALVAR BOOLEAN emitGlobalVariables = FALSE;
LOCALVAR BOOLEAN emitLocalVariables = FALSE;
LOCALVAR BOOLEAN emitHistory = FALSE;
LOCALVAR UINT32 historySince = 0;
LOCALVAR char code_filename[STRING_LENGTH+1+10]; /* room to append type */
LOCALVAR char entrys_filename[STRING_LENGTH+1+10]; /* room to append type */


/* forward references */
LOCALREF void parseCode(char *filename);
LOCALREF BOOLEAN dateToNumber(UINT32 *date,UINT16 year, char *month,UINT16 day);

/* parse filename and copy the root portion to root_filename
**
** Inputs:
**	filename    file to parse
*/
LOCALDEF void setRootName(char *filename)
{
    int i;
    if (root_filename[0] != '\0') return; /* already set */
    if (filename == NULL) return;
    for (i = 0;filename[i] != '\0'; i += 1) {
	if (filename[i] == '.') break;
	if (i >= DIM(root_filename)-1) break;
	root_filename[i] = filename[i];
    };
    root_filename[i] = '\0';
    if (filename[i] != '.') {
	printf("? Could not find root filename in %s\n",filename);
	root_filename[0] = '\0';
    };
    return;
}

/* read a file which is a list of files to read parsing each file along the
** way.
**
** Inputs:
**	filename	name of file to read
*/
LOCALDEF void read_indirect_file(char *filename)
{
    FILE *indfile;
    char line[STRING_LENGTH+1];

    indfile = fopen(filename,"r");
    if (indfile == NULL) {

	printf("%s 1: MAKEDOC Could not open indirect file.\n",filename);
	error_count += 1;
	return;
    };
    line[STRING_LENGTH] = '\0';
    while (NULL != fgets(line,STRING_LENGTH,indfile)) {
	int length;
	char first_nonspace;
	length = strlen(line);
	if (length <= 1) continue;
	first_nonspace = '\0';
	sscanf(line," %c",&first_nonspace);
	if (first_nonspace == '!') continue;
	if (line[length-1] == '\n') line[length-1] = '\0';
	parseCode(line);
    };
    return;
}

/* DOCUMENTATION:USER
** The program can make a header file, a translated code file and a
** documentation file.
**
** The header file is a ".h" file for C programs that follow Mike Newman's
** EXTERNDEF EXTERNVAR conventions. The default file type for the ".h" file is
** .xh (extracted header) and the default file name is the root name of the
** first file or indirect file presented. -hfilename explicitly sets the
** header filename.
**
** The first line of this comment had DOCUMENTATION:USER on it.
**
** At the beginning of a comment the keyword DOCUMENTATION followed by a colon
** and then a type name is recognized by the program. The type name
** is matched against a name provided with the -Dname switch
** on the command line. If the name matches every line in the file untill
** a line that starts with ** DOCUMENTATION_END is placed in the documentation
** file.
**
** During code translation with the -cFilename switch the following
** ANSI constructs are converted to pure K&R form:
**
**	Indented preprocessor directives are outdented. 
**
**	Subroutine headers are converted to place argument type declarations
**	outside of the argument list.
**
**	Argument lists of exactly (void) are changed to empty argument lists.
**
** Here is the USAGE printed by the program:
** DOCUMENTATION_END
*/
LOCALVAR char *helpMessage =
/* DOCUMENTATION:USER */
	"MAKEDOC Can make a header file from C sources,\n"
	"  extract indicated portions of the module to a documentation file,\n"
	"  and convert routine declarations to pre ANSI syntax.\n"
	"\n"
	"Usage is  MAKEDOC <arg> <arg>...  <file1> <file2> ...\n"
	"    Arguments start with - everything else is a file.\n"
	"    @ indicates and indirect file, reading each file in a list.\n"
	"\n"
	"    @filename read each file in a list of files.\n"
	"    -D- no doc file is written.\n"
	"    -D* emit all documentation blocks\n"
	"    -Dtype comments starting with:\n"
	"        /* DOCUMENTATION:\"type\" untill the */\n"
	"      are extracted to the documentation file\n"
	"    -E- put no Entry headers in the documentation file (default)\n"
	"    -EE put EXTERNDEF headers in the documentation file\n"
	"    -EL put LOCALDEF headers in the documentation file\n"
	"    -cfilename sets code output file and turns on this feature\n"
	"        Code is converted from ANSI style subroutine headers to\n"
	"        pre-ANSI K+R style.\n"
	"    -hfilename sets header output file, default is <file1>.xh\n"
	"    -m[-] Emit history to documentation file, default is OFF\n"
	"    -ofilename sets documentation output file, default is <file1>.xd\n"

	"    -P .xh files are written in pre-ANSI form.\n"
	"    -V- put no variable headers in the documentation file (default)\n"
	"    -VE put EXTERNVAR headers in the documentation file\n"
	"    -VL put LOCALVAR headers in the documentation file\n"
	"    -Z Remove any control-Z characters found.\n"
/* DOCUMENTATION_END */
	;
	
/* MAKEDOC main program
*/
EXTERNDEF int main(int argc,char *argv[]) {
    int i;
    root_filename[0] = '\0';
    entrys_filename[0] = '\0';
    if (argc < 2) {
	printf(helpMessage);
	exit(EXIT_FAILURE);
    };
    code = NULL;
    doc_filename[0] = '\0';
    documentationType = "NONE";
    error_count = 0;
    for (i = 1;i < argc;i += 1) {
	if (argv[i][0] == '@') {
	    setRootName(&argv[i][1]);
	    read_indirect_file(&argv[i][1]);
	    continue;
	};
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
		case 'c':
		    if (code != NULL) fclose(code);
		    strcpy(code_filename,&argv[i][2]);
		    code = fopen(code_filename,"w");
		    if (code == NULL) {
			printf("? Could not open %s\n",code_filename);
		    };
		    printf("Translation to pre-ANSI C, creating %s\n",code_filename);
		    break;
		case 'D':
		    if (argv[i][2] == '-') {
			documentationType = "NONE";
		    } else {
			emitDocFile = TRUE;
			documentationType = &argv[i][2];
		    };
		    break;
		case 'E':
		    switch (argv[i][2]) {
			case '-':
			    emitGlobalEntrys = FALSE;
			    emitLocalEntrys = FALSE;
			    break;
			case 'E':
			    emitGlobalEntrys = (argv[i][3] != '-');
			    break;
			case 'L':
			    emitLocalEntrys = (argv[i][3] != '-');
			    break;
		    };
		    if (emitGlobalEntrys || emitLocalEntrys) {
			emitDocFile = TRUE;
		    };
		    break;
		case 'h':
		    if (entrys != NULL) fclose(entrys);
		    strcpy(entrys_filename,&argv[i][2]);
		    entrys = fopen(entrys_filename,"w");
		    if (entrys == NULL) {
			printf("? Could not open file %s\n",entrys_filename);
			entrys_filename[0] = '\0';
		    };
		    break;
		case 'm':
		{
		    int year;
		    char month[STRING_LENGTH+1];
		    int day;
		    int matches;
		    emitHistory = (argv[i][2] != '-');
		    historySince = 0;
		    matches = sscanf(&argv[i][2]," %d %3s %d",&day,month,&year);
		    if (matches == 3) {
			if (!dateToNumber(&historySince,year,month,day)) {
			    historySince = 0;
			};
		    };
		    break;
		}
		case 'o':
		    emitDocFile = TRUE;
		    strcpy(doc_filename,&argv[i][2]);
		    break;
		case 'P':
		    preAnsiFlag = (argv[i][2] != '-');
		    break;
		case 'V':
		    switch (argv[i][2]) {
			case '-':
			    emitGlobalVariables = FALSE;
			    emitLocalVariables = FALSE;
			    break;
			case 'E':
			    emitGlobalVariables = (argv[i][3] != '-');
			    break;
			case 'L':
			    emitLocalVariables = (argv[i][3] != '-');
			    break;
		    };
		    if (emitGlobalVariables || emitLocalVariables) {
			emitDocFile = TRUE;
		    };
		    break;
		case 'Z':
		    eliminateControlZsFlag = (argv[i][2] != '-');
		    break;
		default:
		    printf("? Unrecognized selection after -: %s\n",argv[i]);
		    error_count += 1;
		    break;
	    };
	    continue;
	};
	setRootName(argv[i]);
	parseCode(argv[i]);
    };
    if (docFile != NULL) fclose(docFile);
    if (code != NULL) fclose(code);
    if (entrys != NULL) fprintf(entrys,
				"#endif /* __%s_XH__ */\n",
				root_filename);
    if (entrys != NULL) fclose(entrys);
    if (files_read > 1) {
	printf("\n%d total lines (all modules)\n",total_lines);
    };
    if (error_count > 0) {
	if (entrys_filename[0] != '\0') remove(entrys_filename);
	printf("\n%d errors encountered.\n",error_count);
	exit(EXIT_FAILURE);
    };
    exit(EXIT_SUCCESS);
    return EXIT_SUCCESS;
}

/* Decide if a string is completely blank
 *
 *Inputs:
 *	string		string to test
 * Returns:
 *	TRUE if the string is blank
 */
LOCALDEF BOOLEAN is_blank(char *string)
{
    int i;
    i = 0;
    if (string == NULL) return TRUE;
    while (isspace(string[i])) {
	i += 1;
    };
    if (string[i] == NUL) return TRUE;
    return FALSE;
}


/* Similar to strncmp. Colates numbers in an order
** more appreciated by people.  Treats upper and lower case letters
** as equivalent. Ignores trailing blanks.
**
** Inputs:
**	str1
**	str2
**	maxlen
** Returns:
**	same results as strcmp
*/
LOCALDEF int io_compareName(char *str1,char *str2,unsigned INT16 maxlen)
{
    int i;
    int j;
    INT16 v1;	/* numeric value */
    INT16 v2;
    char c1;
    char c2;
    int l1;	/* counts number of digits in number */
    int l2;

    i = 0;
    v1 = 0;
    while (TRUE) {
	c1 = toupper(str1[i]);
	c2 = toupper(str2[i]);
	if (c1 != c2) break;
	if (c1 == '\0') return 0;
	if (isdigit(c1)) {
	    v1 *= 10;
	    v1 += c1 - '0';
	} else {
	    v1 = 0;
	};
	i += 1;
	if (i >= maxlen) return 0;
    };
    if (c1 == '\0') {
	if (is_blank(&str2[i])) return 0;
    };
    if (c2 == '\0') {
	if (is_blank(&str1[i])) return 0;
    };
    if (!isdigit(c1) && !isdigit(c2)) {
	return (c1 > c2) ? 1 : -1;
    };
    v2 = v1;
    l2 = 0;
    for (j = 0;((i+j) < maxlen) && isdigit(str2[i+j]);j += 1) {
	l2 += 1;
	v2 *= 10;
	v2 += str2[i+j] - '0';
    };
    l1 = 0;
    for (j = 0;((i+j) < maxlen) && isdigit(str1[i+j]);j += 1) {
	l1 += 1;
	v1 *= 10;
	v1 += str1[i+j] - '0';
    };
    if (v1 > v2) return 1;
    if (v1 < v2) return -1;
    return (l1 > l2) ? 1 : -1;
}

/* print a message including file name of file being read,
** line number, contents of the line and the message.
**
** Inputs:
**	msg	    message to use
**	line	    text of line being worked on
**	iserror	    TRUE if this is an error
**	emitFile    TRUE to include file and line number in message
*/
LOCALDEF void out_message(char *msg,char *line,BOOLEAN iserror,
			  BOOLEAN emitFile)
{
    int length;
    if (msg == NULL) msg = "";
    if (line == NULL) line = "";
    if (emitFile) {
	printf("%s %d: ",input_filename,line_number);
    } else {
	printf("\t");
    };
    printf("%s %s",msg,line);
    length = strlen(line);
    if ((length == 0) || (line[length-1] == '\n')) {
	printf("\n");
    };
    if (iserror) error_count += 1;
    return;
}

/* get the first keyword on a line. The keyword returned is valid
** untill the next call to this routine
**
** Inputs:
**	line		line to parse
** Returns:
**	keyword found
*/
LOCALDEF char *first_keyword(char *line)
{
	static char name[STRING_LENGTH+1];
	name[0] = '\0';
	sscanf(line," %s",name);
	return name;
}

/* Find a keyword within a line and replace it with another
** the replacement keyword must be the same length as the initial
** keyword
**
** Inputs:
**	line		line to parse
**	keyword		keyword to look for
**	replacement	replacement for keyword
*/
LOCALDEF void keyword_replace(char *line,char *keyword,char *replacement)
{
	char *sub;
	int i;
	if (strlen(keyword) != strlen(replacement)) {
		out_message("Keyword and replacement different lengths\n",
			line,TRUE,TRUE);
		return;
	};
	sub = strstr(line,keyword);
	if (sub == NULL) return;
	for (i = 0;replacement[i] != '\0';i += 1) {
		sub[i] = replacement[i];
	};
	return;
}

/* test to see if a line contains a star slash
** if a comment ended and started again this will report ended
**
** Inputs:
**	line		line to look in
** Returns:
**	TRUE if star slash was found
*/
LOCALDEF BOOLEAN is_end_comment(char *line)
{
	if (NULL == strstr(line,"*/")) return FALSE;
	return TRUE;
}

/* test to see if a line starts with a preprocessor directive.
**
** Inputs:
**	line	    string to check
** Outputs:
**	start	    index of the start of the preprocessor directive
** Returns:
**	TRUE if there is a directive on the line
*/
LOCALDEF BOOLEAN is_preprocessor(char *line,int *start)
{
    int i;

    if (line == NULL) {
	return FALSE;
    };
    for (i = 0;line[i] != '\0';i += 1) {
	if (line[i] == '#') {
	    if (start != NULL) *start = i;
	    return TRUE;
	};
	if (!isspace(line[i])) break;	
    };
    return FALSE;
}

/* Months of the year */
LOCALVAR char *months[] = {
    "January","February","March","April",
    "May","June","July","August",
    "September","October","November","December",
    NULL
};

/* convert a date to a number
**
** Inputs:
**	year
**	month	    month as a string
**	day
** Outputs:
**	date	    date represented as year * 10000 + month * 100 + day
** Returns:
**	TRUE if the date was sucessfully converted
*/
LOCALDEF BOOLEAN dateToNumber(UINT32 *date,UINT16 year, char *month,UINT16 day)
{
    int iMonth;
    if (day > 31) return FALSE;
    if (day <= 0) return FALSE;
    if ((year < 99) && (year > 30)) year += 1900;
    if (year < 99) year += 2000;
    for (iMonth = 0;iMonth < DIM(months);iMonth += 1) {
	if (0 == strncmp(month,months[iMonth],3)) {
	    *date = (((UINT32)year) * 10000L) + (iMonth+1)*100 + day;
	    return TRUE;
	};
    };
    return FALSE;
}

/* convert a number to a date
**
** Inputs:
**	number
** Returns:
**	date represented as a string
*/
LOCALDEF char *dateFromNumber(UINT32 number)
{
    UINT32 day;
    UINT32 month;
    static char string[20];

    if (number == 0) return "Always";
    day = number % 100;
    month = (number / 100) % 100;
    month -= 1;
    month = month % 12;
    sprintf(string,"%ld %03s %ld",day,months[(UINT8)month],number/10000);
    return string;
}

/* Test to see if a line is a comment starting with a date.
**
** Inputs:
**	line	    line to look in
** Outputs:
**	date	    date represented as year * 10000 + month * 100 + day
** Returns:
**	TRUE if so FALSE if not
*/
LOCALDEF BOOLEAN isDateComment(char *line,UINT32 *date)
{
    int matches;
    UINT16 year;
    char month[STRING_LENGTH+1];
    UINT16 day;

    matches = sscanf(line," ** %d %3s %d",&day,month,&year);
    if (matches != 3) {
	matches = sscanf(line," /* %d %3s %d",&day,month,&year);
	if (matches != 3) {
	    matches = sscanf(line," // %d %3s %d",&day,month,&year);
	};
    };
    if (matches != 3) return FALSE;
    return dateToNumber(date,year,month,day);
}

/* Test to see if a line is a comment starting with the
** words:	Modification History
**
** Inputs:
**	line	    line to look in
** Returns:
**	TRUE if so FALSE if not
*/
LOCALDEF BOOLEAN isModificationComment(char *line)
{
    int matches;
    char name[STRING_LENGTH+1];
    char name2[STRING_LENGTH+1];

    matches = sscanf(line," ** %[0-9A-Za-z_] %[0-9A-Za-z_]",name,name2);
    if (matches != 2) {
	matches = sscanf(line," /* %[0-9A-Za-z_] %[0-9A-Za-z_]",name,name2);
	if (matches != 2) {
	    matches = sscanf(line," // %[0-9A-Za-z_] %[0-9A-Za-z_]",name,name2);
	};
    };
    if (matches != 2) return NULL;
    if (0 != io_compareName("MODIFICATION",name,STRING_LENGTH)) return FALSE;
    if (0 != io_compareName("HISTORY",name2,STRING_LENGTH)) return FALSE;
    return TRUE;
}

/* Test to see if a line matches a control keyword comment of the form:
**	<slash><star>	<keyword> : <value>
** or:
**	**		<keyword> : <value>
** or:
**	//		<keyword> : <value>
** If so return the value found. Otherwise return NULL.
**
** Inputs:
**	keyword	    keyword to match
**	line	    line to look in
** Returns:
**	pointer to value valid untill the next call to this routine
**	NULL if line does not match pattern or keyword does not match.
*/
LOCALDEF char *isKeywordComment(char *keyword,char *line)
{
    int matches;
    char name[STRING_LENGTH+1];
    static char value[STRING_LENGTH+1];

    matches = sscanf(line," ** %[0-9A-Za-z_] : %[0-9A-Za-z_]",name,value);
    if (matches != 2) {
	matches = sscanf(line," /* %[0-9A-Za-z_] : %[0-9A-Za-z_]",name,value);
	if (matches != 2) {
	    matches = sscanf(line," // %[0-9A-Za-z_] : %[0-9A-Za-z_]",
			     name,value);
	};
    };
    if (matches != 2) return NULL;
    if (0 == io_compareName(keyword,name,STRING_LENGTH)) {
	return value;
    };
    return NULL;
}

/* test to see if a line is a documentation end comment
**
** Inputs:
**	line		line to look in
** Returns:
**	TRUE if turns off documentation extract
*/
LOCALDEF BOOLEAN is_doc_end(char *line)
{
    int matches;
    char keyword[STRING_LENGTH+1];

    matches = sscanf(line," /* %[0-9A-Za-z_]",keyword);
    if (matches != 1) {
        matches = sscanf(line," ** %[0-9A-Za-z_]",keyword);
	if (matches != 1) return FALSE;
    };
    if (0 == io_compareName(keyword,"DOCUMENTATION_END",STRING_LENGTH)) {
	return TRUE;
    };
    return FALSE;
}

/* test to see if a line is a documentation start comment
**
** Inputs:
**	line		line to look in
** Returns:
**	TRUE if begins a documentation comment
*/
LOCALDEF BOOLEAN is_doc_start(char *line)
{
	char *value;
	
	value = isKeywordComment("DOCUMENTATION",line);
	if (value == NULL) return FALSE;
	if (0 == io_compareName(documentationType,"*",STRING_LENGTH)) return TRUE;
	if (0 == io_compareName(value,documentationType,STRING_LENGTH)) return TRUE;
	return FALSE;
}

/* test to see if a line starts a comment
**
** Inputs:
**	line		line to look in
** Returns:
**	TRUE if begins a documentation comment
*/
LOCALDEF BOOLEAN is_start_comment(char *line)
{
	int matches;
	char keyword[STRING_LENGTH+1];
	matches = sscanf(line," %s",keyword);
	if (matches != 1) return FALSE;
	if (0 == io_compareName(keyword,"/*",STRING_LENGTH)) {
	    return TRUE;
	};
	return FALSE;
}

/* copy a line to an output file. If it looks like a comment start
** <slash>* ** or <space>* strip that and other leading whitespace.
**
** Inputs:
**	line	line to process and output
*/
LOCALDEF void outDocLine(char *line)
{
	int i;
	if (docFile == NULL) return;
	if (line == NULL) return;
	i = 0;
	if (((line[0] == ' ') || (line[0] == '*') || (line[0] == '/'))
	     && (line[1] == '*')) {
	    i = strspn(line,"\t* /");
	};
	fputs(&line[i],docFile);
	return;
}

/* Extract the nth field from a string. The field is copied into a static
 * buffer that is valid untill the next call to this routine.
 * If there are insufficient fields in the string an empty string is returned.
 *
 * Inputs:
 *	seperators	characters used to seperate fields
 *	string		string to parse fields from
 *	field		field to extract from string
 *	ignoreMultiple	TRUE if multiple separators are counted as one
 * Returns:
 *	pointer to the string created
 */
LOCALDEF char *getField(char *seperators,char *string,int field,
			 BOOLEAN ignoreMultiple)
{
    static char temp[STRING_LENGTH+1];
    int i;
    int start;
    int length;

    temp[0] = NUL;
    if (seperators == NULL) seperators = " \t"; /* default is space and tab */
    start = 0;
    for (i = 1;i < field;i += 1) {	/* find beginning of field n */
	int mid;
	mid = strcspn(&string[start],seperators);
	start += mid;
	if (string[start] == NUL) return temp; /* not enough fields */
	start += 1;			/* step over seperator */
	if (ignoreMultiple) {
	    int j;
	    j = strspn(&string[start],seperators);
	    start += j;
	};
    };
    length = strcspn(&string[start],seperators);
    length = MIN(STRING_LENGTH,length);
    strncpy(temp,&string[start],length);
    temp[length] = NUL;
    return temp;
}

/* start at the end of a string and back up over any square braces
** and whitespace.
**
** Inputs:
**	string	    string to process
** Returns:
**	number of characters left at the front of the string
*/
LOCALDEF int back_over_squarebraces(char *string)
{
    int end;
    int braces = 0;
    int length = strlen(string);
    for (end = length;end > 0;end -= 1) {
	if (isspace(string[end-1])) continue;
	if (string[end-1] == CLOSE_SQUARE_BRACE) {
	    braces += 1;
	    continue;
	};
	if (string[end-1] == OPEN_SQUARE_BRACE) {
	    braces -= 1;
	    if (braces < 0) {
		out_message("squarebrace without partner []",string,TRUE,TRUE);
		return length;
	    };
	    continue;
	};
	if (braces > 0) continue;
	break;
    };
    if (braces != 0) {
	out_message("unmatched square braces []",string,TRUE,TRUE);
	return length;
    };
    return end;
}
    

/* get the name portion of the nth argument to a subroutine
**
** Inputs:
**	string	    string to parse
**	idx	    argument to get
** Returns:
**	name portion of the nth argument valid till the next call to this
**	routine, NULL if no such argument
*/
LOCALDEF char *get_nth_arg_name(char *string,int idx)
{
    static char name[STRING_LENGTH+1];
    char *field;
    int i;
    int end;

    name[0] = '\0';
    if (string == NULL) return NULL;
    if (idx <= 0) return NULL;
    field = getField(",",string,idx,FALSE);
    if (field == NULL) return NULL;
    if (field[0] == '\0') return NULL;
    end = back_over_squarebraces(field);
    if (end <= 0) return NULL;		/* empty field */
    if ((end == 3) && (0 == strncmp(field,"...",3))) {
	return "...";
    };
    for (i = end-1;i >= 0;i -= 1) {
	if (isalnum(field[i])) continue;
	if (field[i] == '_') continue;
	i += 1;
	break;
    };
    if (i < 0) i = 0;
    strncpy(name,&field[i],end-i);
    name[end-i] = '\0';
    return name;
}

/* Find termination of a header and remove the generic termination
** replacing it with an end of statement
**
** Inputs:
**	header	    header to modify, modification is done in place
**	term	    terminators
*/
LOCALDEF void terminate_header(char *header,char *term)
{
    char *stop;
    stop = strpbrk(header,term);
    if (stop == NULL) {
	out_message("BUG: lost track of terminator",header,TRUE,TRUE);
    } else {
	while ((stop > header) &&
	       ((*(stop-1) == '\n') || isspace(*(stop-1)))) {
	    stop -= 1;
	};
	strcpy(stop,";\n");
    };
}

/* translate a header from ANSI C syntax to pre ansi C syntax
** This routines handles arguments seperated by commas and a trailing
** argument terminated by a balanced close paren.
**
** If the argument list is explicitly void the void is removed.
**
** If there are arguments the arguments within the parens are replaced
** by a simple comma seperated list of argument names. After the close paren
** a new line is started and each argument is declared on a single line
** with its type.
**
** Arguments which are arrays are handled. Arguments which terminate with
** close parens are not handled.
**
** Inputs:
**	header	string to translate, some modifications are made in place
**	ofile	file to put the translated header in
**	checkKeyword	TRUE to validate the keyword
*/
LOCALDEF void translate_header(char *header,FILE *ofile,BOOLEAN checkKeyword)
{
    int idx;
    int start;
    int end;
    int parens;
    int length = strlen(header);
    BOOLEAN any_declarations_output;
    char *keyword;
    BOOLEAN argumentsRequired = TRUE;

    if (header == NULL) return;
    if (ofile == NULL) return;		/* no output file */
    keyword = first_keyword(header);
    if ((0 == strcmp(keyword,"LOCALVAR")) ||
	(0 == strcmp(keyword,"/*")) ||
	(0 == strcmp(keyword,"*/")) ||
	(0 == strcmp(keyword,"**")) ||
	(0 == strcmp(keyword,"EXTERNVAR"))) {
	/* data declarations need no translation */
	fputs(header,ofile);
	return;
    };
    if (!((0 == strcmp(keyword,"LOCALDEF")) ||
	  (0 == strcmp(keyword,"LOCALREF")) ||
	  (0 == strcmp(keyword,"EXTERNREF")) ||
	  (0 == strcmp(keyword,"EXTERNDEF")))) {
	if (checkKeyword) out_message("Unexpected keyword",header,TRUE,TRUE);
	fputs(header,ofile);
	return;
    };
    if ((0 == strcmp(keyword,"LOCALREF")) ||
	  (0 == strcmp(keyword,"EXTERNREF"))) {
	argumentsRequired = FALSE;
    };
    /* walk the string from the end looking for the argument list */
    for (idx = length-1;idx > 0;idx -= 1) {
	if (header[idx] == CLOSE_PAREN) {
	    break;
	};
    };
    if (idx == 0) {
	if (argumentsRequired)
	    out_message("could not find subroutine arguments",header,TRUE,TRUE);
	fputs(header,ofile);
	return;
    };
    end = idx;				/* the last paren */
    parens = 1;
    for (idx -= 1;idx > 0;idx -= 1) {
	if (header[idx] == OPEN_PAREN) {
	    parens -= 1;
	    if (parens <= 0) break;
	};
	if (header[idx] == CLOSE_PAREN) {
	    parens += 1;
	    continue;
	};
    };
    if (parens != 0) {
	out_message("could not find beginning of argument list",header,
		    TRUE,TRUE);
	fputs(header,ofile);
	return;
    };
    start = idx;
    fwrite(header,1,start,ofile);	/* stuff before the arg list */
    fputc(OPEN_PAREN,ofile);
    header[end] = '\0';			/* terminate at trailing paren */
    if ((0 == strcmp(keyword,"LOCALREF")) ||
	(0 == strcmp(keyword,"EXTERNREF"))) {
	/* look for single argument of (void) and remove it */
	BOOLEAN flushArgList = FALSE;
	if (preAnsiFlag) flushArgList = TRUE;
	if (!flushArgList) {
	    fputs(&header[start+1],ofile); /* put out the whole arg list */
	};
	fputc(CLOSE_PAREN,ofile);
	header[end] = CLOSE_PAREN;		/* put trailing paren back */
	fwrite(&header[end+1],1,length-end-1,ofile);
	return;
    };
    idx = 1;
    while (TRUE) {
	/* EXTERNDEF or LOCALDEF get here */
	char *nth;
	nth = get_nth_arg_name(&header[start+1],idx);
	if (nth == NULL) break;
	if (nth[0] == '\0') {
	    if (idx != 1) {
		char string[STRING_LENGTH+1];
		sprintf(string,"no name for argument number %d",idx);
		out_message(string,header,TRUE,TRUE);
	    };
	};
	if (0 == strcmp(nth,"void")) {
	    nth = get_nth_arg_name(&header[start+1],idx+1);
	    if (nth != NULL) {
		out_message("further arguments not allowed after void",header,
			    TRUE,TRUE);
	    };
	    break;
	};
	if (0 == strcmp(nth,"...")) {
	    nth = get_nth_arg_name(&header[start+1],idx+1);
	    if (nth != NULL) {
		out_message("... was not the last argument",header,TRUE,TRUE);
	    };
	    break;
	};
	if (idx != 1) fputc(',',ofile);
	fputs(nth,ofile);
	idx += 1;
    };
    fputc(CLOSE_PAREN,ofile);
    idx = 1;
    any_declarations_output = FALSE;
    while (TRUE) {
	char *nth;
	nth = getField(",",&header[start+1],idx,FALSE);
	if (nth == NULL) break;
	if (nth[0] == '\0') break;
	idx += 1;
	if (0 == strcmp(nth,"")) {
	    continue;
	};
	if (0 == strcmp(nth,"void"))
	    continue;
	if (0 == strcmp(nth,"..."))
	    break;
	if (idx == 2) {
	    fputc('\n',ofile);		/* first time start line */
	};
	any_declarations_output = TRUE;
	fputs(nth,ofile);		/* declaration */
	fputs(";\n",ofile);		/* terminator */
    };
    header[end] = CLOSE_PAREN;		/* put trailing paren back */
    if (any_declarations_output && (header[end+1] == '\n')) {
	end += 1;
    };
    fwrite(&header[end+1],1,length-end-1,ofile);
    return;
}

/* remove control-Z characters from a string, the string is modified in place.
**
** Inputs:
**	string	    string to modify
** Returns:
**	TRUE if any were found
*/
LOCALDEF BOOLEAN removeControlZs(char *string)
{
    BOOLEAN anyFound = FALSE;
    int i;

    if (string == NULL) return FALSE;
    for (i = 0;string[i] != '\0';i += 1) {
	if (string[i] == 26) {	/* control-Z */
	    memmove(&string[i],&string[i+1],
		strlen(&string[i]));	/* one extra! (to get terminator) */
	    anyFound = TRUE;
	};
    };
    return anyFound;
}

/* read a source file. look for comments with documentation flag
** put them in documentation file. Look also for entrys put them in
** entrys file.
**
**  Inputs:
**	infile	file to read
*/
LOCALDEF void parseCode(char *infile)
{
    FILE *in;
    enum { REFERENCE,GLOBAL_SUBROUTINE,
	    LOCAL_SUBROUTINE,GLOBAL_VARIABLE,LOCAL_VARIABLE} headerType;
    char line[STRING_LENGTH+1];
    BOOLEAN in_comment = FALSE;
    BOOLEAN emittingAllToDoc = FALSE;
    BOOLEAN emittingAllToHeader = FALSE;
    BOOLEAN in_header = FALSE;
    BOOLEAN inModificationHistory = FALSE;
    BOOLEAN sawModificationHistory = FALSE;
    char arg_header[5*STRING_LENGTH+1];
    char full_header[5*STRING_LENGTH+1];
    int start;
    char *keyword;
    INT32 lastTripleReported = -1;
    char *headerComment;		/* comment associated with header */
    int hcLength = 20*STRING_LENGTH;
    int hcUsed = 0;
    static BOOLEAN mallocFailed = FALSE;

    if (mallocFailed) return;
    if (infile == NULL) return;
    input_filename = infile;
    in = fopen(input_filename,"r");
    if (in == NULL) {
	printf("Could not open input file: %s\n",input_filename);
	error_count += 1;
	return;
    };
    if (entrys == NULL) {
	/* no output file yet start one */
	strncpy(entrys_filename,root_filename,STRING_LENGTH);
	entrys_filename[STRING_LENGTH] = NUL;
	strncat(entrys_filename,".xh",DIM(entrys_filename));
	entrys = fopen(entrys_filename,"w");
    };
    if (entrys == NULL) {
	printf("%s 1: Could not open generated include file\n",
	       entrys_filename);
	printf("Ignoring %s\n",input_filename);
	error_count += 1;
	fclose(in);
	return;
    } else {
	fprintf(entrys,"/* Extracted by MAKEDOC from %s */\n\n",
		input_filename);
	fprintf(entrys,"#ifndef __%s_XH__\n#define __%s_XH__\n",
		root_filename,root_filename);
    };
    if (emitDocFile && (docFile == NULL)) {
	if (doc_filename[0] == '\0') {
	    sprintf(doc_filename,"%s.xd",root_filename);
	};
	docFile = fopen(doc_filename,"w");
	if (docFile == NULL) {
	    printf("%s 1: Could not open documentation file\n",doc_filename);
	    error_count += 1;
	} else {
	    fprintf(docFile,"/* %s file extracted by MAKEDOC from %s */\n\n",
		    doc_filename,input_filename);
	};
    };
    headerComment = (char *)malloc(hcLength);
    if (headerComment == NULL) {
	mallocFailed = TRUE;
	fclose(in);
	return;
    };
    line_number = 0;
    arg_header[sizeof(arg_header) - 1] = '\0';
    line[STRING_LENGTH] = '\0';
    while (NULL != fgets(line,STRING_LENGTH,in)) {
	char *emitAllKeyword;
	line_number += 1;
	if (eliminateControlZsFlag) {
	    removeControlZs(line);
	};
	if (is_blank(line)) {
	    /* comments associated with subroutines must touch the routine */
	    hcUsed = 0;
	    headerComment[0] = '\0';
	};
	if (NULL != strstr(line,"??" "?")) {
	    out_message(NULL,line,FALSE,
			(line_number != (lastTripleReported+1)));
	    lastTripleReported = line_number;
	};
	if (in_comment) {
	    if (in_header) {
		out_message("Comment within header",line,TRUE,TRUE);
	    };
	};
	emitAllKeyword = isKeywordComment("MAKEDOC",line);
	if (emitAllKeyword != NULL) {
	    if (0 == io_compareName(emitAllKeyword,"EMIT_TO_HEADER_OFF",
				    STRING_LENGTH)) {
		emittingAllToHeader = FALSE;
	    };
	};
	if (emittingAllToHeader) {
	    if (entrys != NULL) fputs(line,entrys);
	};
	if (emitAllKeyword != NULL) {
	    if (0 == io_compareName(emitAllKeyword,"EMIT_TO_HEADER_ON",
				    STRING_LENGTH)) {
		emittingAllToHeader = TRUE;
	    };
	};
	if (is_doc_end(line)) {
	    if (emittingAllToDoc) {
		if (docFile != NULL) fputs("\n\n",docFile);
	    };
	    emittingAllToDoc = FALSE;
	};
	if (emittingAllToDoc) {
	    outDocLine(line);
	};
	if (is_start_comment(line)) {
	    if (is_doc_start(line)) {
		emittingAllToDoc = TRUE;
	    };
	    if (!is_end_comment(line)) {
		in_comment = TRUE;
	    };
	};
	if (inModificationHistory && (historySince != 0)) {
	    UINT32 date;
	    if (isDateComment(line,&date)) {
		if (date < historySince) inModificationHistory = FALSE;
	    };
	};
	if (inModificationHistory) {
	    if (!emittingAllToDoc && emitHistory) {
		outDocLine(line);
	    };
	};
	if (isModificationComment(line)) {
	    char string[STRING_LENGTH+1];
	    sprintf(string,"\nHistory since %s from %s:\n",
		    dateFromNumber(historySince),
		    input_filename);
	    outDocLine(string);
	    inModificationHistory = TRUE;
	    sawModificationHistory = TRUE;
	};
	if (in_comment && (docFile != NULL)) {
	    /* accumulate comments for documentation file */
	    int lineLength = strlen(line);
	    char *newComment;
	    if ((hcUsed + lineLength + 1) > hcLength) {
		hcLength += lineLength + 1;
		newComment = (char *)malloc(hcLength);
		if (newComment == NULL) {
		    free(headerComment);
		    mallocFailed = TRUE;
		    printf("Could not allocate %d bytes of memory\n",hcLength);
		    fclose(in);
		    return;
		};
		strncpy(newComment,headerComment,hcUsed);
		newComment[hcUsed] = NUL;
		free(headerComment);
		headerComment = newComment;
	    };
	    strcpy(&headerComment[hcUsed],line);
	    hcUsed += lineLength;
	};
	if (is_end_comment(line)) {
	    in_comment = FALSE;
	    inModificationHistory = FALSE;
	};
	keyword = first_keyword(line);
	if (0 == strcmp(keyword,"EXTERNDEF")) {
	    if (in_header) {
		out_message("found header within another header",line,
			    TRUE,TRUE);
	    } else {
		arg_header[0] = '\0';
		headerType = GLOBAL_SUBROUTINE;
	    };
	    in_header = TRUE;
	    if (in_comment) {
		out_message("EXTERNDEF within comment",line,TRUE,TRUE);
	    };
	} else if (0 == strcmp(keyword,"EXTERNVAR")) {
	    if (in_header) {
		out_message("found header within another header",line,
			    TRUE,TRUE);
	    } else {
		arg_header[0] = '\0';
		headerType = GLOBAL_VARIABLE;
	    };
	    in_header = TRUE;
	    if (in_comment) {
		out_message("EXTERNVAR within comment",line,TRUE,TRUE);
	    };
	} else if ((0 == strcmp(keyword,"LOCALDEF")) ||
		   (0 == strcmp(keyword,"LOCALVAR"))) {
	    if (in_header) {
		out_message("found header within another header",line,
			    TRUE,TRUE);
	    } else {
		arg_header[0] = '\0';
		if (0 == strcmp(keyword,"LOCALDEF")) {
		    headerType = LOCAL_SUBROUTINE;
		} else {
		    headerType = LOCAL_VARIABLE;
		};
	    };
	    in_header = TRUE;
	    if (in_comment) {
		out_message("LOCALDEF or LOCALVAR within comment",line,
			    TRUE,TRUE);
	    };
	} else if ((0 == strcmp(keyword,"LOCALREF")) ||
		   (0 == strcmp(keyword,"EXTERNREF"))) {
	    if (in_header) {
		out_message("found header within another header",line,
			    TRUE,TRUE);
	    } else {
		arg_header[0] = '\0';
		headerType = REFERENCE;
	    };
	    in_header = TRUE;
	    if (in_comment) {
		out_message("LOCALREF or EXTERNREF  within comment",line,
			    TRUE,TRUE);
	    };
	};
	if (is_preprocessor(line,&start)) {
	    if (in_header) {
		out_message("preprocessor lines not allowed in header",line,
			    TRUE,TRUE);
	    } else {
		/* outdent indented preprocessor commands */
		if (code != NULL) fputs(&line[start],code);
		continue;
	    };
	};
	if (in_header) {
	    BOOLEAN terminated;
	    char *stop;
	    char *term = ";{="; /* "}" so paren counters dont get confused */
	    strncat(arg_header,line,(sizeof(arg_header) - 1));
	    if (strlen(arg_header) >= (sizeof(arg_header) - 1)) {
		out_message("subroutine header too long for translation",
			      line,TRUE,TRUE);
	    };
	    stop = strpbrk(line,term);
	    terminated = (stop != NULL);
	    if (terminated) {
		if (code != NULL) {
		    strcpy(full_header,arg_header);
		    translate_header(full_header,code,TRUE);
		};
		if ((entrys != NULL) &&
		    ((headerType == GLOBAL_VARIABLE)
		     || (headerType == GLOBAL_SUBROUTINE))) {
		    terminate_header(arg_header,term);
		    keyword_replace(arg_header,"EXTERNDEF","EXTERNREF");
		    keyword_replace(arg_header,"EXTERNVAR","EXTERNREF");
		    translate_header(arg_header,entrys,FALSE);
		};
		in_header = FALSE;
		if (docFile != NULL) {
		    /* emit header and comment */
		    BOOLEAN emit = FALSE;
		    switch (headerType) {
			case GLOBAL_SUBROUTINE:
			    if (emitGlobalEntrys) emit = TRUE;
			    /* put back name */
			    keyword_replace(arg_header,
					    "EXTERNREF","EXTERNDEF");
			    break;
			case LOCAL_SUBROUTINE:
			    if (emitLocalEntrys) emit = TRUE;
			    break;
			case GLOBAL_VARIABLE:
			    /* put back name */
			    keyword_replace(arg_header,
					    "EXTERNREF","EXTERNVAR");
			    if (emitGlobalVariables) emit = TRUE;
			    break;
			case LOCAL_VARIABLE:
			    if (emitLocalVariables) emit = TRUE;
			    break;
			case REFERENCE:
			default:
			    emit = FALSE;
			    break;
		    };
		    if (emit) {
			terminate_header(arg_header,term);
			fputs(arg_header,docFile);
			if (((headerType == GLOBAL_SUBROUTINE) ||
			     (headerType == LOCAL_SUBROUTINE)) &&
			    (headerComment[0] == '\0')) {
			    out_message("Missing comment for declaration",
					arg_header,TRUE,TRUE);
			    fputs("?Missing comment\n",docFile);
			};
			if (headerComment[0] != '\0') {
			    fputs(headerComment,docFile);
			};
			fputs("\n",docFile);
		    };
		    /* only put out a comment once, with its header */
		    hcUsed = 0;
		    headerComment[0] = '\0';
		};
	    };
	} else {
	    if (code != NULL) fputs(line,code);
	};
    };
    fclose(in);
    if (in_header) {
	out_message("header is not terminated",arg_header,TRUE,TRUE);
	if (code != NULL) fputs(arg_header,code);
    };
    free(headerComment);
    if (!sawModificationHistory) {
	out_message("Modification History not found",NULL,TRUE,TRUE);
    };
    printf("%d lines in module %s\n",line_number,input_filename);
    total_lines += line_number;
    files_read += 1;
    return;
}
