/* This file contains functions which output TSAPI event (message) information
** using the services provided by IOSUBS.C
**
** Copyright (c) by NetPhone, Inc. 1997, All rights reserved.
** Copyright (c) by NetPhone, Inc. 1996, All rights reserved.
**
** MAKEDOC:EMIT_TO_HEADER_ON
#define tsdumpsEdit 26
** MAKEDOC:EMIT_TO_HEADER_OFF
**
** Modification History:
**   6Nov97 MJNewman 26: Put out newline for caller ID reports.
**  29Oct97 PJConley 25: Update span info enums to match new nb_privat.h
**   7Oct97 MJNewman 24: Report call info structure.
**  29Sep97 MJNewman 23: Remove common prefixes from many TSAPI names to
**		reduce the string storage required. An obscure error talking of
**		being unable to start the device occured. Add expected digits.
**  26Sep97 MJNewman 22: Add lineStyle, moreLineInfo and digital span info.
**   9Sep97 MJNewman 21: Add new field to system info.
**   5Sep97 MJNewman 20: Add reports of system and board information.
**  15Apr97 MJNewman 19: Report flash timeout.
**  14Apr97 MJNewman 18: Report more fields of line info and callerID info.
**   6Mar97 MJNewman 17: Report more for callerID info
**  12Feb97 MJNewman 16: Remove edit global symbol. Make first line production
**		common code.
**  10Feb97 VABerry  15: Report noAnswerTime correctly as tenths.
**  12Jan97 MJNewman 14: Report more things in callerID info.
**  18Dec96 MJNewman 13: Added report of ROMboardOrder
**  18Dec96 VABerry  12: Modified to support change in enums for Sound Support
**  18Dec96 VABerry  11: Added support for setting and displaying Filter Bit
**		fields.
**   1Nov96 MJNewman 10: Add report of stream info.
**  29Oct96 MJNewman 9:	Report callerID time with syntax. No info if missing.
**		Timestamps.
**  22Oct96 MJNewman 8:	Update private message output.
**  16Aug96 MJNewman 7:	Get operator number.
**  12Aug96 MJNewman 6:	Fix originated report of called number.
**  26Jun96 MJNewman 5:	Fix pointer to raw data.
**  26Jun96 MJNewman 4:	Raw callerID report.
**  24Jun96 MJNewman:	Add CallerID info.
**   5May96 MJNewman:	Created.
*/

#include <ctype.h>
#include <time.h>

#include <windows.h>

/* Sometimes required for UNIX platforms */
#include <sys/types.h>

#include "acs.h"
#include "csta.h"

#include "basetype.h"
#include "iosubs.h"
#include "nbprivat.h"

#include "iosubs.xh"
#include "tsdumps.xh"
#include "query.xh"

/* Log the version of the board firmware here. */
LOCALVAR UINT16 firmwareMajor = 2;
LOCALVAR UINT16 firmwareMinor = 2;
LOCALVAR UINT16 firmwareEdit = 0;

FILTER FilterString::callFilter[] = { 
    {"CF_NONE",			"CF_CLEAR",	0x0000},
    {"CF_BIT",			"",		0x0000},
    {"CF_BIT1",			"",		0x0001},
    {"CF_BIT2",			"",		0x0002},
    {"TRANSFERRED",		"CFT",		CF_TRANSFERRED},
    {"SERVICE_INITIATED",	"CFSI",		CF_SERVICE_INITIATED},
    {"RETRIEVED",		"CFR",		CF_RETRIEVED},
    {"QUEUED",			"CFQ",		CF_QUEUED},
    {"ORIGINATED",		"CFO",		CF_ORIGINATED},
    {"NETWORK_REACHED",		"CFNR",		CF_NETWORK_REACHED},
    {"HELD",			"CFH",		CF_HELD},
    {"FAILED",			"CFF",		CF_FAILED},
    {"ESTABLISHED",		"CFE",		CF_ESTABLISHED},
    {"DIVERTED",		"CFDI",		CF_DIVERTED},
    {"DELIVERED",		"CFDE",		CF_DELIVERED},
    {"CONNECTION_CLEARED",	"CFCOC",	CF_CONNECTION_CLEARED},
    {"CONFERENCED",		"CFC",		CF_CONFERENCED},
    {"CALL_CLEARED",		"CFCAC",	CF_CALL_CLEARED},
    {NULL,			NULL,		0}
};


FILTER FilterString::featureFilter[] = {
    {"FF_NONE",			"FF_CLEAR",	0x00},
    {"FF_BIT",			"",		0x00},
    {"FF_BIT1",			"",		0x01},
    {"FF_BIT2",			"",		0x02},
    {"FF_BIT3",			"",		0x04},
    {"FF_BIT4",			"",		0x08},
    {"MESSAGE_WAITING",		"FFMW",		FF_MESSAGE_WAITING},
    {"FORWARDING",		"FFF",		FF_FORWARDING},
    {"DO_NOT_DISTURB",		"FFDND",	FF_DO_NOT_DISTURB},
    {"CALL_INFORMATION",	"FFCI",		FF_CALL_INFORMATION},
    {NULL,			NULL,		0}
};

FILTER FilterString::agentFilter[] = {
    {"AF_NONE",			"AF_CLEAR",	0x00},
    {"AF_BIT",			"",		0x00},
    {"AF_BIT1",			"",		0x01},
    {"AF_BIT2",			"",		0x02},
    {"WORK_READY",		"AFWR",		AF_WORK_READY},
    {"WORK_NOT_READY",		"AFWNR",	AF_WORK_NOT_READY},
    {"READY",			"AFR",		AF_READY},
    {"NOT_READY",		"AFNR",		AF_NOT_READY},
    {"LOGGED_OFF",		"AFLOFF",	AF_LOGGED_OFF},
    {"LOGGED_ON",		"AFLON",	AF_LOGGED_ON},
    {NULL,			NULL,		0}
};

FILTER FilterString::maintenanceFilter[] = {
    {"MF_NONE",			"MF_CLEAR",	0x00},
    {"MF_BIT",			"",		0x00},
    {"MF_BIT1",			"",		0x01},
    {"MF_BIT2",			"",		0x02},
    {"MF_BIT3",			"",		0x04},
    {"MF_BIT4",			"",		0x08},
    {"MF_BIT5",			"",		0x10},
    {"MF_BIT6",			"",		0x20},
    {"BACK_IN_SERVICE",		"MFBIS",	MF_BACK_IN_SERVICE},
    {"OUT_OF_SERVICE",		"MFOOS",	MF_OUT_OF_SERVICE},
    {NULL,			NULL,		0}
};

FILTER FilterString::privateFilter[] = {
    {"PF_NONE",			"PF_CLEAR",	0x00000000L},
    {"PF_BIT",			"",		0x00000000L},
    {"BIT1",			"",		0x00000001L},
    {"DTMF",			"",		0x00000002L},
    {"BIT3",			"",		0x00000004L},
    {"BIT4",			"",		0x00000008L},
    {"BIT5",			"",		0x00000010L},
    {"BIT6",			"",		0x00000020L},
    {"BIT7",			"",		0x00000040L},
    {"BIT8",			"",		0x00000080L},
    {"BIT9",			"",		0x00000100L},
    {"BIT10",			"",		0x00000200L},
    {"BIT11",			"",		0x00000400L},
    {"BIT12",			"",		0x00000800L},
    {"BIT13",			"",		0x00001000L},
    {"BIT14",			"",		0x00002000L},
    {"BIT15",			"",		0x00004000L},
    {"BIT16",			"",		0x00008000L},
    {"BIT17",			"",		0x00010000L},
    {"BIT18",			"",		0x00020000L},
    {"BIT19",			"",		0x00040000L},
    {"BIT20",			"",		0x00080000L},
    {"BIT21",			"",		0x00100000L},
    {"BIT22",			"",		0x00200000L},
    {"BIT23",			"",		0x00400000L},
    {"BIT24",			"",		0x00800000L},
    {"BIT25",			"",		0x01000000L},
    {"BIT26",			"",		0x02000000L},
    {"BIT27",			"",		0x04000000L},
    {"BIT28",			"",		0x08000000L},
    {"BIT29",			"",		0x10000000L},
    {"BIT30",			"",		0x20000000L},
    {"BIT31",			"",		0x40000000L},
    {"BIT32",			"",		0x80000000L},
    {NULL,			NULL,		0}
};

char *FilterString::getFilterString(int type, unsigned long filter)
{
    int i;
    filstring[0] = NUL;			/* initialize string */
    switch(type) {
	case FILTER_CALL:
	    for (i = 2; callFilter[i].string != NULL; i++) {
		if (filter & callFilter[i].mask) {
		    strcat(filstring, callFilter[i].string);
		    strcat(filstring, "|");
		};
	    };
	    filstring[strlen(filstring)-1] = NUL; /* strip last | */
	    return filstring;
	    break;
	case FILTER_FEATURE:
	    for (i = 2; featureFilter[i].string != NULL; i++) {
		if (filter & featureFilter[i].mask) {
		    strcat(filstring, featureFilter[i].string);
		    strcat(filstring, "|");
		};
	    };
	    filstring[strlen(filstring)-1] = NUL; /* strip last | */
	    return filstring;
	    break;
	case FILTER_AGENT:
	    for (i = 2; agentFilter[i].string != NULL; i++) {
		if (filter & agentFilter[i].mask) {
		    strcat(filstring, agentFilter[i].string);
		    strcat(filstring, "|");
		};
	    };
	    filstring[strlen(filstring)-1] = NUL; /* strip last | */
	    return filstring;
	    break;
	case FILTER_MAINTENANCE:
	    for (i = 2; maintenanceFilter[i].string != NULL; i++) {
		if (filter & maintenanceFilter[i].mask) {
		    strcat(filstring, maintenanceFilter[i].string);
		    strcat(filstring, "|");
		};
	    };
	    filstring[strlen(filstring)-1] = NUL; /* strip last | */
	    return filstring;
	    break;
	case FILTER_PRIVATE:
	    for (i = 2; privateFilter[i].string != NULL; i++) {
		if (filter & privateFilter[i].mask) {
		    strcat(filstring, privateFilter[i].string);
		    strcat(filstring, "|");
		};
	    };
	    filstring[strlen(filstring)-1] = NUL; /* strip last | */
	    return filstring;
	    break;
	    
	default:
	    return NULL;
	    break;
    };
    return NULL;
}

int FilterString::setFilter(char *filter, long bit)
{
    int i;

    /* Check call filters */
    if (!strcmp(filter, callFilter[0].string)) {
	globalFilter.call = 0x0000;
	return 0;
    };
    for (i = 0; callFilter[i].string != NULL; i++) {
	if ((!strcmp(filter, callFilter[i].string)) ||
	    (!strcmp(filter, callFilter[i].abbreviation))){
	    if (!strcmp(filter, "CF_BIT")) {
		callFilter[i].mask = 1 << ((int)bit - 1);
	    };
	    if (globalFilter.call & (unsigned short) callFilter[i].mask) {
		globalFilter.call &= (unsigned short) ~callFilter[i].mask;
	    }
	    else {
		globalFilter.call |= (unsigned short) callFilter[i].mask;
	    };
	    return 0;
	};
    };
    /* Check feature filters */
    if (!strcmp(filter, featureFilter[0].string)) {
	globalFilter.feature = 0x00;
	return 0;
    };
    for (i = 0; featureFilter[i].string != NULL; i++) {
	if ((!strcmp(filter, featureFilter[i].string)) ||
	    (!strcmp(filter, featureFilter[i].abbreviation))){
	    if (!strcmp(filter, "FF_BIT")) {
		featureFilter[i].mask = (1 << ((int)bit-1));
	    };
	    if (globalFilter.feature & (unsigned char) featureFilter[i].mask) {
		globalFilter.feature &= (unsigned char) ~featureFilter[i].mask;
	    }
	    else {
		globalFilter.feature |= (unsigned char) featureFilter[i].mask;
	    };
	    return 0;
	};
    };
    /* Process Agent Filters */
    if (!strcmp(filter, agentFilter[0].string)) {
	globalFilter.agent = 0x00;
	return 0;
    };
    for (i = 0; agentFilter[i].string != NULL; i++) {
	if ((!strcmp(filter, agentFilter[i].string)) ||
	    (!strcmp(filter, agentFilter[i].abbreviation))){
	    if (!strcmp(filter, "AF_BIT")) {
		agentFilter[i].mask = 1 << ((int)bit-1);
	    };
	    if (globalFilter.agent & (unsigned char) agentFilter[i].mask) {
		globalFilter.agent &= (unsigned char) ~agentFilter[i].mask;
	    }
	    else {
		globalFilter.agent |= (unsigned char) agentFilter[i].mask;
	    };
	    return 0;
	};
    };
    /* Process Maintenance Filters */
    if (!strcmp(filter, maintenanceFilter[0].string)) {
	globalFilter.maintenance = 0x00;
	return 0;
    };
    for (i = 0; maintenanceFilter[i].string != NULL; i++) {
	if ((!strcmp(filter, maintenanceFilter[i].string)) ||
	    (!strcmp(filter, maintenanceFilter[i].abbreviation))){
	    if (!strcmp(filter, "MF_BIT")) {
		maintenanceFilter[i].mask = 1 << ((int)bit-1);
	    };
	    if (globalFilter.maintenance & (unsigned char) maintenanceFilter[i].mask) {
		globalFilter.maintenance &= (unsigned char) ~maintenanceFilter[i].mask;
	    }
	    else {
		globalFilter.maintenance |= (unsigned char) maintenanceFilter[i].mask;
	    };
	    return 0;
	};
    };
    /* Process Private Filters */
    if (!strcmp(filter, privateFilter[0].string)) {
	globalFilter.privateFilter = 0x0000;
	return 0;
    };
    for (i = 0; privateFilter[i].string != NULL; i++) {
	if ((!strcmp(filter, privateFilter[i].string)) ||
	    (!strcmp(filter, privateFilter[i].abbreviation))){
	    if (!strcmp(filter, "PF_BIT")) {
		privateFilter[i].mask = 1L << ((int)bit-1);
	    };
	    if (globalFilter.privateFilter & privateFilter[i].mask) {
		globalFilter.privateFilter & ~privateFilter[i].mask;
	    }
	    else {
		globalFilter.privateFilter |= privateFilter[i].mask;
	    };
	    return 0;
	};
    };
    /* Not what we wanted */
    return -1;
}

/* column for time printout */
const int timeColumn =  40;

// This ACS handle is considered local
LOCALVAR ACSHandle_t localACShandle = 0;

// put out an invokeID stamp on each message when true
LOCALVAR BOOLEAN invokeIDstampsOn = TRUE;

/* Set the invokeIDstamps flag
**
** Inputs:
**	flag	new value for the flag
** Returns:
**	old value of the flag.
*/
EXTERNDEF BOOLEAN o_enableInvokeIDstamps(BOOLEAN flag)
{
    BOOLEAN temp;
    temp = invokeIDstampsOn;
    invokeIDstampsOn = flag;
    return temp;
}

// put out an invokeID delay on each message when true
LOCALVAR BOOLEAN invokeIDdelayOn = TRUE;

/* Set the invokeIDstamps flag
**
** Inputs:
**	flag	new value for the flag
** Returns:
**	old value of the flag.
*/
EXTERNDEF BOOLEAN o_enableInvokeIDdelayStamps(BOOLEAN flag)
{
    BOOLEAN temp;
    temp = invokeIDdelayOn;
    invokeIDdelayOn = flag;
    return temp;
}

// put out a timestamp on each message when true
LOCALVAR BOOLEAN timestampsOn = FALSE;

/* Set the timestamps flag
**
** Inputs:
**	flag	new value for the flag
** Returns:
**	old value of the flag.
*/
EXTERNDEF BOOLEAN o_enableTimestamps(BOOLEAN flag)
{
    BOOLEAN temp;
    temp = timestampsOn;
    timestampsOn = flag;
    return temp;
}

/* Sets what is considered the local acs handle
**
** Inputs:
**	handle
*/
EXTERNDEF void o_setLocalACShandle(ACSHandle_t handle)
{
    localACShandle = handle;
    return;
}

// Array of ACS type names. The real ACS name starts with ACS_ then these names.
LOCALVAR char* ACSTypeNames[] =
			       {
    "OPEN_STREAM",			/* 1 */
    "OPEN_STREAM_CONF",
    "CLOSE_STREAM",
    "CLOSE_STREAM_CONF",
    "ABORT_STREAM",
    "UNIVERSAL_FAILURE_CONF",
    "UNIVERSAL_FAILURE",
    "KEY_REQUEST",
    "KEY_REPLY",
    "NAME_SRV_REQUEST",
    "NAME_SRV_REPLY",
    "?unknown?"
			       };

// Array of names of CSTA type. The real CSTA name starts with CSTA_ then these
// strings.
LOCALVAR char *CSTATypeNames[] = {
    "ALTERNATE_CALL",			/* 1 */
    "ALTERNATE_CALL_CONF",		/* 2 */ 
    "ANSWER_CALL",			/* 3 */
    "ANSWER_CALL_CONF",			/* 4 */
    "CALL_COMPLETION",			/* 5 */
    "CALL_COMPLETION_CONF",		/* 6 */
    "CLEAR_CALL",			/* 7 */
    "CLEAR_CALL_CONF",			/* 8 */
    "CLEAR_CONNECTION",			/* 9 */
    "CLEAR_CONNECTION_CONF",		/* 10 */
    "CONFERENCE_CALL",			/* 11 */
    "CONFERENCE_CALL_CONF",		/* 12 */
    "CONSULTATION_CALL",		/* 13 */
    "CONSULTATION_CALL_CONF",		/* 14 */
    "DEFLECT_CALL",			/* 15 */
    "DEFLECT_CALL_CONF",		/* 16 */
    "PICKUP_CALL",			/* 17 */
    "PICKUP_CALL_CONF",			/* 18 */
    "GROUP_PICKUP_CALL",		/* 19 */
    "GROUP_PICKUP_CALL_CONF",		/* 20 */
    "HOLD_CALL",			/* 21 */
    "HOLD_CALL_CONF",			/* 22 */
    "MAKE_CALL",			/* 23 */
    "MAKE_CALL_CONF",			/* 24 */
    "MAKE_PREDICTIVE_CALL",		/* 25 */
    "MAKE_PREDICTIVE_CALL_CONF",	/* 26 */
    "QUERY_MWI",			/* 27 */
    "QUERY_MWI_CONF",			/* 28 */
    "QUERY_DND",			/* 29 */
    "QUERY_DND_CONF",			/* 30 */
    "QUERY_FWD",			/* 31 */
    "QUERY_FWD_CONF",			/* 32 */
    "QUERY_AGENT_STATE",		/* 33 */
    "QUERY_AGENT_STATE_CONF",		/* 34 */
    "QUERY_LAST_NUMBER",		/* 35 */
    "QUERY_LAST_NUMBER_CONF",		/* 36 */
    "QUERY_DEVICE_INFO",		/* 37 */
    "QUERY_DEVICE_INFO_CONF",		/* 38 */
    "RECONNECT_CALL",			/* 39 */
    "RECONNECT_CALL_CONF",		/* 40 */
    "RETRIEVE_CALL",			/* 41 */
    "RETRIEVE_CALL_CONF",		/* 42 */
    "SET_MWI",				/* 43 */
    "SET_MWI_CONF",			/* 44 */
    "SET_DND",				/* 45 */
    "SET_DND_CONF",			/* 46 */
    "SET_FWD",				/* 47 */
    "SET_FWD_CONF",			/* 48 */
    "SET_AGENT_STATE",			/* 49 */
    "SET_AGENT_STATE_CONF",		/* 50 */
    "TRANSFER_CALL",			/* 51 */
    "TRANSFER_CALL_CONF",		/* 52 */
    "UNIVERSAL_FAILURE_CONF",		/* 53 */
    "CALL_CLEARED",			/* 54 */
    "CONFERENCED",			/* 55 */
    "CONNECTION_CLEARED",		/* 56 */
    "DELIVERED",			/* 57 */
    "DIVERTED",				/* 58 */
    "ESTABLISHED",			/* 59 */
    "FAILED",				/* 60 */
    "HELD",				/* 61 */
    "NETWORK_REACHED",			/* 62 */
    "ORIGINATED",			/* 63 */
    "QUEUED",				/* 64 */
    "RETRIEVED",			/* 65 */
    "SERVICE_INITIATED",		/* 66 */
    "TRANSFERRED",			/* 67 */
    "CALL_INFORMATION",			/* 68 */
    "DO_NOT_DISTURB",			/* 69 */
    "FORWARDING",			/* 70 */
    "MESSAGE_WAITING",			/* 71 */
    "LOGGED_ON",			/* 72 */
    "LOGGED_OFF",			/* 73 */
    "NOT_READY",			/* 74 */
    "READY",				/* 75 */
    "WORK_NOT_READY",			/* 76 */
    "WORK_READY",			/* 77 */
    "ROUTE_REGISTER_REQ",		/* 78 */
    "ROUTE_REGISTER_REQ_CONF",		/* 79 */
    "ROUTE_REGISTER_CANCEL",		/* 80 */
    "ROUTE_REGISTER_CANCEL_CONF",	/* 81 */
    "ROUTE_REGISTER_ABORT",		/* 82 */
    "ROUTE_REQUEST",			/* 83 */
    "ROUTE_SELECT_REQUEST",		/* 84 */
    "RE_ROUTE_REQUEST",			/* 85 */
    "ROUTE_USED",			/* 86 */
    "ROUTE_END",			/* 87 */
    "ROUTE_END_REQUEST",		/* 88 */
    "ESCAPE_SVC",			/* 89 */
    "ESCAPE_SVC_CONF",			/* 90 */
    "ESCAPE_SVC_REQ",			/* 91 */
    "ESCAPE_SVC_REQ_CONF",		/* 92 */
    "PRIVATE",				/* 93 */
    "PRIVATE_STATUS",			/* 94 */
    "SEND_PRIVATE",			/* 95 */
    "BACK_IN_SERVICE",			/* 96 */
    "OUT_OF_SERVICE",			/* 97 */
    "REQ_SYS_STAT",			/* 98 */
    "SYS_STAT_REQ_CONF",		/* 99 */
    "SYS_STAT_START",			/* 100 */
    "SYS_STAT_START_CONF",		/* 101 */
    "SYS_STAT_STOP",			/* 102 */
    "SYS_STAT_STOP_CONF",		/* 103 */
    "CHANGE_SYS_STAT_FILTER",		/* 104 */
    "CHANGE_SYS_STAT_FILTER_CONF",	/* 105 */
    "SYS_STAT",				/* 106 */
    "SYS_STAT_ENDED",			/* 107 */
    "SYS_STAT_REQ",			/* 108 */
    "REQ_SYS_STAT_CONF",		/* 109 */
    "SYS_STAT_EVENT_SEND",		/* 110 */
    "MONITOR_DEVICE",			/* 111 */
    "MONITOR_CALL",			/* 112 */
    "MONITOR_CALLS_VIA_DEVICE",		/* 113 */
    "MONITOR_CONF",			/* 114 */
    "CHANGE_MONITOR_FILTER",		/* 115 */
    "CHANGE_MONITOR_FILTER_CONF",	/* 116 */
    "MONITOR_STOP",			/* 117 */
    "MONITOR_STOP_CONF",		/* 118 */
    "MONITOR_ENDED",			/* 119 */
    "SNAPSHOT_CALL",			/* 120 */
    "SNAPSHOT_CALL_CONF",		/* 121 */
    "SNAPSHOT_DEVICE",			/* 122 */
    "SNAPSHOT_DEVICE_CONF",		/* 123 */
    "GETAPI_CAPS",			/* 124 */
    "GETAPI_CAPS_CONF",			/* 125 */
    "GET_DEVICE_LIST",			/* 126 */
    "GET_DEVICE_LIST_CONF",		/* 127 */
    "QUERY_CALL_MONITOR",		/* 128 */
    "QUERY_CALL_MONITOR_CONF",		/* 129 */
};

// The real CSTA name starts with TSERVER_ then these strings.
LOCALVAR char* ACSFailureNames[] = {
    "STREAM_FAILED",			/* 0 */
    "NO_THREAD",			/* 1 */
    "BAD_DRIVER_ID",
    "DEAD_DRIVER",
    "MESSAGE_HIGH_WATER_MARK",
    "FREE_BUFFER_FAILED",
    "SEND_TO_DRIVER",
    "RECEIVE_FROM_DRIVER",
    "REGISTRATION_FAILED",
    "SPX_FAILED",
    "TRACE",
    "NO_MEMORY",
    "ENCODE_FAILED",
    "DECODE_FAILED",
    "BAD_CONNECTION",
    "BAD_PDU",
    "NO_VERSION",
    "ECB_MAX_EXCEEDED",
    "NO_ECBS",
    "NO_SDB",
    "NO_SDB_CHECK_NEEDED",
    "SDB_CHECK_NEEDED",
    "BAD_SDB_LEVEL",
    "BAD_SERVERID",
    "BAD_STREAM_TYPE",
    "BAD_PASSWORD_OR_LOGIN",
    "NO_USER_RECORD",
    "NO_DEVICE_RECORD",
    "DEVICE_NOT_ON_LIST",
    "USERS_RESTRICTED_HOME",
    "NO_AWAYPERMISSION",
    "NO_HOMEPERMISSION",
    "NO_AWAY_WORKTOP",
    "BAD_DEVICE_RECORD",
    "DEVICE_NOT_SUPPORTED",
    "INSUFFICIENT_PERMISSION",
    "NO_RESOURCE_TAG",
    "INVALID_MESSAGE",
    "EXCEPTION_LIST",
    "NOT_ON_OAM_LIST",
    "PBX_ID_NOT_IN_SDB",
    "USER_LICENSES_EXCEEDED",
    "OAM_DROP_CONNECTION",
    "NO_VERSION_RECORD",
    "OLD_VERSION_RECORD",
    "BAD_PACKET",
    "OPEN_FAILED",
    "OAM_IN_USE",
    "DEVICE_NOT_ON_HOME_LIST",
    "DEVICE_NOT_ON_CALL_CONTROL_LIST",
    "DEVICE_NOT_ON_AWAY_LIST",
    "DEVICE_NOT_ON_ROUTE_LIST",
    "DEVICE_NOT_ON_MONITOR_DEVICE_LIST",
    "DEVICE_NOT_ON_MONITOR_CALL_DEVICE_LIST",
    "NO_CALL_CALL_MONITOR_PERMISSION",
    "HOME_DEVICE_LIST_EMPTY",
    "CALL_CONTROL_LIST_EMPTY",
    "AWAY_LIST_EMPTY",
    "ROUTE_LIST_EMPTY",
    "MONITOR_DEVICE_LIST_EMPTY",
    "MONITOR_CALL_DEVICE_LIST_EMPTY",
    "USER_AT_HOME_WORKTOP",
    "DEVICE_LIST_EMPTY",
    "BAD_GET_DEVICE_LEVEL",
    "DRIVER_UNREGISTERED",
    "NO_ACS_STREAM",
    "DROP_OAM",
    "ECB_TIMEOUT",
    "BAD_ECB",
    "ADVERTISE_FAILED",
    "NETWARE_FAILURE",
    "TDI_QUEUE_FAULT",
    "DRIVER_CONGESTION",
    "NO_TDI_BUFFERS",
    "OLD_INVOKEID",
    "HWMARK_TO_LARGE",
    "SET_ECB_TO_LOW",
    "NO_RECORD_IN_FILE",
    "ECB_OVERDUE",
};

// The real CSTA names are DRIVER_ followed by one of these names.
LOCALVAR char* ACSdriverFailureNames[] = {
    "DUPLICATE_ACSHANDLE",
    "INVALID_ACS_REQUEST",
    "ACS_HANDLE_REJECTION",
    "INVALID_CLASS_REJECTION",
    "GENERIC_REJECTION",
    "RESOURCE_LIMITATION",
    "ACSHANDLE_TERMINATION",
    "LINK_UNAVAILABLE",
    "OAM_IN_USE",
};

LOCALVAR char* ClassNames[] =
			   {
    "ACSREQUEST",
    "ACSUNSOLICITED",
    "ACSCONFIRMATION",
    "CSTAREQUEST",
    "CSTAUNSOLICITED",
    "CSTACONFIRMATION",
    "CSTAEVENTREPORT",
    "?Unknown?"
			   };

LOCALVAR char * agentStateNames[] = {
    "AG_NOT_READY",
    "AG_NULL",
    "AG_READY",
    "AG_WORK_NOT_READY",
    "AG_WORK_READY"
};

/* Form a number enclosed in two question marks
**
** Inputs:
**	num	number
** Returns:
**	string valid untill next call to this routine
*/
LOCALDEF char *errorNumber(int num)
{
    static char buffer[20];

    sprintf(buffer,"?%d?",num);
    return buffer;
}

/* Get the ASCII form of a class name.
**
** Inputs:
**	evtClass	class id to report
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALDEF char *getClassName(EventClass_t evtClass)
{
    if (evtClass >= DIM(ClassNames)) {
	return errorNumber(evtClass);
    };
    return ClassNames[evtClass];
}

/* Get the ASCII form of a type name
**
** Inputs:
**	evtClass    class this is a type of
**	evtType	    type to report
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALDEF char *getTypeName(EventClass_t evtClass,EventType_t evtType)
{
    static char buffer[64];
    switch(evtClass) {
	case ACSREQUEST:
	case ACSUNSOLICITED:
	case ACSCONFIRMATION:
	    if ((evtType > 0) && (evtType <= DIM(ACSTypeNames))) {
		sprintf(buffer,"ACS_%s",ACSTypeNames[evtType-1]);
		return buffer;
	    };
	    break;
	case CSTAREQUEST:
	case CSTAUNSOLICITED:
	case CSTACONFIRMATION:
	case CSTAEVENTREPORT:
	    if ((evtType > 0) && (evtType <= DIM(CSTATypeNames))) {
		sprintf(buffer,"CSTA_%s",CSTATypeNames[evtType-1]);
		return CSTATypeNames[evtType-1];
	    };
	    break;
	default:
	    break;
    };
    return errorNumber(evtType);
}

/* Get the ASCII form of an acs failure error code
**
** Inputs:
**	boardType
** Returns:
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALVAR char *getBoardTypeName(PBXboardType_t boardType)
{
    switch ((int)boardType) {
	default:
	case npPvtBoardTypeUnknown:
	    return "Unknown";
	case npPvtBoardType618:
	    return "PBX618";
	case npPvtBoardType1010:
	    return "PBX1010";
	case npPvtBoardTypeT124:
	    return "PBXT124";
	case npPvtBoardType024:
	    return "PBX024";
    };
}

/* Get the ASCII form of a line style
**
** Inputs:
**	lineStyle
** Returns:
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALVAR char *getLineStyleName(lineStyle_t lineStyle)
{
    switch ((int)lineStyle) {
	default:
	    return "Unknown";
	case npPvtAnalogExt:
	    return "Extension";
	case npPvtAnalogTrunk:
	    return "AnalogTrunk";
//	case npPvtDigitalExtension:
//	    return "DigitalExtension";
	case npPvtDigitalTrunk:
	    return "DigitalTrunk";
    };
}

/* Get the ASCII form of an acs failure error code
**
** Inputs:
**	error	    error code
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALDEF char * getACSfailureName(ACSUniversalFailure_t error)
{
    static char buffer[64];
    if (((int)error >= 0) && (error < DIM(ACSFailureNames))) {
	sprintf(buffer,"TSERVER_%s",ACSFailureNames[error]);
	return buffer;
    };
    if (((((int)error)-1000) >= 0) &&
		((((int)error)-1000) < DIM(ACSdriverFailureNames))) {
	sprintf(buffer,"DRIVER_%s",ACSdriverFailureNames[error-1000]);
	return buffer;
    };
    return errorNumber(error);
}

/* Get the ASCII form of a CSTAEventCause_t
**
** Inputs:
**	cause	    cause to report
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALDEF char *getCauseName(CSTAEventCause_t cause)
{
    switch (cause) {
	case 0:
	case EC_NONE:
	    return NULL;
	case EC_ACTIVE_MONITOR:
	    return "Active_Monitor";
	case EC_ALTERNATE:
	    return "Alternate";
	case EC_BUSY:
	    return "Busy";
	case EC_CALL_BACK:
	    return "Call_Back";
	case EC_CALL_CANCELLED:
	    return "Call_Cancelled";
	case EC_CALL_FORWARD_ALWAYS:
	    return "Call_Forward_Always";
	case EC_CALL_FORWARD_BUSY:
	    return "Call_Forward_Busy";
	case EC_CALL_FORWARD_NO_ANSWER:
	    return "Call_Forward_No_Answer";
	case EC_CALL_FORWARD:
	    return "Call_Forward";
	case EC_CALL_NOT_ANSWERED:
	    return "Call_Not_Answered";
	case EC_CALL_PICKUP:
	    return "Call_Pickup";
	case EC_CAMP_ON:
	    return "Camp_On";
	case EC_DEST_NOT_OBTAINABLE:
	    return "Dest_Not_Obtainable";
	case EC_DO_NOT_DISTURB:
	    return "Do_Not_Disturb";
	case EC_INCOMPATIBLE_DESTINATION:
	    return "Incompatible_Destination";
	case EC_INVALID_ACCOUNT_CODE:
	    return "Invalid_Account_Code";
	case EC_KEY_CONFERENCE:
	    return "Key_Conference";
	case EC_LOCKOUT:
	    return "Lockout";
	case EC_MAINTENANCE:
	    return "Maintenance";
	case EC_NETWORK_CONGESTION:
	    return "Network_Congestion";
	case EC_NETWORK_NOT_OBTAINABLE:
	    return "Network_Not_Obtainable";
	case EC_NEW_CALL:
	    return "New_Call";
	case EC_NO_AVAILABLE_AGENTS:
	    return "No_Available_Agents";
	case EC_OVERRIDE:
	    return "Override";
	case EC_PARK:
	    return "Park";
	case EC_OVERFLOW:
	    return "Overflow";
	case EC_RECALL:
	    return "Recall";
	case EC_REDIRECTED:
	    return "Redirected";
	case EC_REORDER_TONE:
	    return "Reorder_Tone";
	case EC_RESOURCES_NOT_AVAILABLE:
	    return "Resources_Not_Available";
	case EC_SILENT_MONITOR:
	    return "Silent_Monitor";
	case EC_TRANSFER:
	    return "Transfer";
	case EC_TRUNKS_BUSY:
	    return "Trunks_Busy";
	case EC_VOICE_UNIT_INITIATOR:
	    return "Voice_Unit_Initiator";
	default:
	    return errorNumber(cause);
    };
}

/* Get a single character ASCII form of a LocalConnectionState_t
**
** Inputs:
**	state	connection state to report
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALDEF char *getLocalStateCharacter(LocalConnectionState_t state)
{
    switch (state) {
	case CS_NONE:
	    return NULL;
	case CS_NULL:
	    return NULL;
	case CS_INITIATE:
	    return "I";
	case CS_ALERTING:
	    return "A";
	case CS_CONNECT:
	    return "C";
	case CS_HOLD:
	    return "H";
	case CS_QUEUED:
	    return "Q";
	case CS_FAIL:
	    return "F";
	default:
	    return errorNumber(state);
    };
}

/* Get the ASCII form of a LocalConnectionState_t
**
** Inputs:
**	state	connection state to report
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALDEF char *getLocalStateName(LocalConnectionState_t state)
{
    switch (state) {
	case CS_NONE:
	    return NULL;
	case CS_NULL:
	    return NULL;
	case CS_INITIATE:
	    return "Initiated";
	case CS_ALERTING:
	    return "Alerting";
	case CS_CONNECT:
	    return "Connected";
	case CS_HOLD:
	    return "Held";
	case CS_QUEUED:
	    return "Queued";
	case CS_FAIL:
	    return "Failed";
	default:
	    return errorNumber(state);
    };
}

/* Get the ASCII form of an agent state name.
**
** Inputs:
**	state	    agent state to report
** Returns:
**	ASCII form of the name valid untill next call to this routine.
*/
LOCALDEF char *getAgentStateName(AgentState_t state)
{
    if (((int)state < 0) || (state >= DIM(agentStateNames))) {
	return errorNumber(state);
    };
    return agentStateNames[state];
}

LOCALVAR char * deviceTypeNames[] = {
    "STATION",
    "LINE",
    "BUTTON",
    "ACD",
    "TRUNK",
    "OPERATOR",
    "STATION_GROUP",
    "LINE_GROUP",
    "BUTTON_GROUP",
    "ACD_GROUP",
    "TRUNK_GROUP",
    "OPERATOR_GROUP",
    "OTHER"
};

/* Get the ASCII form of a device Type Name
**
** Inputs:
**	devType
** Returns:
**	ASCII form of the device type
*/
LOCALDEF char *getDeviceTypeName(DeviceType_t devType)
{
    if (((int)devType < 0) || (devType >= DIM(agentStateNames))) {
	return errorNumber(devType);
    };
    return deviceTypeNames[devType];
}


/* report the name of a soundID
**
** Inputs:
**	soundID
** Returns:
**	pointer to name of the soundID
*/
LOCALDEF char *getSoundIDname(SOUND_ID soundID)
{
    switch (soundID) {
	case npPvtAudioQuiet:
	    return "Quiet";
	case npPvtAudioStereo:
	    return "Stereo Jack";
	case npPvtContinuousDialToneSound:
	    return "Dial Tone";
	case npPvtContinuousRingToneSound:
	    return "Ring Tone";
	case npPvtContinuousAlertToneSound:
	    return "Alert Tone";
	case npPvtContinuousBusyToneSound:
	    return "Busy Tone";
	default:
	    return errorNumber(soundID);
    };
}


LOCALVAR char * CSTAfailureNames[] = {
    "GENERIC_UNSPECIFIED",
    "GENERIC_OPERATION",
    "REQUEST_INCOMPATIBLE_WITH_OBJECT",
    "VALUE_OUT_OF_RANGE",
    "OBJECT_NOT_KNOWN",
    "INVALID_CALLING_DEVICE",
    "INVALID_CALLED_DEVICE",
    "INVALID_FORWARDING_DESTINATION",
    "PRIVILEGE_VIOLATION_ON_SPECIFIED_DEVICE",
    "PRIVILEGE_VIOLATION_ON_CALLED_DEVICE",
    "PRIVILEGE_VIOLATION_ON_CALLING_DEVICE",
    "INVALID_CSTA_CALL_IDENTIFIER",
    "INVALID_CSTA_DEVICE_IDENTIFIER",
    "INVALID_CSTA_CONNECTION_IDENTIFIER",
    "INVALID_DESTINATION",
    "INVALID_FEATURE",
    "INVALID_ALLOCATION_STATE",
    "INVALID_CROSS_REF_ID",
    "INVALID_OBJECT_TYPE",
    "SECURITY_VIOLATION",
    "?NO_FAILURE_OF_TYPE_20?",
    "GENERIC_STATE_INCOMPATIBILITY",
    "INVALID_OBJECT_STATE",
    "INVALID_CONNECTION_ID_FOR_ACTIVE_CALL",
    "NO_ACTIVE_CALL",
    "NO_HELD_CALL",
    "NO_CALL_TO_CLEAR",
    "NO_CONNECTION_TO_CLEAR",
    "NO_CALL_TO_ANSWER",
    "NO_CALL_TO_COMPLETE",
    "?NO_FAILURE_OF_TYPE_30?",
    "GENERIC_SYSTEM_RESOURCE_AVAILABILITY",
    "SERVICE_BUSY",
    "RESOURCE_BUSY",
    "RESOURCE_OUT_OF_SERVICE",
    "NETWORK_BUSY",
    "NETWORK_OUT_OF_SERVICE",
    "OVERALL_MONITOR_LIMIT_EXCEEDED",
    "CONFERENCE_MEMBER_LIMIT_EXCEEDED",
    "?NO_FAILURE_OF_TYPE_39?",
    "?NO_FAILURE_OF_TYPE_40?",
    "GENERIC_SUBSCRIBED_RESOURCE_AVAILABILITY",
    "OBJECT_MONITOR_LIMIT_EXCEEDED",
    "EXTERNAL_TRUNK_LIMIT_EXCEEDED",
    "OUTSTANDING_REQUEST_LIMIT_EXCEEDED",
    "?NO_FAILURE_OF_TYPE_45?",
    "?NO_FAILURE_OF_TYPE_46?",
    "?NO_FAILURE_OF_TYPE_47?",
    "?NO_FAILURE_OF_TYPE_48?",
    "?NO_FAILURE_OF_TYPE_49?",
    "?NO_FAILURE_OF_TYPE_50?",
    "GENERIC_PERFORMANCE_MANAGEMENT",
    "PERFORMANCE_LIMIT_EXCEEDED",
    "?NO_FAILURE_OF_TYPE_53?",
    "?NO_FAILURE_OF_TYPE_54?",
    "?NO_FAILURE_OF_TYPE_55?",
    "?NO_FAILURE_OF_TYPE_56?",
    "?NO_FAILURE_OF_TYPE_57?",
    "?NO_FAILURE_OF_TYPE_58?",
    "?NO_FAILURE_OF_TYPE_59?",
    "?NO_FAILURE_OF_TYPE_60?",
    "SEQUENCE_NUMBER_VIOLATED",
    "TIME_STAMP_VIOLATED",
    "PAC_VIOLATED",
    "SEAL_VIOLATED",
    "?NO_FAILURE_OF_TYPE_65?",
    "?NO_FAILURE_OF_TYPE_66?",
    "?NO_FAILURE_OF_TYPE_67?",
    "?NO_FAILURE_OF_TYPE_68?",
    "?NO_FAILURE_OF_TYPE_69?",
    "GENERIC_UNSPECIFIED_REJECTION",
    "GENERIC_OPERATION_REJECTION",
    "DUPLICATE_INVOCATION_REJECTION",
    "UNRECOGNIZED_OPERATION_REJECTION",
    "MISTYPED_ARGUMENT_REJECTION",
    "RESOURCE_LIMITATION_REJECTION",
    "ACS_HANDLE_TERMINATION_REJECTION",
    "SERVICE_TERMINATION_REJECTION",
    "REQUEST_TIMEOUT_REJECTION",
    "REQUESTS_ON_DEVICE_EXCEEDED_REJECTION",
};

/* Get the ASCII form of a CSTA universal failure
**
** Inputs:
**	error
** Returns:
**	ASCII form of the failure name
*/
LOCALDEF char *getCSTAfailureName(CSTAUniversalFailure_t error)
{
    if (((int)error < 0) || (error >= DIM(CSTAfailureNames))) {
	return errorNumber(error);
    };
    return CSTAfailureNames[error];
}

/* Report the list of device class names
**
** Inputs:
**	devClass
** Returns:
**	ASCII string of device classes valid untill next call to this routine
*/
LOCALDEF char *getDeviceClassList(DeviceClass_t devClass)
{
    static char classList[20];
    BOOLEAN anyFound = FALSE;

    classList[0] = NUL;
    if ((devClass & DC_VOICE) != 0) {
	anyFound = TRUE;
	strcat(classList,"Voice");
    };
    if ((devClass & DC_DATA) != 0) {
	if (anyFound) strcat(classList,",");
	anyFound = TRUE;
	strcat(classList,"Data");
    };
    if ((devClass & DC_IMAGE) != 0) {
	if (anyFound) strcat(classList,",");
	anyFound = TRUE;
	strcat(classList,"Image");
    };
    if ((devClass & DC_OTHER) != 0) {
	if (anyFound) strcat(classList,",");
	strcat(classList,"Other");
    };
    return classList;
}

#if 0
/* Get the ASCII form of a deviceIDstatus
**
** Inputs:
**	devStatus
** Returns:
**	ASCII form of the status
*/
LOCALDEF char *getDevIDstatusName(DeviceIDStatus_t devStatus)
{
    switch (devStatus) {
	case ID_PROVIDED:
	    return "ID_PROVIDED";
	case ID_NOT_KNOWN:
	    return "ID_NOT_KNOWN";
	case ID_NOT_REQUIRED:
	    return "ID_NOT_REQUIRED";
	default:
	    return errorNumber(devStatus);
    };
}
#endif

/* Get the ASCII form of a deviceIDtype
**
** Inputs:
**	devType
** Returns:
**	ASCII form of the type
*/
LOCALDEF char *getDevIDtypeName(DeviceIDType_t devType)
{
    switch (devType) {
	case DEVICE_IDENTIFIER:
	    return "DeviceID";
	case IMPLICIT_PUBLIC:
	    return "ImplicitPublic";
	case EXPLICIT_PUBLIC_UNKNOWN:
	    return "PublicUnknown";
	case EXPLICIT_PUBLIC_INTERNATIONAL:
	    return "PublicInternational";
	case EXPLICIT_PUBLIC_NATIONAL:
	    return "PublicNational";
	case EXPLICIT_PUBLIC_NETWORK_SPECIFIC:
	    return "PublicNetworkSpecific";
	case EXPLICIT_PUBLIC_SUBSCRIBER:
	    return "PublicSubscriber";
	case EXPLICIT_PUBLIC_ABBREVIATED:
	    return "PublicAbbreviated";
	case IMPLICIT_PRIVATE:
	    return "ImplicitPrivate";
	case EXPLICIT_PRIVATE_UNKNOWN:
	    return "PrivateUnknown";
	case EXPLICIT_PRIVATE_LEVEL3_REGIONAL_NUMBER:
	    return "Level3RegionalNumber";
	case EXPLICIT_PRIVATE_LEVEL2_REGIONAL_NUMBER:
	    return "Level2RegionalNumber";
	case EXPLICIT_PRIVATE_LEVEL1_REGIONAL_NUMBER:
	    return "Level1RegionalNumber";
	case EXPLICIT_PRIVATE_PTN_SPECIFIC_NUMBER:
	    return "PtnSpecificNumber";
	case EXPLICIT_PRIVATE_LOCAL_NUMBER:
	    return "LocalNumber";
	case EXPLICIT_PRIVATE_ABBREVIATED:
	    return "PrivateAbbreviated";
	case OTHER_PLAN:
	    return "OtherPlan";
	case TRUNK_IDENTIFIER:
	    return "TrunkID";
	case TRUNK_GROUP_IDENTIFIER:
	    return "TrunkGroupID";
	default:
	    return errorNumber(devType);
    };
}

/* Emit first line information common to all messages
**
** Inputs:
**	comnt		comment string
**	verbose		TRUE if verbose form is being used
*/
LOCALDEF void o_firstLineStart(char *comnt,BOOLEAN verbose)
{
    if (!verbose) return;
    o_lstring(comnt,0,FALSE);
    if (timestampsOn) o_lstring(timestamp(),0,TRUE);
    o_lstring("",0,TRUE);		/* get a space if needed */
    return;
}

/* Output an ACS unsolicited event
**
** Inputs:
**	eType	type of the event
**	unsol	event to report
*/
LOCALDEF void o_ACSunsolicited(EventType_t eType,
			       ACSUnsolicitedEvent FAR *unsol)
{
    if (unsol == NULL) return;
    switch (eType) {
	case ACS_UNIVERSAL_FAILURE:
	    o_lstring(getACSfailureName(unsol->u.failureEvent.error),0,FALSE);
	    break;
	default:
	    o_lprintf(0,FALSE,
		      "? Type %s is not an expected ACSUnsolicitedEvent",
		      getTypeName(ACSUNSOLICITED,eType));
    };
    return;
}


/* Output an InvokeID, eventually this might report some logged
** textual information about the invokeID
**
** Inputs:
**	invokeID
**	column		column for first character
**	flag		TRUE if whitespace must preceed character
*/
LOCALDEF void o_invokeID(InvokeID_t invokeID,int column,BOOLEAN flag)
{
    if (invokeIDstampsOn) o_lnumber(invokeID,column,flag,10);
    return;
}

/* Output an InvokeID with delta time
**
** Inputs:
**	invokeID
**	column		column for first character
**	flag		TRUE if whitespace must preceed character
*/
LOCALDEF void o_invokeIDwithTime(InvokeID_t invokeID,int column,BOOLEAN flag)
{
    o_invokeID(invokeID,column,flag);
    o_invokeIDdelta(invokeID,column+8,TRUE);
    return;
}

/* Output an ACS confirmation event
**
** Inputs:
**	eType	event type
**	event
**	comnt	comment string with which to prefix each line
**	verbose	TRUE for multiline form of report
*/
LOCALDEF void o_ACSconfirmation(EventType_t eType,
				ACSConfirmationEvent *event,char *comnt,
				BOOLEAN verbose)
{
    const int indent = 8;
    if (event == NULL) return;
    o_firstLineStart(comnt,verbose);
    o_lstring(getTypeName(ACSCONFIRMATION,eType),0,FALSE);
    o_invokeIDwithTime(event->invokeID,timeColumn,TRUE);
    switch (eType) {
	case ACS_OPEN_STREAM_CONF:
	    if (!verbose) break;
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Api:",indent,TRUE);
	    o_lstring(event->u.acsopen.apiVer,indent+4,TRUE);
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Lib:",indent,TRUE);
	    o_lstring(event->u.acsopen.libVer,indent+4,TRUE);
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Tsrv:",indent,TRUE);
	    o_lstring(event->u.acsopen.tsrvVer,indent+4,TRUE);
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("dev:",indent,TRUE);
	    o_lstring(event->u.acsopen.drvrVer,indent+4,TRUE);
	    sscanf(event->u.acsopen.drvrVer," %d.%d-%d",
		   &firmwareMajor,&firmwareMinor,&firmwareEdit);
	    break;
	case ACS_CLOSE_STREAM_CONF:
	    break;
	case ACS_UNIVERSAL_FAILURE_CONF:
	    o_lstring(getACSfailureName(event->u.failureEvent.error),
		      30,TRUE);
	    break;
	default:
	    o_lprintf(indent,TRUE,"unknown type %s",errorNumber(eType));
	    break;
    };
    if (verbose) o_newline(0);
    return;
}

/* Output a CSTA report event
**
** Inputs:
**	eType	event type
**	event
**	comnt	comment string with which to prefix each line
**	verbose TRUE for multiline form
*/
LOCALDEF void o_CSTAreport(EventType_t eType,CSTAEventReport *event,
				char *comnt,BOOLEAN verbose)
{
    if (event == NULL) return;
    o_firstLineStart(comnt,verbose);
    switch (eType) {
	default:
	    o_lprintf(0,FALSE,"CSTAreport ?%d?",eType);
	    break;
    };
    if (verbose) o_newline(0);
    return;
}

/* output a deviceID
**
** Inputs:
**	deviceID	device to report
**	column		column for first character
**	flag		TRUE if whitespace must preceed string
**	
*/
EXTERNDEF void o_deviceID(DeviceID_t FAR *deviceID,int column,BOOLEAN flag)
{
    char string[sizeof(DeviceID_t)+1];

    if (deviceID == NULL) return;
    strncpy(string,(char *)deviceID,sizeof(DeviceID_t));
    string[sizeof(DeviceID_t)] = NUL;
    o_lstring(string,column,flag);
    return;
}
    
/* output a connection ID
**
** Inputs:
**	connID		connection to report
**	column		column for first character
**	flag		TRUE if whitespace must preceed string
*/
EXTERNDEF void o_connectionID(ConnectionID_t FAR *connID,int column,BOOLEAN flag)
{
    if (connID == NULL) return;
    o_deviceID(&connID->deviceID,column,flag);
    if (connID->callID != 0) {
	o_lstring("..",0,FALSE);
	o_lnumber(connID->callID,0,FALSE,16);
    };
    if (connID->devIDType == STATIC_ID) return;
    o_lstring(":D",0,FALSE);
    return;
}
    
/* output an extendedDeviceID
**
** Inputs:
**	deviceID	device to report
**	column		column for first character
**	flag		TRUE if whitespace must preceed string
** Returns:
**	TRUE if the deviceID is present and was output
*/
EXTERNDEF BOOLEAN o_extendedDeviceID(ExtendedDeviceID_t FAR *deviceID,
				    int column,BOOLEAN flag)
{
    if (deviceID == NULL) return FALSE;
    if (deviceID->deviceIDStatus != ID_PROVIDED) return FALSE;
    o_deviceID(&deviceID->deviceID,column,flag);
    if (deviceID->deviceIDType == DEVICE_IDENTIFIER) return TRUE;
    o_lstring(getDevIDtypeName(deviceID->deviceIDType),0,TRUE);
    return TRUE;
}
    
/* output a connection list
**
** Inputs:
**	conn		connection list to report
**	column		column for first character
**	comnt		comment string to prefix each line
*/
LOCALDEF void o_connectionList(ConnectionList_t *conn,
			       int column,char *comnt)
{
    int i;
    if (conn == NULL) return;
    for (i = 0;i < conn->count;i += 1) {
	o_lstring(comnt,0,FALSE);
	o_connectionID(&conn->connection[i].party,column,FALSE);
	o_extendedDeviceID(&conn->connection[i].staticDevice,0,TRUE);
	o_newline(0);
    };
    return;
}

/* output private file info and time since last invoke ID. Reset
** invokeID time.
**
** Inputs:
**	file	pointer to file data
*/
LOCALDEF void o_privateFile(NetPhonePrivateFile_t *file,int column,BOOLEAN flag)
{
    if (file == NULL) return;
    o_lstring(file->fileName,column,flag);
    o_char('@',0,FALSE);
    o_lnumber(file->position,0,FALSE,10);
    o_invokeIDdelta(file->invokeID,0,TRUE);
    o_lstring("Id:",0,TRUE);
    o_invokeID(file->invokeID,0,TRUE);
    return;
}


/* Get the ASCII form of a ForwardingType_t
**
** Inputs:
**	fwdType
** Returns:
**	ASCII form of the type
*/
LOCALDEF char *getForwardingTypeName(ForwardingType_t fwdType)
{
    switch (fwdType) {
	case FWD_IMMEDIATE:
	    return "Immediate";		/* 0 */
	case FWD_BUSY:
	    return "Busy";		/* 1 */
	case FWD_NO_ANS:
	    return "NoAnswer";		/* 2 */
	case FWD_BUSY_INT:
	    return "BusyInternal";	/* 3 */
	case FWD_BUSY_EXT:
	    return "BusyExternal";	/* 4 */
	case FWD_NO_ANS_INT:
	    return "NoAnswerInternal";	/* 5 */
	case FWD_NO_ANS_EXT:
	    return "NoAnswerExternal";	/* 6 */
	default:
	    return errorNumber(fwdType);
    };
}


/* output a private message
**
** Inputs:
**	pInfo	    pointer to forwarding info
**	column	    column for first character
**	flag	    TRUE if whitespace must preceed string
*/
LOCALDEF void o_forwardingInfo(ForwardingInfo_t *pInfo,int column,
			       BOOLEAN flag)
{
    if (pInfo == NULL) return;
    o_deviceID(&pInfo->forwardDN,column,flag);
    o_lstring((pInfo->forwardingOn) ? "ON" : "OFF",column+12,TRUE);
    o_lstring(getForwardingTypeName(pInfo->forwardingType),column+16,TRUE);
    return;
}

/* get a private type name
**
** Inputs:
**	type
** Returns:
**	pointer to ASCII name of the type
*/
LOCALDEF char *getPrivateTypeName(privateMessageID_t type)
{
    switch (type) {
	case npPvtUnknown:		/* 0 */
	    return "npPvtUnknown";
	case npPvtStartPrivateStatus:	/* 1 */
	    return "npPvtStartPrivateStatus";
	case npPvtStopPrivateStatus:	/* 2 */
	    return "npPvtStopPrivateStatus";
	case npPvtPlay:			/* 10 */
	    return "npPvtPlay";
	case npPvtPlayStarted:		/* 11 */
	    return "npPvtPlayStarted";
	case npPvtPlayStop:		/* 12 */
	    return "npPvtPlayStop";
	case npPvtPlayEnded:		/* 13 */
	    return "npPvtPlayEnded";
	case npPvtPlayNextFile:		/* 14 */
	    return "npPvtPlayNextFile";
	case npPvtRecord:		/* 15 */
	    return "npPvtRecord";
	case npPvtRecordStarted:	/* 16 */
	    return "npPvtRecordStarted";
	case npPvtRecordStop:		/* 17 */
	    return "npPvtRecordStop";
	case npPvtRecordEnded:		/* 18 */
	    return "npPvtRecordEnded";
	case npPvtGetDigits:		/* 20 */
	    return "npPvtGetDigits";
	case npPvtStopGetDigits:	/* 21 */
	    return "npPvtStopGetDigits";
	case npPvtDigitsSeen:		/* 22 */
	    return "npPvtDigitsSeen";
	case npPvtPlayDigits:		/* 23 */
	    return "npPvtPlayDigits";
	case npPvtGetDialedDigits:	/* 24 */
	    return "npPvtGetDialedDigits";
	case npPvtReportDialedDigits:	/* 25 */
	    return "npPvtReportDialedDigits";
	case npPvtRequestSwrID:		/* 30 */
	    return "npPvtRequestSwrID";
	case npPvtReportSwrID:		/* 31 */
	    return "npPvtReportSwrID";
	case npPvtRequestHdwrID:	/* 32 */
	    return "npPvtRequestHdwrID";
	case npPvtReportHdwrID:		/* 33 */
	    return "npPvtReportHdwrID";
	case npPvtGetBoardsPresent:	/* 34 */
	    return "npPvtGetBoardsPresent";
	case npPvtReportBoardsPresent:	/* 35 */
	    return "npPvtReportBoardsPresent";
	case npPvtRequestDriverID:	/* 36 */
	    return "npPvtRequestDriverID";
	case npPvtReportDriverID:	/* 37 */
	    return "npPvtReportDriverID";
	case npPvtGetStreamInfo:	/* 38 */
	    return "npPvtGetStreamInfo";
	case npPvtReportStreamInfo:	/* 39 */
	    return "npPvtReportStreamInfo";
	case npPvtGetTranslation:	/* 50 */
	    return "npPvtGetTranslation";
	case npPvtSetTranslation:	/* 51 */
	    return "npPvtSetTranslation";
	case npPvtReportTranslation:	/* 52 */
	    return "npPvtReportTranslation";
	case npPvtSetLineUserInfo:	/* 69 */
	    return "npPvtSetLineUserInfo";
	case npPvtSetLineInfo1:		/* 53 */
	    return "npPvtSetLineInfo1";
	case npPvtGetLineInfo:		/* 54 */
	    return "npPvtGetLineInfo";
	case npPvtReportLineInfo:	/* 55 */
	    return "npPvtReportLineInfo";
	case npPvtSetCTIctrl:		/* 56 */
	    return "npPvtSetCTIctrl";
	case npPvtGetCTIctrl:		/* 57 */
	    return "npPvtGetCTIctrl";
	case npPvtReportCTIctrl:	/* 58 */
	    return "npPvtReportCTIctrl";
	case npPvtEnableAdministration:	/* 59 */
	    return "npPvtEnableAdministration";
	case npPvtSetAdminPassword:	/* 60 */
	    return "npPvtSetAdminPassword";
	case npPvtSetSystemInfo1:	/* 61 */
	    return "npPvtSetSystemInfo1";
	case npPvtGetSystemInfo:	/* 62 */
	    return "npPvtGetSystemInfo";
	case npPvtReportSystemInfo:	/* 63 */
	    return "npPvtReportSystemInfo";
	case npPvtRequestBroadcast:	/* 64 */
	    return "npPvtRequestBroadcast";
	case npPvtReportBroadcast:	/* 65 */
	    return "npPvtReportBroadcast";
	case npPvtGetBoardOrder:	/* 66 */
	    return "npPvtGetBoardOrder";
	case npPvtSetBoardOrder:	/* 67 */
	    return "npPvtSetBoardOrder";
	case npPvtReportBoardOrder:	/* 68 */
	    return "npPvtReportBoardOrder";
	case npPvtGetStatusInfo:	/* 70 */
	    return "npPvtGetStatusInfo";
	case npPvtReportStatusInfo:	/* 71 */
	    return "npPvtReportStatusInfo";
	case npPvtConnectSound:		/* 72 */
	    return "npPvtConnectSound";
	case npPvtDisconnectSound:	/* 73 */
	    return "npPvtDisconnectSound";
	case npPvtGetSound:		/* 74 */
	    return "npPvtGetSound";
	case npPvtReportSound:		/* 75 */
	    return "npPvtReportSound";
	case npPvtRequestInterlock:	/* 76 */
	    return "npPvtRequestInterlock";
	case npPvtReleaseInterlock:	/* 77 */
	    return "npPvtReleaseInterlock";
	case npPvtReportInterlock:	/* 78 */
	    return "npPvtReportInterlock";
	case npPvtMakeCTIcall:		/* 79 */
	    return "npPvtMakeCTIcall";
	case npPvtGetCallerIDinfo:	/* 80 */
	    return "npPvtGetCallerIDinfo";
	case npPvtReportCallerIDinfo:	/* 81 */
	    return "npPvtReportCallerIDinfo";
	case npPvtGetRawCallerID:	/* 82 */
	    return "npPvtGetRawCallerID";
	case npPvtReportRawCallerID:	/* 83 */
	    return "npPvtReportRawCallerID";
	case npPvtSetOperatorNumber:	/* 84 */
	    return "npPvtSetOperatorNumber";
	case npPvtGetOperatorNumber:	/* 85 */
	    return "npPvtGetOperatorNumber";
	case npPvtReportOperatorNumber:	/* 86 */
	    return "npPvtReportOperatorNumber";
	case npPvtGetROMboardOrder:	/* 87 */
	    return "npPvtReportBroadcast";
	case npPvtReportROMboardOrder:	/* 88 */
	    return "npPvtGetBoardOrder";
	case npPvtSetSystemInfo2:	/* 89 */
	    return "npPvtSetSystemInfo2";
	case npPvtGetSoftwareVersion:	/* 90 */
	    return "npPvtGetSoftwareVersion";
	case npPvtReportSoftwareVersion: /* 91 */
	    return "npPvtReportSoftwareVersion";
	case npPvtGetSystemResources:	/* 92 */
	    return "npPvtGetSystemResources";
	case npPvtGetBoardResources:	/* 93 */
	    return "npPvtGetBoardResources";
	case npPvtReportSystemResources: /* 94 */
	    return "npPvtReportSystemResources";
	case npPvtReportBoardResources:	/* 95 */
	    return "npPvtReportBoardResources";
	case npPvtSetSystemInfo:	/* 96 */
	    return "npPvtSetSystemInfo";
	case npPvtSetLineInfo:		/* 97 */
	    return "npPvtSetLineInfo";
	case npPvtSetDSpanInfo:		/* 98 */
	    return "npPvtSetDSpanInfo";
	case npPvtGetDSpanInfo:		/* 99 */
	    return "npPvtGetDSpanInfo";
	case npPvtReportDSpanInfo:	/* 100 */
	    return "npPvtReportDSpanInfo";
	case npPvtSetMoreLineInfo:	/* 101 */
	    return "npPvtSetMoreLineInfo";
	case npPvtGetMoreLineInfo:	/* 102 */
	    return "npPvtGetMoreLineInfo";
	case npPvtReportMoreLineInfo:	/* 103 */
	    return "npPvtReportMoreLineInfo";
	case npPvtSetDigitalTrunkFlags:	/* 104 */
	    return "npPvtSetDigitalTrunkFlags";
	case npPvtGetCallInfo:		/* 105 */
	    return "npPvtGetCallInfo";
	case npPvtReportCallInfo:	/* 106 */
	    return "npPvtReportCallInfo";
	case npPvtGetParkInfo:		/* 107 */
	    return "npPvtGetParkInfo";
	case npPvtParkCall:		/* 108 */
	    return "npPvtParkCall";
	case npPvtParkCallOccured:	/* 109 */
	    return "npPvtParkCallOccured";
	default:
	    return errorNumber(type);
    };
}

/* output a line information structure. Assumes that the leading comment
** and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	comnt	comment to prefix lines
**	indent
*/
EXTERNDEF void o_lineInfo(NetPhonePrivateLineInfo_t *pInfo,char *comnt,
			 int indent)
{
    if (pInfo == NULL) return;
    o_lstring(getLineStyleName(pInfo->lineStyle),indent,TRUE);
    o_lstring("Class",0,TRUE);
    o_lnumber(pInfo->classOfService,0,TRUE,10);
    o_lstring("Pickup group",0,TRUE);
    o_lnumber(pInfo->pickupGroup,0,TRUE,10);
    if (pInfo->doNotDisturb) {
	o_lstring("DND",0,TRUE);
    };
    if (pInfo->callWaitingEnabled) {
	o_lstring("Call waiting ON",0,TRUE);
    };
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("No answer time",indent,TRUE);
    o_fixed_number(pInfo->noAnswerTime,0,TRUE,10,1);
    o_lstring("Seconds",0,TRUE);
    o_lnumber(pInfo->waitNrings,0,TRUE,10);
    if (pInfo->ifBusy[0] != NUL) {
	o_newline(0);
	o_lstring(comnt,0,FALSE);
	o_lstring("If Busy:",indent,TRUE);
	o_lstring(pInfo->ifBusy,0,TRUE);
    };
    if (pInfo->ifNoAnswer[0] != NUL) {
	o_newline(0);
	o_lstring(comnt,0,FALSE);
	o_lstring("If no answer:",indent,TRUE);
	o_lstring(pInfo->ifNoAnswer,0,TRUE);
    };
    if (pInfo->ifImmediate[0] != NUL) {
	o_newline(0);
	o_lstring(comnt,0,FALSE);
	o_lstring("Immediate:",indent,TRUE);
	o_lstring(pInfo->ifImmediate,0,TRUE);
    };
    if ((firmwareMajor > 2)
	|| ((firmwareMajor == 2)
	    && ((firmwareMinor > 2) || (firmwareEdit > 230)) ) ) {
	if (pInfo->devicePhysName[0] != NUL) {
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Physical deviceID:",indent,TRUE);
	    o_lstring(pInfo->devicePhysName,0,TRUE);
	};
	if (pInfo->deviceXlateName[0] != NUL) {
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Xlation table deviceID:",indent,TRUE);
	    o_lstring(pInfo->deviceXlateName,0,TRUE);
	};
    };
    o_newline(0);
    return;
}

/* output a more line information structure. Assumes that the leading comment
** and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	comnt	comment to prefix lines
**	indent
*/
EXTERNDEF void o_moreLineInfo(NetPhonePrivateMoreLineInfo_t *pInfo,char *comnt,
			 int indent)
{
    if (comnt) {};
    o_lstring(getLineStyleName(pInfo->lineStyle),indent,TRUE);
    if (pInfo->devicePhysName[0] != NUL) {
	o_lstring("Phys:",indent,TRUE);
	o_lstring(pInfo->devicePhysName,0,TRUE);
    };
    o_lstring("DTrunk Flags: 0x",0,TRUE);
    o_lnumber(pInfo->digitalTrunkFlags,0,FALSE,16);
    o_newline(0);
    return;
}

/* output a call info structure
**
** Inputs:
**	pInfo	structure to output
**	indent
*/
EXTERNDEF void o_callInfo(NetPhonePrivateCallInfo_t *pInfo,int indent)
{
    o_lnumber(pInfo->numberLines,indent,TRUE,10);
    o_lstring("members",0,TRUE);
    if (pInfo->clientCTIctrl != 0) {
	o_lstring("CTI:",indent,TRUE);
	o_lnumber(pInfo->clientCTIctrl,0,TRUE,10);
    };
    if (pInfo->parkID != 0) {
	o_lstring("Park:",indent,TRUE);
	o_lnumber(pInfo->parkID,0,TRUE,10);
    };
    return;
}

/* get Span's mode name
**
** Inputs:
**	mode
** Returns:
**	pointer to ASCII name of the type
*/
EXTERNDEF char *getDSmodeName(nbPvtDSmode_t mode)
{
    switch (mode) {
	default:
	    return "Unknown";
	case nbPvtDSmodeT1:
	    return "T1";
	case nbPvtDSmodeE1:
	    return "E1";
    };
}

/* get Span's framing name
**
** Inputs:
**	mode
** Returns:
**	pointer to ASCII name of the type
*/
EXTERNDEF char *getDSframeName(nbPvtDSframe_t frame)
{
    switch (frame) {
	default:
	    return "Unknown";
	case nbPvtDSframeSF:
	    return "SF";
	case nbPvtDSframeESF:
	    return "ESF";
    };
}

/* get Span's signaling protocol name
**
** Inputs:
**	mode
** Returns:
**	pointer to ASCII name of the type
*/
LOCALDEF char *getDSlineSignalingName(nbPvtDSlineSignal_t signaling)
{
    switch (signaling) {
	default:
	     return "Unknown";
	case nbPvtDSsigEandM:
	    return "EandM";
	case nbPvtDSsigLoopStart:
	    return "LoopStart";
	case nbPvtDSsigLoopStartRLCF:
	    return "LoopStartRLCF";
	case nbPvtDSsigGndStart:
	    return "GndStart";
	case nbPvtDSsigGndStartRLCF:
	    return "GndStartRLCF";
	case nbPvtDSsigCustLcfRBat:
	    return "CustLcfRBat";
	case nbPvtDSsigNetLcfRBat:
	    return "NetLcfRBat";
	case nbPvtDSsigCIloopSuper:
	    return "CIloopSuper";
    };
}

/* get Span's line coding scheme name
**
** Inputs:
**	mode
** Returns:
**	pointer to ASCII name of the type
*/
LOCALDEF char *getDSlineCodingName(nbPvtDSlineCode_t code)
{
    switch (code) {
	default:
	    return "Unknown";
	case nbPvtDSlineCodeNRZ:
	    return "NRZ";
	case nbPvtDSlineCodeNotUsed:
	    return "NotUsed";
	case nbPvtDSlineCodeAMI:
	    return "AMI";
	case nbPvtDSlineCodeB8ZS:
	    return "B8ZS";
    };
}

/* get Span's line start name
**
** Inputs:
**	mode
** Returns:
**	pointer to ASCII name of the type
*/
LOCALDEF char *getDSlineStartName(nbPvtDSlineStart_t startType)
{
    switch (startType) {
	default:
	    return "Unknown";
	case nbPvtDSstartIMM:
	    return "Immediate";
	case nbPvtDSstartDEL:
	    return "Delay";
	case nbPvtDSstartVDEL:
	    return "DelayVerify";
	case nbPvtDSstartWINK:
	    return "WinkStart";
	case nbPvtDSstartLOOP:
	    return "LoopStart";
	case nbPvtDSstartGND:
	    return "GroundStart";
    };
}


/* get Span's span signaling mode
**
** Inputs:
**	mode
** Returns:
**	pointer to ASCII name of the type
*/
LOCALDEF char *getDSsignalingMode(nbPvtDSspanSignal_t mode)
{
    switch (mode) {
	default:
	    return "Unknown";
	case nbPvtDSspanSigCAS:
	    return "CAS(BitRobbed)";
	case nbPvtDSspanSigPRI:
	    return "PRI";
    };
}

/* get Span's span connection type
**
** Inputs:
**	mode
** Returns:
**	pointer to ASCII name of the type
*/
LOCALDEF char *getDSconnectName(nbPvtDSconnect_t connectType)
{
    switch (connectType) {
	default:
	    return "Unknown";
	case nbPvtDSconnectNet:
	    return "ConnectNet";
	case nbPvtDSconnectPBX:
	    return "ConnectPBX";
    };
}

/* output a more line information structure. Assumes that the leading comment
** and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	comnt	comment to prefix lines
**	indent
*/
EXTERNDEF void o_DSpanInfo(NetPhonePrivateDSpanInfo_t *pInfo,char *comnt,
			 int indent)
{
    o_lprintf(indent,TRUE,"%06ld%04ld",pInfo->serial.fabDate,
	      pInfo->serial.number);
    o_lstring("Index:",0,TRUE);
    o_lnumber(pInfo->bdIndex,0,TRUE,10);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring(getDSmodeName(pInfo->mode),indent,TRUE);
    o_lstring(getDSframeName(pInfo->frame),0,TRUE);
    o_lstring(getDSlineSignalingName(pInfo->lineSignal),0,TRUE);
    o_lstring(getDSlineCodingName(pInfo->lineCode),0,TRUE);
    o_lstring(getDSlineStartName(pInfo->lineStart),0,TRUE);
    o_lstring(getDSsignalingMode(pInfo->spanSignal),0,TRUE);
    o_lstring(getDSconnectName(pInfo->connectType),0,TRUE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("Expected digits:",indent,TRUE);
    o_lnumber(pInfo->expectedDIDdigits,0,TRUE,10);
    if (pInfo->userSpecName[0] != NUL) {
	o_lstring("User Name",0,TRUE);
	o_lstring(pInfo->userSpecName,0,TRUE);
    };
    o_newline(0);
    return;
}

/* Output the time string from a callerID
**	Typical string "10291452"
**
** Inputs:
**	pTime		string to output
**	indent
**	flag		TRUE if whitespace must preceed string
*/
LOCALDEF void o_callerIDtime(char *pTime,
			     int indent,BOOLEAN flag)
{
    if (pTime == NULL) return;
    if (strlen(pTime) < 8) {
	o_lstring(pTime,indent,flag);
	return;
    };
    o_char(pTime[0],indent,flag);
    o_char(pTime[1],0,FALSE);
    o_char('/',0,FALSE);
    o_char(pTime[2],0,FALSE);
    o_char(pTime[3],0,FALSE);
    o_char(pTime[4],0,TRUE);
    o_char(pTime[5],0,FALSE);
    o_char(':',0,FALSE);
    o_char(pTime[6],0,FALSE);
    o_char(pTime[7],0,FALSE);
    return;
}

/* output a callerID information structure
**
** Inputs:
**	pInfo	structure to output
**	indent
**	comnt	
*/
EXTERNDEF void o_callerIDinfo(NetPhonePrivateCallerIDinfo_t *pInfo,
			 int indent,char *comnt)
{
    if (pInfo == NULL) return;
    o_lstring(comnt,0,FALSE);
    o_lstring("Calling:",indent,FALSE);
    o_lstring(pInfo->callingNumber,0,TRUE);
    if (pInfo->isTrunk) o_lstring("Trunk",indent+12,TRUE);
    o_lstring((pInfo->idFound) ? "Found" : "Missing",indent+18,TRUE);
    if (pInfo->idWasBlocked) o_lstring("Blocked",0,TRUE);
    if (pInfo->idOutArea) o_lstring("Out of area",0,TRUE);
    o_newline(0);
    if ((firmwareMajor > 2)
	|| ((firmwareMajor == 2)
	    && ((firmwareMinor > 2) || (firmwareEdit > 230)) ) ) {
	if (pInfo->devicePhysName[0] != NUL) {
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Physical deviceID:",indent,TRUE);
	    o_lstring(pInfo->devicePhysName,0,TRUE);
	    o_newline(0);
	};
	if (pInfo->deviceXlateName[0] != NUL) {
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Xlation table deviceID:",indent,TRUE);
	    o_lstring(pInfo->deviceXlateName,0,TRUE);
	    o_newline(0);
	};
    };
    if (!pInfo->idFound) return;
    o_lstring(comnt,0,FALSE);
    o_lstring(pInfo->callingName,indent,TRUE);
    o_callerIDtime(pInfo->callTime,0,TRUE);
    o_newline(0);
    return;
}

/* This structure describes the fields of the caller ID header
** It is also a complete description of the Single Data Message Format
** message.
*/
typedef struct {
    UINT8 type;
    UINT8 length;
    char month[2];
    char day[2];
    char hour[2];
    char minute[2];
    char digits[10];
} CALLER_ID;

/* This structure describes the fields of the caller ID Calling Line DN
** parameter.
*/
typedef struct {
    UINT8 type;
    UINT8 length;
    char digits[10];
} CALLER_ID_DN;

/* output the raw data of a callerID structure.
** Assumes that the leading comment
** and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	indent
**	comnt
*/
EXTERNDEF void o_rawCallerID(char *pInfo,int indent,char *comnt)
{
    int i;
    CALLER_ID far *sdmf;

    sdmf = (CALLER_ID *)pInfo;
    o_lstring("length",indent,TRUE);
    o_lnumber(sdmf->length,0,TRUE,10);
    switch(sdmf->type) {
	case 0x04:
	    o_lstring("Single",0,TRUE);
	    o_maxLstring(sdmf->month,0,TRUE,8);
	    o_maxLstring(sdmf->digits,0,TRUE,10);
	    break;
	case 0x80:
	{
	    /* Multiple Data Message Format */
	    int paramStart = 0;
	    o_lstring("Multi",0,TRUE);
	    while (paramStart <= sdmf->length-2) {
		CALLER_ID_DN *mdmf;
		mdmf = (CALLER_ID_DN *)&pInfo[paramStart+2];
		paramStart += mdmf->length + 2;
		if (mdmf->type == 0x01) {
		    CALLER_ID *pMdmf2 = (CALLER_ID *)mdmf;
		    o_newline(0);
		    o_lstring(comnt,0,FALSE);
		    o_maxLstring(pMdmf2->month,indent,TRUE,8);
		    continue;
		};
		if ((mdmf->type == 0x02) || (mdmf->type == 0x04)) {
		    o_newline(0);
		    o_lstring(comnt,0,FALSE);
		    o_lstring("Number:",indent,FALSE);
		    o_maxLstring(mdmf->digits,indent,TRUE,
				 MIN(mdmf->length,20));
		    continue;
		};
		if (mdmf->type == 0x07) {
		    o_newline(0);
		    o_lstring(comnt,0,FALSE);
		    o_lstring("Name:",indent,FALSE);
		    o_maxLstring(mdmf->digits,indent,TRUE,
				 MIN(mdmf->length,20));
		    continue;
		};
	    };
	    break;
	}
	default:
	    o_lstring("Unknown",0,TRUE);
	    break;
    };
    for(i=0;i<128;i+= 1) {
	if (i%10 == 0) {
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	};
	o_lnumber(pInfo[i],indent,TRUE,16);
    };
    o_newline(0);
    return;
}

/* output a NetPhonePrivateStatusInfo_t structure
** comment and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	comnt	comment to prefix lines
**	indent
*/
EXTERNDEF void o_statusInfo(NetPhonePrivateStatusInfo_t *pInfo,char *comnt,
			 int indent)
{
    int i;
    if (pInfo == NULL) return;
    o_lstring("Status Info for block:",0,TRUE);
    o_lnumber(pInfo->block,0,TRUE,10);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("DTMF requests",indent,TRUE);
    o_lnumber(pInfo->dtmfRequests,0,TRUE,10);
    o_lstring("DTMFs now in use",0,TRUE);
    o_lnumber(pInfo->dtmfInUse,0,TRUE,10);
    o_newline(0);
    for (i = 0;i < DIM(pInfo->dtmfGrants);i += 1) {
#define SEND_DTMF 5
#define PROGRESS_DTMF 6
	if (pInfo->dtmfGrants[i] != 0) {
	    o_lstring(comnt,0,FALSE);
	    o_lstring("DTMF",indent+4,FALSE);
	    o_rnumber(i,indent+12,TRUE,10);
	    o_lstring("grants",0,TRUE);
	    o_rnumber(pInfo->dtmfGrants[i],indent+24,TRUE,10);
	    if (i == SEND_DTMF) o_lstring("SEND_DTMF",indent+32,TRUE);
	    if (i == PROGRESS_DTMF) o_lstring("PROGRESS_DTMF",indent+32,TRUE);
	    o_newline(0);
	};
    };
    o_lstring(comnt,0,FALSE);
    o_lstring("Record requests",indent,TRUE);
    o_lnumber(pInfo->recordRequests,0,TRUE,10);
    o_lstring("Record channels now in use",0,TRUE);
    o_lnumber(pInfo->recordInUse,0,TRUE,10);
    o_newline(0);
    for (i = 0;i < DIM(pInfo->recordGrants);i += 1) {
	if (pInfo->recordGrants[i] != 0) {
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Chanl",indent+4,FALSE);
	    o_rnumber(i,indent+12,TRUE,10);
	    o_lstring("grants",0,TRUE);
	    o_rnumber(pInfo->recordGrants[i],indent+24,TRUE,10);
	    o_newline(0);
	};
    };
    o_lstring(comnt,0,FALSE);
    o_lstring("Play requests",indent,TRUE);
    o_lnumber(pInfo->playRequests,0,TRUE,10);
    o_lstring("Play channels now in use",0,TRUE);
    o_lnumber(pInfo->playInUse,0,TRUE,10);
    o_newline(0);
    for (i = 0;i < DIM(pInfo->playGrants);i += 1) {
	if (pInfo->playGrants[i] != 0) {
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Chanl",indent+4,FALSE);
	    o_rnumber(i,indent+12,TRUE,10);
	    o_lstring("grants",0,TRUE);
	    o_rnumber(pInfo->playGrants[i],indent+24,TRUE,10);
	    o_newline(0);
	};
    };
    return;
}

/* output a NetPhonePrivateResources_t structure
** comment and prefix information for the first line is already output.
**
** Inputs:
**	pRes	structure to output
**	comnt	comment to prefix lines
**	indent
*/
EXTERNDEF void o_privateResources(NetPhonePrivateResources_t *pRes,char *comnt,
				  int indent)
{
    if (pRes == NULL) return;
    o_lstring("Board",indent,FALSE);
    o_lnumber(pRes->boardID,0,TRUE,10);
    o_lstring("Number boards",0,TRUE);
    o_lnumber(pRes->numberBoards,0,TRUE,10);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("Extns",indent,FALSE);
    o_lnumber(pRes->totalExtensions,0,TRUE,10);
    o_lstring("Trunks",0,TRUE);
    o_lnumber(pRes->totalTrunks,0,TRUE,10);
    o_lstring("AudioChannels",0,TRUE);
    o_lnumber(pRes->totalAudioChannels,0,TRUE,10);
    o_lstring("DTMFdetectors",0,TRUE);
    o_lnumber(pRes->totalDTMFdetectors,0,TRUE,10);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("analogTrunks",indent,FALSE);
    o_lnumber(pRes->analogTrunks,0,TRUE,10);
    o_lstring("dialToneDetectors",0,TRUE);
    o_lnumber(pRes->dialToneDetectors,0,TRUE,10);
    o_lstring("progressDetectors",0,TRUE);
    o_lnumber(pRes->progressDetectors,0,TRUE,10);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("digitalSpans",indent,FALSE);
    o_lnumber(pRes->digitalSpans,0,TRUE,10);
    o_lstring("digitalTrunks",0,TRUE);
    o_lnumber(pRes->digitalTrunks,0,TRUE,10);
    o_newline(0);
    return;
}

/* output a NetPhonePrivateVersion_t structure
** comment and prefix information for the first line is already output.
**
** Inputs:
**	pVer	structure to output
**	indent
*/
EXTERNDEF void o_privateVersion(NetPhonePrivateVersion_t *pVer,int indent)
{
    if (pVer == NULL) return;
    o_lnumber(pVer->numberBoards,indent,TRUE,10);
    o_lstring("Boards  Board:",0,TRUE);
    o_lnumber(pVer->boardID,0,TRUE,10);
    o_lstring("Type",0,TRUE);
    o_lstring(getBoardTypeName(pVer->boardType),0,TRUE);
    o_lstring(pVer->version,0,TRUE);
    return;
}

/* output a boardList structure
** comment and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	comnt	comment to prefix lines
**	indent
*/
EXTERNDEF void o_boardList(NetPhonePrivateBoardList_t *pInfo,char *comnt,
			 int indent)
{
    int i;
    for (i = 0;i < DIM(pInfo->boardID);i += 1) {
	if ((pInfo->boardID[i].fabDate == 0) && (pInfo->boardID[i].number == 0)) {
	    continue;
	};
	o_newline(0);
	o_lstring(comnt,0,FALSE);
	o_lstring("Board",indent,TRUE);
	o_lnumber(i,0,TRUE,10);
	o_lprintf(indent+10,TRUE,"%06ld%04ld",pInfo->boardID[i].fabDate,
		  pInfo->boardID[i].number);
    };
    o_newline(0);
    return;
}

/* output a system info structure
** comment and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	comnt	comment to prefix lines
**	indent
*/
EXTERNDEF void o_systemInfo(NetPhonePrivateSystemInfo_t *pInfo,char *comnt,
			    int indent)
{
    if (pInfo == NULL) {
	o_newline(0);
	return;
    };
    o_lstring("Trunks for incomming calls",indent,FALSE);
    o_lnumber(pInfo->trunksForIncoming,0,TRUE,10);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("Type of hold audio:",indent,FALSE);
    o_lstring(getSoundIDname(pInfo->audioOnHold),0,TRUE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("Rotate trunk usage:",indent,FALSE);
    o_lstring(pInfo->rotateTrunkUsage ? "ON" : "OFF",0,TRUE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("Digit play time:",indent,FALSE);
    o_lnumber(pInfo->digitPlayTime,0,TRUE,10);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    o_lstring("Digit quiet time:",indent,FALSE);
    o_lnumber(pInfo->digitQuietTime,0,TRUE,10);
    o_newline(0);
    if ((firmwareMajor > 2)
	|| ((firmwareMajor == 2)
	    && ((firmwareMinor > 2) || (firmwareEdit > 230)) ) ) {
	o_lstring(comnt,0,FALSE);
	o_lstring("Flash timeout:",indent,FALSE);
	o_lnumber(pInfo->flashTimeout,0,TRUE,10);
	o_newline(0);
    };
    if ((firmwareMajor > 2)
	|| ((firmwareMajor == 2)
	    && ((firmwareMinor > 7) || (firmwareEdit > 508)) ) ) {
	o_lstring(comnt,0,FALSE);
	o_lstring("CallerID:",indent,FALSE);
	o_lstring(((pInfo->callerIDunavailable) ? "DISABLED" : "ENABLED"),
		  0,TRUE);
	o_newline(0);
    };
    if ((firmwareMajor > 2)
	|| ((firmwareMajor == 2) && (firmwareMinor > 7))) {
	o_lstring(comnt,0,FALSE);
	o_lstring("Park ringback time (tenths):",indent,FALSE);
	o_lnumber(pInfo->parkRingbackTime,0,TRUE,10);
	o_newline(0);
    };
    return;
}

/* output a stream info structure
** comment and prefix information for the first line is already output.
**
** Inputs:
**	pInfo	structure to output
**	indent
*/
EXTERNDEF void o_streamInfo(NetPhonePrivateStreamInfo_t *pInfo,int indent)
{
    if (pInfo == NULL) {
	return;
    };
    o_lstring("Strm:",indent,FALSE);
    o_lnumber(pInfo->streamID,0,TRUE,10);
    o_lstring("SessionID:",0,TRUE);
    o_lnumber(pInfo->sessionID,0,TRUE,16);
    if (!pInfo->inUse) {
	o_lstring("UNUSED",0,TRUE);
    };
    if (!pInfo->privateStatusEnabled) {
	o_lstring("PVT",0,TRUE);
    };
    if (!pInfo->adminEnabled) {
	o_lstring("ADMIN",0,TRUE);
    };
    o_lstring("Dev:",0,TRUE);
    o_lnumber(pInfo->numberDeviceXrefs,0,TRUE,10);
    o_lstring("call:",0,TRUE);
    o_lnumber(pInfo->numberCallXrefs,0,TRUE,10);
    o_lstring("callOn:",0,TRUE);
    o_lnumber(pInfo->numberCallOnDeviceXrefs,0,TRUE,10);
    return;
}

/* output a private message
**
** Inputs:
**	pvt	message to output
**	comnt	comment to prefix lines
**	indent
*/
LOCALDEF void o_privateMessage(NetPhonePrivateMessage_t *pvt,char *comnt,
			       int indent)
{

    static clock_t playEndedTime = 0;
    BOOLEAN newLineOutput = FALSE;
    if (pvt == NULL) return;
    o_firstLineStart(comnt,TRUE);
    o_lstring(getPrivateTypeName(pvt->header.msgType),
	      0,FALSE);
    o_connectionID((ConnectionID_t*)&pvt->header.connection,0,TRUE);
    switch (pvt->header.msgType) {
	case npPvtUnknown:		/* 0 */
	case npPvtStartPrivateStatus:	/* 1 */
	case npPvtStopPrivateStatus:	/* 2 */
	    break;
	case npPvtPlayStarted:		/* 11 */
	{   float delta;
	    delta = getInvokeIDstamp(pvt->data.file.invokeID) - playEndedTime;
	    delta /= CLK_TCK;
	    if (delta > 0) {
		o_rstring("Since end:",timeColumn,TRUE);
		o_lfloat(delta,0,TRUE);
	    };
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_privateFile(&pvt->data.file,indent+16,TRUE);
	    break;
	}
	case npPvtPlay:			/* 10 */
	case npPvtPlayStop:		/* 12 */
	case npPvtPlayNextFile:		/* 14 */
	case npPvtRecord:		/* 15 */
	case npPvtRecordStarted:	/* 16 */
	case npPvtRecordStop:		/* 17 */
	case npPvtRecordEnded:		/* 18 */
	    o_privateFile(&pvt->data.file,indent+16,TRUE);
	    break;
	case npPvtPlayEnded:		/* 13 */
	    o_privateFile(&pvt->data.file,0,TRUE);
	    playEndedTime = getInvokeIDstamp(pvt->data.file.invokeID);
	    break;
	case npPvtGetDigits:		/* 20 */
	case npPvtStopGetDigits:	/* 21 */
	case npPvtDigitsSeen:		/* 22 */
	case npPvtPlayDigits:		/* 23 */
	case npPvtGetDialedDigits:	/* 24 */
	case npPvtReportDialedDigits:	/* 25 */
	    o_lnumber(pvt->data.digits.count,indent+16,TRUE,10);
	    o_lstring("digits for clientID",0,TRUE);
	    o_lnumber(pvt->data.digits.clientID,0,TRUE,10);
	    o_lstring("Digits:",0,TRUE);
	    o_maxLstring(pvt->data.digits.digits,0,TRUE,
			 MIN(20,pvt->data.digits.count));
	    break;
	case npPvtRequestSwrID:		/* 30 */
	case npPvtRequestHdwrID:	/* 32 */
	case npPvtRequestDriverID:	/* 36 */
	    break;
	case npPvtReportSwrID:		/* 31 */
	case npPvtReportHdwrID:		/* 33 */
	case npPvtReportDriverID:	/* 37 */
	    o_maxLstring(pvt->data.textReport.string,indent+16,TRUE,
			DIM(pvt->data.textReport.string));
	    break;
	case npPvtGetStreamInfo:	/* 38 */
	case npPvtReportStreamInfo:	/* 39 */
	    o_streamInfo(&pvt->data.streamInfo,indent+16);
	    break;
	case npPvtGetTranslation:	/* 50 */
	case npPvtSetTranslation:	/* 51 */
	case npPvtReportTranslation:	/* 52 */
	    o_lnumber(pvt->data.xlate.position,indent+16,TRUE,10);
	    o_lstring("of",0,TRUE);
	    o_lnumber(pvt->data.xlate.numPositions,0,TRUE,10);
	    o_char(':',0,FALSE);
	    o_maxLstring(pvt->data.xlate.pattern,0,TRUE,
			DIM(pvt->data.xlate.pattern));
	    o_lstring("-->",0,TRUE);
	    o_maxLstring(pvt->data.xlate.result,0,TRUE,
			DIM(pvt->data.xlate.result));
	    break;
	case npPvtSetLineUserInfo:	/* 69 */
	case npPvtSetLineInfo:		/* 53 */
	case npPvtSetLineInfo1:		/* 54 */
	case npPvtGetLineInfo:		/* 97 */
	case npPvtReportLineInfo:	/* 55 */
	    o_lineInfo(&pvt->data.lineInfo,comnt,indent+16);
	    newLineOutput = TRUE;
	    break;
	case npPvtSetCTIctrl:		/* 56 */
	case npPvtGetCTIctrl:		/* 57 */
	case npPvtReportCTIctrl:	/* 58 */
	    o_lnumber(pvt->data.CTIinfo.clientCTIctrl,indent+16,TRUE,10);
	    o_lstring("0x",indent+16,TRUE);
	    o_lnumber(pvt->data.CTIinfo.clientCTIctrl,0,FALSE,16);
	    break;
	case npPvtEnableAdministration:	/* 59 */
	case npPvtSetAdminPassword:	/* 60 */
	case npPvtSetOperatorNumber:	/* 84 */
	case npPvtGetOperatorNumber:	/* 85 */
	case npPvtReportOperatorNumber:	/* 86 */
	    o_maxLstring(pvt->data.textReport.string,indent+16,TRUE,
			DIM(pvt->data.textReport.string));
	    break;
	case npPvtSetSystemInfo:	/* 61 */
	case npPvtGetSystemInfo:	/* 62 */
	case npPvtReportSystemInfo:	/* 63 */
	    o_systemInfo(&pvt->data.sysInfo,comnt,indent+16);
	    newLineOutput = TRUE;
	    break;
	case npPvtRequestBroadcast:	/* 64 */
	case npPvtReportBroadcast:	/* 65 */
	    o_maxLstring(pvt->data.broadcast.name,indent+16,TRUE,
			 DIM(pvt->data.broadcast.name));
	    o_maxLstring(pvt->data.broadcast.data,indent+30,TRUE,
			 DIM(pvt->data.broadcast.data));
	    break;
	case npPvtGetROMboardOrder:	/* 87 */
	case npPvtReportROMboardOrder:	/* 88 */
	case npPvtGetBoardsPresent:	/* 34 */
	case npPvtReportBoardsPresent:	/* 35 */
	case npPvtGetBoardOrder:	/* 66 */
	case npPvtSetBoardOrder:	/* 67 */
	case npPvtReportBoardOrder:	/* 68 */
	    o_boardList(&pvt->data.boards,comnt,indent);
	    newLineOutput = TRUE;
	    break;
	case npPvtGetStatusInfo:	/* 70 */
	case npPvtReportStatusInfo:	/* 71 */
	    o_statusInfo(&pvt->data.status,comnt,indent);
	    break;
	case npPvtConnectSound:		/* 72 */
	case npPvtDisconnectSound:	/* 73 */
	case npPvtGetSound:		/* 74 */
	case npPvtReportSound:		/* 75 */
	    o_lstring(getSoundIDname(pvt->data.sound.sound),indent+16,TRUE);
	    o_lstring("on:",0,TRUE);
	    o_lfloat((((float)pvt->data.sound.onTime)/40),0,TRUE);
	    o_lstring("off:",0,TRUE);
	    o_lfloat((((float)pvt->data.sound.offTime)/40),0,TRUE);
	    break;
	case npPvtRequestInterlock:	/* 76 */
	case npPvtReleaseInterlock:	/* 77 */
	case npPvtReportInterlock:	/* 78 */
	    o_lstring((pvt->data.interlock.wasTaken) ? "Taken" : "InUse",
		      indent+20,TRUE);
	    o_lnumber(pvt->data.interlock.value,indent+20,TRUE,10);
	    break;
	case npPvtMakeCTIcall:		/* 79 */
	    o_lnumber(pvt->data.CTIinfo.clientCTIctrl,indent+16,TRUE,10);
	    o_lstring("0x",indent+16,TRUE);
	    o_lnumber(pvt->data.CTIinfo.clientCTIctrl,0,FALSE,16);
	    break;
	case npPvtGetCallerIDinfo:	/* 80 */
	case npPvtReportCallerIDinfo:	/* 81 */
	    o_newline(0);
	    o_callerIDinfo(&pvt->data.callerID,indent+16,comnt);
	    newLineOutput = TRUE;
	    break;
	case npPvtGetRawCallerID:	/* 82 */
	case npPvtReportRawCallerID:	/* 83 */
	    o_rawCallerID(&pvt->data.textReport.string[0],indent+16,comnt);
	    break;
	case npPvtGetSoftwareVersion:	/* 90 */
	case npPvtReportSoftwareVersion: /* 91 */
	    o_privateVersion(&pvt->data.version,indent+16);
	    break;
	case npPvtGetSystemResources:	/* 92 */
	case npPvtGetBoardResources:	/* 93 */
	case npPvtReportSystemResources: /* 94 */
	case npPvtReportBoardResources:	/* 95 */
	    o_privateResources(&pvt->data.resources,comnt,indent+16);
	    newLineOutput = TRUE;
	    break;
	case npPvtSetDSpanInfo:		/* 98 */
	case npPvtGetDSpanInfo:		/* 99 */
	case npPvtReportDSpanInfo:	/* 100 */
	    o_DSpanInfo(&pvt->data.DSpan,comnt,indent+16);
	    newLineOutput = TRUE;
	    break;
	case npPvtSetMoreLineInfo:	/* 101 */
	case npPvtGetMoreLineInfo:	/* 102 */
	case npPvtReportMoreLineInfo:	/* 103 */
	case npPvtSetDigitalTrunkFlags:	/* 104 */
	    o_moreLineInfo(&pvt->data.moreLineInfo,comnt,indent+16);
	    newLineOutput = TRUE;
	    break;
	case npPvtGetCallInfo:		/* 105 */
	case npPvtReportCallInfo:	/* 106 */
	case npPvtGetParkInfo:		/* 107 */
	case npPvtParkCall:		/* 108 */
	case npPvtParkCallOccured:	/* 109 */
	    o_callInfo(&pvt->data.callInfo,indent+16);
	    break;
 	default:
	    o_lstring("Unknown private message",indent,FALSE);
	    break;
    };
    if (!newLineOutput) o_newline(0);
    return;
}

/* Output a CSTA unsolicited event
**
** Inputs:
**	eType	event type
**	event
**	pvt	private data
**	comnt	comment string with which to prefix each line
**	verbose	TRUE for multiline form
*/
LOCALDEF void o_CSTAunsolicited(EventType_t eType,
				CSTAUnsolicitedEvent *event,
				NetPhonePrivateMessage_t *pvt,
				char *comnt,BOOLEAN verbose)
{
    const int indent = 8;
    if (event == NULL) return;
    switch (eType) {
	case CSTA_CALL_CLEARED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Call cleared:",indent,FALSE);
	    o_connectionID(&event->u.callCleared.clearedCall,0,TRUE);
	    o_lstring(getLocalStateName(event->u.callCleared.localConnectionInfo),
			      0,TRUE);
	    o_lstring(getCauseName(event->u.callCleared.cause),0,TRUE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_CONFERENCED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Conference:",indent,FALSE);
	    o_newline(0);
	    o_lstring("Primary call:",indent+4,FALSE);
	    o_connectionID(&event->u.conferenced.primaryOldCall,0,TRUE);
	    o_lstring(getCauseName(event->u.conferenced.cause),
		      0,TRUE);
	    o_newline(0);
	    o_lstring("Secondary call:",indent+4,FALSE);
	    o_connectionID(&event->u.conferenced.secondaryOldCall,0,TRUE);
	    if (event->u.conferenced.confController.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring("Controller:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.conferenced.confController,
				   0,TRUE);
	    };
	    if (event->u.conferenced.addedParty.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring("Added party:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.conferenced.addedParty,0,TRUE);
		o_lstring(getLocalStateName(event->u.conferenced.localConnectionInfo),
			  0,TRUE);
	    };
	    o_newline(0);
	    o_connectionList(&event->u.conferenced.conferenceConnections,
			     indent+4,comnt);
	    return;
	case CSTA_CONNECTION_CLEARED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Connection cleared:",indent,FALSE);
	    o_connectionID(&event->u.connectionCleared.droppedConnection,
			   0,TRUE);
	    o_lstring(getCauseName(event->u.connectionCleared.cause),
		      0,TRUE);
	    if (event->u.connectionCleared.releasingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Subject:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.connectionCleared.releasingDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;	    
	case CSTA_DELIVERED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Delivered:",indent,FALSE);
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Alerting:",indent+4,FALSE);
	    o_connectionID(&event->u.delivered.connection,0,TRUE);
	    o_lstring(getCauseName(event->u.delivered.cause),
		      0,TRUE);
	    if (event->u.delivered.alertingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Alerting:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.delivered.alertingDevice,0,TRUE);
		o_lstring(getLocalStateName(event->u.delivered.localConnectionInfo),
			  0,TRUE);
	    };
	    if (event->u.delivered.callingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Calling:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.delivered.callingDevice,0,TRUE);
	    };
	    if (event->u.delivered.calledDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Called:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.delivered.calledDevice,0,TRUE);
	    };
	    if (event->u.delivered.lastRedirectionDevice.deviceIDType
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Last redirection:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.delivered.calledDevice,0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;	    
	case CSTA_DIVERTED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Diverted:",indent,FALSE);
	    o_connectionID(&event->u.diverted.connection,0,TRUE);
	    o_lstring(getCauseName(event->u.diverted.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.diverted.localConnectionInfo),
		      0,TRUE);
	    if (event->u.diverted.divertingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Diverting:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.diverted.divertingDevice,0,TRUE);
	    };
	    if (event->u.diverted.divertingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("New destination:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.diverted.newDestination,0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;	    
	case CSTA_ESTABLISHED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Established:",indent,FALSE);
	    o_connectionID(&event->u.established.establishedConnection,0,TRUE);
	    o_lstring(getCauseName(event->u.established.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.established.localConnectionInfo),
		      0,TRUE);
	    if (event->u.established.answeringDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Answering:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.established.answeringDevice,
				   0,TRUE);
	    };
	    if (event->u.established.callingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Calling:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.established.callingDevice,0,TRUE);
	    };
	    if (event->u.established.calledDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Called:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.established.calledDevice,0,TRUE);
	    };
	    if (event->u.established.lastRedirectionDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Calling:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.established.lastRedirectionDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_FAILED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Failed:",indent,FALSE);
	    o_connectionID(&event->u.failed.failedConnection,0,TRUE);
	    o_lstring(getCauseName(event->u.failed.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.failed.localConnectionInfo),
		      0,TRUE);
	    if (event->u.failed.failingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Failing:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.failed.failingDevice,
				   0,TRUE);
	    };
	    if (event->u.failed.calledDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Called:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.failed.calledDevice,0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_HELD:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Held:",indent,FALSE);
	    o_connectionID(&event->u.held.heldConnection,0,TRUE);
	    o_lstring(getCauseName(event->u.held.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.held.localConnectionInfo),
		      0,TRUE);
	    if (event->u.held.holdingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Holding:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.held.holdingDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_NETWORK_REACHED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Network reached:",indent,FALSE);
	    o_connectionID(&event->u.networkReached.connection,0,TRUE);
	    o_lstring(getCauseName(event->u.networkReached.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.networkReached.localConnectionInfo),
		      0,TRUE);
	    if (event->u.networkReached.trunkUsed.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Trunk:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.networkReached.trunkUsed,
				   0,TRUE);
	    };
	    if (event->u.networkReached.calledDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Called:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.networkReached.calledDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_ORIGINATED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Originated:",indent,FALSE);
	    o_connectionID(&event->u.originated.originatedConnection,0,TRUE);
	    o_lstring(getCauseName(event->u.originated.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.originated.localConnectionInfo),
		      0,TRUE);
	    if (event->u.originated.callingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Calling:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.originated.callingDevice,
				   0,TRUE);
	    };
	    if (event->u.originated.calledDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Called:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.originated.calledDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_QUEUED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Queued",indent,FALSE);
	    o_lnumber(event->u.queued.numberQueued,0,FALSE,10);
	    o_lstring(":",0,FALSE);
	    o_connectionID(&event->u.queued.queuedConnection,0,TRUE);
	    o_lstring(getCauseName(event->u.queued.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.queued.localConnectionInfo),
		      0,TRUE);
	    if (event->u.queued.queue.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Queue:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.queued.queue,
				   0,TRUE);
	    };
	    if (event->u.queued.callingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Calling:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.queued.callingDevice,
				   0,TRUE);
	    };
	    if (event->u.queued.calledDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Called:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.queued.calledDevice,
				   0,TRUE);
	    };
	    if (event->u.queued.queue.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Last redirection:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.queued.lastRedirectionDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_RETRIEVED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Retrieved:",indent,FALSE);
	    o_connectionID(&event->u.retrieved.retrievedConnection,0,TRUE);
	    o_lstring(getCauseName(event->u.retrieved.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.retrieved.localConnectionInfo),
		      0,TRUE);
	    if (event->u.retrieved.retrievingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Retrieving:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.retrieved.retrievingDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_SERVICE_INITIATED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Service initiated:",indent,FALSE);
	    o_connectionID(&event->u.serviceInitiated.initiatedConnection,
			   0,TRUE);
	    o_lstring(getCauseName(event->u.serviceInitiated.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.serviceInitiated.localConnectionInfo),
		      0,TRUE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_TRANSFERRED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Transferred:",indent,FALSE);
	    o_connectionID(&event->u.transferred.primaryOldCall,0,TRUE);
	    o_lstring(getCauseName(event->u.transferred.cause),
		      0,TRUE);
	    o_lstring(getLocalStateName(event->u.transferred.localConnectionInfo),
		      0,TRUE);
	    o_newline(0);
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Secondary:",indent+2,FALSE);
	    o_connectionID(&event->u.transferred.secondaryOldCall,0,TRUE);
	    if (event->u.transferred.transferringDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Transferring:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.transferred.transferringDevice,
				   0,TRUE);
	    };
	    if (event->u.transferred.transferredDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Transferred:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.transferred.transferredDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_CALL_INFORMATION:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Call information:",indent,FALSE);
	    o_connectionID(&event->u.callInformation.connection,0,TRUE);
	    if (event->u.callInformation.device.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Device:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.callInformation.device,
				   0,TRUE);
	    };
	    {
		char string[sizeof(AccountInfo_t)+1];
		o_newline(0);
		strncpy(string,(char *)&event->u.callInformation.accountInfo,
			sizeof(AccountInfo_t));
		string[sizeof(AccountInfo_t)] = NUL;
		o_lstring(comnt,0,FALSE);
		o_lstring("Account Info:",indent+4,FALSE);
		o_lstring(string,0,FALSE);
	    }
	    {
		char string[sizeof(AuthCode_t)+1];
		o_newline(0);
		strncpy(string,
			(char *)&event->u.callInformation.authorisationCode,
			sizeof(AuthCode_t));
		string[sizeof(AuthCode_t)] = NUL;
		o_lstring(comnt,0,FALSE);
		o_lstring("Authorisation code:",indent+4,FALSE);
		o_lstring(string,0,FALSE);
	    }
	    if (verbose) o_newline(0);
	    break;
	case CSTA_DO_NOT_DISTURB:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Do not disturb:",indent,FALSE);
	    o_lstring(((event->u.doNotDisturb.doNotDisturbOn)
		       ? "ON" : "OFF"),0,TRUE);
	    if (event->u.doNotDisturb.device.deviceIDStatus
		== ID_PROVIDED) {
		o_extendedDeviceID(&event->u.doNotDisturb.device,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_FORWARDING:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Forwarding:",indent,FALSE);
	    if (event->u.forwarding.device.deviceIDStatus
		== ID_PROVIDED) {
		o_extendedDeviceID(&event->u.forwarding.device,
				   0,TRUE);
	    };
	    o_forwardingInfo(&event->u.forwarding.forwardingInformation,
			     indent+30,TRUE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_MESSAGE_WAITING:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Message waiting:",indent,FALSE);
	    o_lstring(((event->u.messageWaiting.messageWaitingOn) ? "ON" : "OFF"),
		      0,TRUE);
	    if (event->u.messageWaiting.deviceForMessage.deviceIDStatus
		== ID_PROVIDED) {
		o_extendedDeviceID(&event->u.messageWaiting.deviceForMessage,
				   0,TRUE);
	    };
	    if (event->u.messageWaiting.invokingDevice.deviceIDStatus
		== ID_PROVIDED) {
		o_newline(0);
		o_lstring(comnt,0,FALSE);
		o_lstring("Device:",indent+4,FALSE);
		o_extendedDeviceID(&event->u.messageWaiting.invokingDevice,
				   0,TRUE);
	    };
	    if (verbose) o_newline(0);
	    break;
	case CSTA_LOGGED_ON:
	    o_firstLineStart(comnt,verbose);
//	    event->u.loggedOn;
	    o_lstring("CSTA_LOGGED_ON",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_LOGGED_OFF:
	    o_firstLineStart(comnt,verbose);
//	    event->u.loggedOff;
	    o_lstring("CSTA_LOGGED_OFF",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_NOT_READY:
	    o_firstLineStart(comnt,verbose);
//	    event->u.notReady;
	    o_lstring("CSTA_NOT_READY",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_READY:
	    o_firstLineStart(comnt,verbose);
//	    event->u.ready;
	    o_lstring("CSTA_READY",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_WORK_NOT_READY:
	    o_firstLineStart(comnt,verbose);
//	    event->u.workNotReady;
	    o_lstring("CSTA_WORK_NOT_READY",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_WORK_READY:
	    o_firstLineStart(comnt,verbose);
//	    event->u.workReady;
	    o_lstring("CSTA_WORK_READY",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_PRIVATE_STATUS:
	    o_privateMessage(pvt,comnt,indent);
	    return;
	case CSTA_BACK_IN_SERVICE:
	    o_firstLineStart(comnt,verbose);
//	    event->u.backInService;
	    o_lstring("CSTA_BACK_IN_SERVICE",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_OUT_OF_SERVICE:
	    o_firstLineStart(comnt,verbose);
//	    event->u.outOfService;
	    o_lstring("CSTA_OUT_OF_SERVICE",indent,FALSE);
	    if (verbose) o_newline(0);
	    break;
	case CSTA_MONITOR_ENDED:
	    o_firstLineStart(comnt,verbose);
	    o_lstring("Monitor ended:",indent,FALSE);
	    o_lstring(getCauseName(event->u.monitorEnded.cause),
		      0,TRUE);
	    if (verbose) o_newline(0);
	    break;
	default:
	    o_firstLineStart(comnt,verbose);
	    o_lprintf(0,FALSE,"CSTAunsolicited ?%d?",eType);
	    if (verbose) o_newline(0);
	    break;
    };
    return;
}

/* output register request ID
**
** Inputs:
**	reqID	    register request
**	column		column for first character
**	flag		TRUE if whitespace must preceed string
*/
LOCALDEF void o_registerReqID(RouteRegisterReqID_t *reqID,
			      int column,BOOLEAN flag)
{
    if (reqID == NULL) return;
    o_lstring("@@@NIY registerReqID",column,flag);
    return;
}

/* output one string on a line
**
** Inputs:
**	string
*/
LOCALDEF void o_APIcap(char *comnt,char *string)
{
    o_lstring(comnt,0,FALSE);
    o_lstring(string,0,FALSE);
    o_newline(0);
}

/* output snapshot call
**
** Inputs:
**	snapData	snapshot to report
**	column		column for first character
**	comnt	comment string with which to prefix each line
*/
LOCALDEF void o_snapshotCall(CSTASnapshotCallData_t *snapData,int column,
			     char *comnt)
{
    CSTASnapshotCallResponseInfo_t FAR *pInfo;
    int i;
    if (snapData == NULL) return;
    pInfo = snapData->info;
    if (pInfo == NULL) return;
    for (i = 0;i < MIN(24,snapData->count);i += 1) {
	o_lstring(comnt,0,FALSE);
	o_connectionID(&pInfo->callIdentifier,column,TRUE);
	o_lstring(getLocalStateName(pInfo->localConnectionState),
		  column+20,TRUE);
	o_extendedDeviceID(&pInfo->deviceOnCall,column+30,TRUE);
	o_newline(0);
	pInfo += 1;
    };
    return;
}

/* output CSTACallState_t
**
** Inputs:
**	pState		pointer to info
**	indent		column for first character
**	flag		TRUE if whitespace must preceed string
*/
LOCALDEF void o_callState(CSTACallState_t FAR *pState,int indent,BOOLEAN flag)
{
    BOOLEAN firstTime = FALSE;
    int i;
    int count;
    LocalConnectionState_t FAR *pList;
    if (pState == NULL) return;
    pList = pState->state;
    if (pList == NULL) return;
    count = (int)pState->count;
    if ((count < 0) || (count > 24)) {
	o_lprintf(indent,FALSE,"? Illegal CSTACallState_t count %d",count);
	return;
    };
    for (i = 0;i < count;i += 1) {
	char *name;
	name = getLocalStateCharacter(pList[i]);
	if (name == NULL) continue;
	o_lnumber(i,indent,(firstTime) ? flag : (BOOLEAN)TRUE,10);
	o_char(':',0,FALSE);
	o_lstring(name,0,FALSE);
	firstTime = FALSE;
    };
    return;
}

/* output CSTASnapshotDeviceResponseInfo_t
**
** Inputs:
**	pInfo	pointer to info
**	indent		column for first character
**	flag		TRUE if whitespace must preceed string
*/
LOCALDEF void o_deviceResponseInfo(CSTASnapshotDeviceResponseInfo_t FAR *info,
				  int indent,BOOLEAN flag)
{
    if (info == NULL) return;
    o_connectionID(&info->callIdentifier,indent,flag);
    o_callState(&info->localCallState,indent,TRUE);
    return;
}
    
/* output snapshot of a device
**
** Inputs:
**	snapData	snapshot data
**	indent		column for first character
**	comnt	comment string with which to prefix each line
*/
LOCALDEF void o_snapshotDevice(CSTASnapshotDeviceData_t *pSnapData,int indent,
			       char *comnt)
{
    CSTASnapshotDeviceResponseInfo_t FAR *pInfo;
    int i;
    int count;
    if (pSnapData == NULL) return;
    count = (int)pSnapData->count;
    if ((count < 0) || (count > 24)) {
	o_lstring(comnt,0,FALSE);
	o_lprintf(indent,FALSE,"? Illegal CSTASnapshotDeviceData_t count %d",
		  count);
	o_newline(0);
	return;
    };
    pInfo = pSnapData->info;
    for (i = 0;i < count;i += 1) {
	o_lstring(comnt,0,FALSE);
	o_deviceResponseInfo(&pInfo[i],indent,TRUE);
	o_newline(0);
    };
    return;
}

/* Print the filters
*/
EXTERNDEF void o_printfilter(CSTAMonitorFilter_t filter, char *title,
			     int column, char *comnt, BOOLEAN dIfEmpty)
{
    char string[STRING_LENGTH+1];
    int gotSome = filter.call||filter.feature||filter.agent||
		  filter.maintenance||filter.privateFilter;
    if ((!dIfEmpty) && (!gotSome))
	return;
    o_lstring(comnt,0,FALSE);
    o_lstring(title, column, FALSE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    sprintf(string, "Call       : (0x%4.4x) ", filter.call);
    o_lstring(string,column+4,FALSE);
    o_lstring(filStr.getFilterString(FILTER_CALL, (unsigned long)filter.call), 0, FALSE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    sprintf(string, "feature    : (0x%2.2x) ", filter.feature);
    o_lstring(string,column+4,FALSE);
    o_lstring(filStr.getFilterString(FILTER_FEATURE, (unsigned long)filter.feature), 0, FALSE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    sprintf(string, "agent      : (0x%2.2x) ", filter.agent);
    o_lstring(string,column+4,FALSE);
    o_lstring(filStr.getFilterString(FILTER_AGENT, (unsigned long)filter.agent), 0, FALSE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    sprintf(string, "maintenance: (0x%2.2x) ", filter.maintenance);
    o_lstring(string,column+4,FALSE);
    o_lstring(filStr.getFilterString(FILTER_MAINTENANCE, (unsigned long)filter.maintenance),
	      0, FALSE);
    o_newline(0);
    o_lstring(comnt,0,FALSE);
    sprintf(string, "private    : (0x%8.8lx) ", filter.privateFilter);
    o_lstring(string,column+4,FALSE);
    o_lstring(filStr.getFilterString(FILTER_PRIVATE, (unsigned long)filter.privateFilter),
	      0, FALSE);
    o_newline(0);
    
    return;
}

/* output filters
**
** Inputs:
**	snapData	snapshot data
**	column		column for first character
**	comnt	comment string with which to prefix each line
*/
LOCALDEF void o_monitorFilters(CSTAMonitorFilter_t *filter,int column,
			       char *comnt)
{
    if (filter == NULL) return;
    o_printfilter(*filter, "Filters:", column, comnt, FALSE);
    return;
}

/* output forward list of a device
**
** Inputs:
**	forward
**	column		column for first character
**	comnt		comment prefix for each line
*/
LOCALDEF void o_forwardList(ListForwardParameters_t *forward,int column,
			    char *comnt)
{
    int i;
    int count;
    if (forward == NULL) return;
    if ((forward->count < 0) || (forward->count > DIM(forward->param))) {
	o_lstring(comnt,column,FALSE);
	o_lstring("?Unreasonable number of forwarding paramters",0,FALSE);
	o_newline(0);
    };
    count = MIN(forward->count,DIM(forward->param));
    for (i = 0;i < count;i += 1) {
	o_lstring(comnt,0,FALSE);
	o_forwardingInfo(&forward->param[i],column,FALSE);
	o_newline(0);
    };
    return;
}

/* Output a CSTA confirmation event
**
** Inputs:
**	eType	event type
**	event
**	pvt	private data associated with the event
**	comnt	comment string with which to prefix each line
**	verbose	TRUE for multiline form
*/
LOCALDEF void o_CSTAconfirmation(EventType_t eType,
				 CSTAConfirmationEvent *event,
				 NetPhonePrivateMessage_t *pvt,
				 char *comnt,BOOLEAN verbose)
{
    const int indent = 8;
    if (event == NULL) return;
    o_firstLineStart(comnt,verbose);
    o_lstring(getTypeName(CSTACONFIRMATION,eType),0,FALSE);
    o_invokeIDwithTime(event->invokeID,timeColumn,TRUE);
    o_newline(0);
    if (!verbose) return;
    switch (eType) {
	/* return from a case to suppress final newline */
	case CSTA_ALTERNATE_CALL_CONF:
	    break;
	case CSTA_ANSWER_CALL_CONF:
	    break;
	case CSTA_CALL_COMPLETION_CONF:
	    break;
	case CSTA_CLEAR_CALL_CONF:
	    break;
	case CSTA_CLEAR_CONNECTION_CONF:
	    break;
	case CSTA_CONFERENCE_CALL_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("New call:",indent,TRUE);
	    o_connectionID(&event->u.conferenceCall.newCall,indent,TRUE);
	    o_newline(0);
	    o_connectionList(&event->u.conferenceCall.connList,indent+4,comnt);
	    break;
	case CSTA_CONSULTATION_CALL_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("New call:",indent,TRUE);
	    o_connectionID(&event->u.consultationCall.newCall,indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_DEFLECT_CALL_CONF:
	    break;
	case CSTA_PICKUP_CALL_CONF:
	    break;
	case CSTA_GROUP_PICKUP_CALL_CONF:
	    break;
	case CSTA_HOLD_CALL_CONF:
	    break;
	case CSTA_MAKE_CALL_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Making call:",indent,TRUE);
	    o_connectionID(&event->u.makeCall.newCall,indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_MAKE_PREDICTIVE_CALL_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Making predictive call:",indent,TRUE);
	    o_connectionID(&event->u.makePredictiveCall.newCall,indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_QUERY_MWI_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Mwi:",indent,TRUE);
	    o_lstring((event->u.queryMwi.messages) ? "ON" : "OFF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_QUERY_DND_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Dnd:",indent,TRUE);
	    o_lstring((event->u.queryDnd.doNotDisturb) ? "ON" : "OFF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_QUERY_FWD_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Forwards:",indent,TRUE);
	    o_newline(0);
	    o_forwardList(&event->u.queryFwd.forward,indent,comnt);
	    return;
	case CSTA_QUERY_AGENT_STATE_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Agent state:",indent,TRUE);
	    o_lstring(getAgentStateName(event->u.queryAgentState.agentState),
		      0,TRUE);
	    o_newline(0);
	    break;
	case CSTA_QUERY_LAST_NUMBER_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Last number:",indent,TRUE);
	    o_deviceID(&event->u.queryLastNumber.lastNumber,indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_QUERY_DEVICE_INFO_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Device Info:",indent,TRUE);
	    o_deviceID(&event->u.queryDeviceInfo.device,indent,TRUE);
	    o_lstring(getDeviceTypeName(event->u.queryDeviceInfo.deviceType),
		      0,TRUE);
	    o_lstring(getDeviceClassList(event->u.queryDeviceInfo.deviceClass),
		      0,TRUE);
	    o_newline(0);
	    break;
	case CSTA_RECONNECT_CALL_CONF:
	    break;
	case CSTA_RETRIEVE_CALL_CONF:
	    break;
	case CSTA_SET_MWI_CONF:
	    break;
	case CSTA_SET_DND_CONF:
	    break;
	case CSTA_SET_FWD_CONF:
	    break;
	case CSTA_SET_AGENT_STATE_CONF:
	    break;
	case CSTA_TRANSFER_CALL_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Transfer:",indent,TRUE);
	    o_connectionID(&event->u.transferCall.newCall,indent,TRUE);
	    o_newline(0);
	    o_connectionList(&event->u.transferCall.connList,indent+4,comnt);
	    return;
	case CSTA_UNIVERSAL_FAILURE_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring(getCSTAfailureName(event->u.universalFailure.error),
		      indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_ROUTE_REGISTER_REQ_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_registerReqID(&event->u.routeRegister.registerReqID,
			    indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_ROUTE_REGISTER_CANCEL_CONF:
	    break;
	case CSTA_ESCAPE_SVC_CONF:
	    o_privateMessage(pvt,comnt,indent);
	    return;
	case CSTA_SYS_STAT_REQ_CONF:
	    // @@@resolve confusion over system status req events
	    // @@@NetPhone does not have these yet.
	    // @@@event->u.sysStatReq;
	    o_lstring(comnt,0,FALSE);
	    o_lstring("@@@ CSTA_SYS_STAT_REQ_CONF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_SYS_STAT_START_CONF:
	    // @@@NetPhone does not have these yet.
	    //@@@ event->u.sysStatStart;
	    o_lstring(comnt,0,FALSE);
	    o_lstring("@@@ CSTA_SYS_STAT_START_CONF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_SYS_STAT_STOP_CONF:
	    // @@@NetPhone does not have these yet.
	    // @@@event->u.sysStatStop;
	    o_lstring(comnt,0,FALSE);
	    o_lstring("@@@ CSTA_SYS_STAT_STOP_CONF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_CHANGE_SYS_STAT_FILTER_CONF:
	    // @@@NetPhone does not have these yet.
	    // @@@event->u.changeSysStatFilter;
	    o_lstring(comnt,0,FALSE);
	    o_lstring("@@@ CSTA_SYS_STAT_FILTER_CONF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_REQ_SYS_STAT_CONF:
	    /* @@@why is there a req_sys_stat and a sys_stat_req? */
	    o_lstring(comnt,0,FALSE);
	    o_lstring("@@@ CSTA_SYS_STAT_CONF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_MONITOR_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Monitor XrefID:",indent,TRUE);
	    o_lnumber(event->u.monitorStart.monitorCrossRefID,0,TRUE,10);
	    o_lstring("Filters:",indent,TRUE);
	    o_newline(0);
	    o_monitorFilters(&event->u.monitorStart.monitorFilter,indent,comnt);
	    return;
	case CSTA_CHANGE_MONITOR_FILTER_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("Change filters:",indent,TRUE);
	    o_newline(0);
	    o_monitorFilters(&event->u.changeMonitorFilter.monitorFilter,
			     indent,comnt);
	    return;
	case CSTA_MONITOR_STOP_CONF:
	    break;
	case CSTA_SNAPSHOT_CALL_CONF:
	    o_snapshotCall(&event->u.snapshotCall.snapshotData,
			       indent,comnt);
	    return;
	case CSTA_SNAPSHOT_DEVICE_CONF:
	    o_snapshotDevice(&event->u.snapshotDevice.snapshotData,
			     indent,comnt);
	    return;
	case CSTA_GETAPI_CAPS_CONF:
	    if (event->u.getAPICaps.alternateCall) o_APIcap(comnt,"alternateCall");
	    if (event->u.getAPICaps.answerCall) o_APIcap(comnt,"answerCall");
	    if (event->u.getAPICaps.callCompletion) o_APIcap(comnt,"callCompletion");
	    if (event->u.getAPICaps.clearCall) o_APIcap(comnt,"clearCall");
	    if (event->u.getAPICaps.clearConnection) o_APIcap(comnt,"clearConnection");
	    if (event->u.getAPICaps.conferenceCall) o_APIcap(comnt,"conferenceCall");
	    if (event->u.getAPICaps.consultationCall) o_APIcap(comnt,"consultationCall");
	    if (event->u.getAPICaps.deflectCall) o_APIcap(comnt,"deflectCall");
	    if (event->u.getAPICaps.pickupCall) o_APIcap(comnt,"pickupCall");
	    if (event->u.getAPICaps.groupPickupCall) o_APIcap(comnt,"groupPickupCall");
	    if (event->u.getAPICaps.holdCall) o_APIcap(comnt,"holdCall");
	    if (event->u.getAPICaps.makeCall) o_APIcap(comnt,"makeCall");
	    if (event->u.getAPICaps.makePredictiveCall) o_APIcap(comnt,"makePredictiveCall");
	    if (event->u.getAPICaps.queryMwi) o_APIcap(comnt,"queryMwi");
	    if (event->u.getAPICaps.queryDnd) o_APIcap(comnt,"queryDnd");
	    if (event->u.getAPICaps.queryFwd) o_APIcap(comnt,"queryFwd");
	    if (event->u.getAPICaps.queryAgentState) o_APIcap(comnt,"queryAgentState");
	    if (event->u.getAPICaps.queryLastNumber) o_APIcap(comnt,"queryLastNumber");
	    if (event->u.getAPICaps.queryDeviceInfo) o_APIcap(comnt,"queryDeviceInfo");
	    if (event->u.getAPICaps.reconnectCall) o_APIcap(comnt,"reconnectCall");
	    if (event->u.getAPICaps.retrieveCall) o_APIcap(comnt,"retrieveCall");
	    if (event->u.getAPICaps.setMwi) o_APIcap(comnt,"setMwi");
	    if (event->u.getAPICaps.setDnd) o_APIcap(comnt,"setDnd");
	    if (event->u.getAPICaps.setFwd) o_APIcap(comnt,"setFwd");
	    if (event->u.getAPICaps.setAgentState) o_APIcap(comnt,"setAgentState");
	    if (event->u.getAPICaps.transferCall) o_APIcap(comnt,"transferCall");
	    if (event->u.getAPICaps.eventReport) o_APIcap(comnt,"eventReport");
	    if (event->u.getAPICaps.callClearedEvent) o_APIcap(comnt,"callClearedEvent");
	    if (event->u.getAPICaps.conferencedEvent) o_APIcap(comnt,"conferencedEvent");
	    if (event->u.getAPICaps.connectionClearedEvent) o_APIcap(comnt,"connectionClearedEvent");
	    if (event->u.getAPICaps.deliveredEvent) o_APIcap(comnt,"deliveredEvent");
	    if (event->u.getAPICaps.divertedEvent) o_APIcap(comnt,"divertedEvent");
	    if (event->u.getAPICaps.establishedEvent) o_APIcap(comnt,"establishedEvent");
	    if (event->u.getAPICaps.failedEvent) o_APIcap(comnt,"failedEvent");
	    if (event->u.getAPICaps.heldEvent) o_APIcap(comnt,"heldEvent");
	    if (event->u.getAPICaps.networkReachedEvent) o_APIcap(comnt,"networkReachedEvent");
	    if (event->u.getAPICaps.originatedEvent) o_APIcap(comnt,"originatedEvent");
	    if (event->u.getAPICaps.queuedEvent) o_APIcap(comnt,"queuedEvent");
	    if (event->u.getAPICaps.retrievedEvent) o_APIcap(comnt,"retrievedEvent");
	    if (event->u.getAPICaps.serviceInitiatedEvent) o_APIcap(comnt,"serviceInitiatedEvent");
	    if (event->u.getAPICaps.transferredEvent) o_APIcap(comnt,"transferredEvent");
	    if (event->u.getAPICaps.callInformationEvent) o_APIcap(comnt,"callInformationEvent");
	    if (event->u.getAPICaps.doNotDisturbEvent) o_APIcap(comnt,"doNotDisturbEvent");
	    if (event->u.getAPICaps.forwardingEvent) o_APIcap(comnt,"forwardingEvent");
	    if (event->u.getAPICaps.messageWaitingEvent) o_APIcap(comnt,"messageWaitingEvent");
	    if (event->u.getAPICaps.loggedOnEvent) o_APIcap(comnt,"loggedOnEvent");
	    if (event->u.getAPICaps.loggedOffEvent) o_APIcap(comnt,"loggedOffEvent");
	    if (event->u.getAPICaps.notReadyEvent) o_APIcap(comnt,"notReadyEvent");
	    if (event->u.getAPICaps.readyEvent) o_APIcap(comnt,"readyEvent");
	    if (event->u.getAPICaps.workNotReadyEvent) o_APIcap(comnt,"workNotReadyEvent");
	    if (event->u.getAPICaps.workReadyEvent) o_APIcap(comnt,"workReadyEvent");
	    if (event->u.getAPICaps.backInServiceEvent) o_APIcap(comnt,"backInServiceEvent");
	    if (event->u.getAPICaps.outOfServiceEvent) o_APIcap(comnt,"outOfServiceEvent");
	    if (event->u.getAPICaps.privateEvent) o_APIcap(comnt,"privateEvent");
	    if (event->u.getAPICaps.routeRequestEvent) o_APIcap(comnt,"routeRequestEvent");
	    if (event->u.getAPICaps.reRoute) o_APIcap(comnt,"reRoute");
	    if (event->u.getAPICaps.routeSelect) o_APIcap(comnt,"routeSelect");
	    if (event->u.getAPICaps.routeUsedEvent) o_APIcap(comnt,"routeUsedEvent");
	    if (event->u.getAPICaps.routeEndEvent) o_APIcap(comnt,"routeEndEvent");
	    if (event->u.getAPICaps.monitorDevice) o_APIcap(comnt,"monitorDevice");
	    if (event->u.getAPICaps.monitorCall) o_APIcap(comnt,"monitorCall");
	    if (event->u.getAPICaps.monitorCallsViaDevice) o_APIcap(comnt,"monitorCallsViaDevice");
	    if (event->u.getAPICaps.changeMonitorFilter) o_APIcap(comnt,"changeMonitorFilter");
	    if (event->u.getAPICaps.monitorStop) o_APIcap(comnt,"monitorStop");
	    if (event->u.getAPICaps.monitorEnded) o_APIcap(comnt,"monitorEnded");
	    if (event->u.getAPICaps.snapshotDeviceReq) o_APIcap(comnt,"snapshotDeviceReq");
	    if (event->u.getAPICaps.snapshotCallReq) o_APIcap(comnt,"snapshotCallReq");
	    if (event->u.getAPICaps.escapeService) o_APIcap(comnt,"escapeService");
	    if (event->u.getAPICaps.privateStatusEvent) o_APIcap(comnt,"privateStatusEvent");
	    if (event->u.getAPICaps.escapeServiceEvent) o_APIcap(comnt,"escapeServiceEvent");
	    if (event->u.getAPICaps.escapeServiceConf) o_APIcap(comnt,"escapeServiceConf");
	    if (event->u.getAPICaps.sendPrivateEvent) o_APIcap(comnt,"sendPrivateEvent");
	    if (event->u.getAPICaps.sysStatReq) o_APIcap(comnt,"sysStatReq");
	    if (event->u.getAPICaps.sysStatStart) o_APIcap(comnt,"sysStatStart");
	    if (event->u.getAPICaps.sysStatStop) o_APIcap(comnt,"sysStatStop");
	    if (event->u.getAPICaps.changeSysStatFilter) o_APIcap(comnt,"changeSysStatFilter");
	    if (event->u.getAPICaps.sysStatReqEvent) o_APIcap(comnt,"sysStatReqEvent");
	    if (event->u.getAPICaps.sysStatReqConf) o_APIcap(comnt,"sysStatReqConf");
	    if (event->u.getAPICaps.sysStatEvent) o_APIcap(comnt,"sysStatEvent");
	    return;
	case CSTA_GET_DEVICE_LIST_CONF:
	    // @@@ event->u.getDeviceList;
	    o_lstring(comnt,0,FALSE);
	    o_lstring("@@@ CSTA_GET_DEVICE_LIST_CONF",indent,TRUE);
	    o_newline(0);
	    break;
	case CSTA_QUERY_CALL_MONITOR_CONF:
	    o_lstring(comnt,0,FALSE);
	    o_lstring("@@@ CSTA_QUERY_CALL_MONITOR_CONF",indent,TRUE);
	    o_newline(0);
	    break;
	default:
	    o_lstring(comnt,0,FALSE);
	    o_lprintf(indent,TRUE,"CSTAconfirmation ?%d?",eType);
	    o_newline(0);
	    break;
    };
    return;
}

/* Output an event header
**
** Inputs:
**	eHdr	event header to report
**	verbose	TRUE to print many line form
*/
EXTERNDEF void o_eventHeader(void *eHdr,BOOLEAN verbose)
{
    ACSEventHeader_t *hdr = (ACSEventHeader_t *)eHdr;
    if (hdr == NULL) return;
    if (verbose) {
	if (hdr->acsHandle != localACShandle) {
	    o_lstring("Handle:",0,FALSE);
	    o_lnumber(hdr->acsHandle,0,TRUE,10);
	};
    };
    if (verbose) o_lstring("Class:",0,TRUE);
    o_lstring(getClassName(hdr->eventClass),0,verbose);
    if (verbose) o_lstring("Type:",0,TRUE);
    o_lstring(getTypeName(hdr->eventClass,hdr->eventType),0,verbose);
    return;
}

/* Print an event
**
** Inputs:
**	evt	event to report
**	pvt	private data associated with the event
**	verbose	TRUE to print many line form
**	comnt	comment string with which to prefix each line
*/
EXTERNDEF void o_event(CSTAEvent_t *event,NetPhonePrivateMessage_t *pvt,
		       BOOLEAN verbose,char *comnt) {
    if (event == NULL) return;
    switch (event->eventHeader.eventClass) {
	case ACSREQUEST:
	    o_firstLineStart(comnt,verbose);
	    o_eventHeader(&event->eventHeader,verbose);
	    o_newline(0);
	    if (verbose) {
		o_lstring(comnt,0,FALSE);
		o_lstring("? Strange to find an ACSREQUEST from the server",
			  0,FALSE);
		o_newline(0);
	    };
	    break;
	case ACSUNSOLICITED:
	    o_firstLineStart(comnt,verbose);
	    o_ACSunsolicited(event->eventHeader.eventType,
			     &event->event.acsUnsolicited);
	    if (verbose) o_newline(0);
	    break;
	case ACSCONFIRMATION:
	    o_ACSconfirmation(event->eventHeader.eventType,
			      &event->event.acsConfirmation,comnt,verbose);
	    break;
	case CSTAREQUEST:
	    o_firstLineStart(comnt,verbose);
	    o_eventHeader(&event->eventHeader,verbose);
	    o_newline(0);
	    if (verbose) {
		o_lstring(comnt,0,FALSE);
		o_lstring("? Found a CSTAREQUEST from the server",0,FALSE);
		o_newline(0);
	    };
	    break;
	case CSTAUNSOLICITED:
	    o_CSTAunsolicited(event->eventHeader.eventType,
			      &event->event.cstaUnsolicited,pvt,comnt,verbose);
	    break;
	case CSTACONFIRMATION:
	    o_CSTAconfirmation(event->eventHeader.eventType,
			       &event->event.cstaConfirmation,pvt,
			       comnt,verbose);
	    break;
	case CSTAEVENTREPORT:
	    o_CSTAreport(event->eventHeader.eventType,
			 &event->event.cstaEventReport,comnt,verbose);
	    break;
	default:
	    o_firstLineStart(comnt,verbose);
	    o_eventHeader(&event->eventHeader,verbose);
	    o_newline(0);
	    break;
    };
}


/* This area maintains timestamps based on the invokeID when the stamp is
** valid the delay since the original invoke is reported with the message.
*/

typedef struct timeStamp {
    InvokeID_t invokeID;
    BOOLEAN valid;
    clock_t olderTime;
    clock_t lastTime;
} TimeStamp_t;

LOCALVAR TimeStamp_t invokeStamps[128];

/* Set timestamp associated with invokeID
**
** Inputs:
**	invokeID
** Returns:
**	Time since invokeID was last seen in seconds
*/
EXTERNVAR float stampInvokeID(InvokeID_t invokeID)
{
    clock_t now = clock();
    int index = (int)(invokeID % DIM(invokeStamps));
    if (invokeStamps[index].valid
	&& (invokeStamps[index].invokeID == invokeID)) {
	clock_t then;
	then = invokeStamps[index].lastTime;
	invokeStamps[index].olderTime = then;
	invokeStamps[index].lastTime = now;
	return  (now - then)/CLK_TCK;
    };
    invokeStamps[index].valid = TRUE;
    invokeStamps[index].invokeID = invokeID;
    invokeStamps[index].olderTime = now;
    invokeStamps[index].lastTime = now;
    return 0;
};

/* get timestamp associated with invokeID
**
** Inputs:
**	invokeID
** Returns:
**	Time since invokeID was last seen in seconds
*/
EXTERNVAR clock_t getInvokeIDstamp(InvokeID_t invokeID)
{
    int index = (int)(invokeID % DIM(invokeStamps));
    if (invokeStamps[index].valid
	&& (invokeStamps[index].invokeID == invokeID)) {
	return invokeStamps[index].lastTime;
    };
    return 0;
};

/* output the delta time between the start and last invokeID timestamp
**
** Inputs:
**	invokeID
**	column		column for first character
**	flag		TRUE if whitespace must preceed string
** Returns:
**	TRUE if time was present
*/
EXTERNDEF BOOLEAN o_invokeIDdelta(InvokeID_t invokeID,
				  int column,BOOLEAN flag)
{
    int index = (int)(invokeID % DIM(invokeStamps));
    if (invokeStamps[index].valid
	&& (invokeStamps[index].invokeID == invokeID)) {
	float delta;
	delta = invokeStamps[index].lastTime - invokeStamps[index].olderTime;
	delta /= CLK_TCK;
	if (invokeIDdelayOn) o_lfloat(delta,column,flag);
	return TRUE;
    };
    return FALSE;
}
