/* This module provides a call logging service.
**
** Copyright (c) by NetPhone Inc. 1997 All rights reserved.
**
** MAKEDOC:EMIT_TO_HEADER_ON
#define calllogEdit 3
** MAKEDOC:EMIT_TO_HEADER_OFF
**
** Modification History:
**  18Mar97 MJNewman 3: Use windows style output
**  17Mar97 MJNewman 2: Do not output call if no output file set..
**  14Feb97 MJNewman 1: Created.
*/

#include "calllog.h"
#include <ctype.h>
#include <cstring.h>

static TSet<CallInfo>activeCalls(16);

/*
 * Add WIN32 support
 */
#if defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
static HANDLE callLogFile;
#else
static HFILE callLogFile;
#endif

BOOLEAN CallInfo::reportUnansweredCalls = TRUE;

BOOLEAN CallInfo::outputCRenabled = FALSE;

time_t CallInfo::lastSnapshotTime = 0;

enum PBXtype CallInfo::currentPBXformat = PBXtypeMitelSX2000;

CallInfo::CallInfo(const long id) {
    callID = id;
    time(&creationTime);
    ringingTime = creationTime;
    answeringTime = 0;
    hangupTime = 0;
    callerSet = FALSE;
    callerNumber = string("?Unknown?");
    toSet = FALSE;
    toNumber = string("?Unknown?");
    trunkName = string("L1#1");		/* lie about the trunk name for now */
    callType = callTypeUnknown;
    dialedSet = FALSE;
    dialedNumber = string("?Unknown?");
    /*
     * initialization for new members
     */
    category = string("U");
    snapshotInitiated = FALSE;
    lastSeen = creationTime;
    lastHeld = 0;
    totalTimeHeld = 0;
    xferInfo = string("NONE");
    answerNumber = string("NONE");
    return;
}

/*
 * Add the assignment operator
 */
CallInfo& CallInfo::operator = (const CallInfo &r) {
    callID = r.callID;
    creationTime = r.creationTime;
    ringingTime = r.ringingTime;
    answeringTime = r.answeringTime;
    hangupTime = r.hangupTime;
    callerSet = r.callerSet;
    callerNumber = r.callerNumber;
    toSet = r.toSet;
    toNumber = r.toNumber;
    trunkName = r.trunkName;
    callType = r.callType;
    dialedSet = r.dialedSet;
    dialedNumber = r.dialedNumber;
    category = r.category;
    snapshotInitiated = r.snapshotInitiated;
    lastSeen = r.lastSeen;
    lastHeld = r.lastHeld;
    totalTimeHeld = r.totalTimeHeld;
    xferInfo = r.xferInfo;
    answerNumber = r.answerNumber;
    return (*this);
}

void CallInfo::o_reportNetPhone(int indent) const {
    const int datimeColumn = indent + 2;
/*
 * Add ringing time information
 */
    const int ringTimeColumn = datimeColumn + 18;
    const int typeColumn = ringTimeColumn + 6;
    const int trunkColumn = typeColumn + 2;
    const int elapsedColumn = trunkColumn + 10;
    const int callerColumn = elapsedColumn + 2;
    const int calleeColumn = callerColumn + 18;
/*
 * Add call hold time information
 */
    const int heldTimeColumn = calleeColumn + 22;
/*
 * Add answering number information
 */
    const int answerColumn = heldTimeColumn + 1;
/*
 * Add the transferred call information
 */
    const int originCallerColumn = answerColumn + 6;
    long int amountOfTime;
    if (!toSet) return;
    o_lstring(category.c_str(),indent,FALSE);
    if (answeringTime) {
	o_unixSlashTime(answeringTime,datimeColumn,TRUE);
	amountOfTime = (hangupTime > answeringTime) ? (hangupTime - answeringTime) : 0;
    } else {
	if (!reportUnansweredCalls) {
	    return;
	};
	o_unixSlashTime(creationTime,datimeColumn,TRUE);
	amountOfTime = 0;
    };
/*
 * Add ringing time information
 */
    long int timeToAnswer = ((answeringTime)? answeringTime : hangupTime) - ringingTime;
    o_ramount_of_time((timeToAnswer > 0)? timeToAnswer : 0, ringTimeColumn, TRUE);
    char typeCode;
    switch (callType) {
	case callTypeUnknown:
	    typeCode = 'U';
	    break;
	case callTypeInternal:
	    typeCode = 'I';
	    break;
	case callTypeExternal:
	    typeCode = 'E';
	    break;
	case callTypeExternalOutbound:
	    typeCode = 'O';
	    break;
    };
    o_char(typeCode,typeColumn,TRUE);
    o_lstring(((callType == callTypeInternal) ? toNumber.c_str() : trunkName.c_str()),
	      trunkColumn,TRUE);
    o_ramount_of_time(amountOfTime,elapsedColumn,TRUE);
    o_lstring(callerNumber.c_str(),callerColumn,TRUE);
    o_lstring(toNumber.c_str(),calleeColumn,TRUE);
/*
 * Add call hold time information
 */
    o_ramount_of_time(totalTimeHeld, heldTimeColumn, TRUE);
/*
 * Add number which answers the call
 */    
    if ( (answerNumber != toNumber) && (answerNumber != "NONE") )
	o_lstring(answerNumber.c_str(), answerColumn, TRUE);
/*
 * Add transfer information
 */    
    if (xferInfo != "NONE") o_lstring(xferInfo.c_str(), originCallerColumn, TRUE);
    if (outputCRenabled) o_char('\r',0,FALSE);
    o_newline(0);
    return;
}

void CallInfo::o_reportDefinityG3(int indent) const {
    long int amountOfTime;
    int hours;
    int minutes;
    int tenths;
    char string[30];
    struct tm *time_now;
    if (!toSet) return;
    if (answeringTime == 0) return;	/* unanswered call */
    amountOfTime = (hangupTime != 0) ? (hangupTime - answeringTime) : 0;
    time_now = localtime(&answeringTime);
    strftime(string,DIM(string),"%H%M",time_now);
    o_lstring(string,indent,FALSE);
    hours = (int)(amountOfTime / 3600);
    minutes = (int)((amountOfTime % 3600) / 60);
    tenths = (int)(amountOfTime % 60);
    tenths *= 10;
    tenths /= 60;			/* report in tenths */
    if (hours > 9) {
	hours = 9;
	minutes = 59;
	tenths = 9;
    };
    sprintf(string,"%01d%02d%01d",hours,minutes,tenths);
    o_lstring(string,indent+4,FALSE); 
    switch (callType) {
	case callTypeUnknown:
	case callTypeInternal:
	case callTypeExternal:
	    o_char('I',0,FALSE);
	    strncpy(string,normalizeNumber(callerNumber.c_str()).c_str(),
		    DIM(string));
	    o_lstring(string,27,FALSE);
	    strncpy(string,normalizeNumber(toNumber.c_str()).c_str(),DIM(string));
	    string[4] = NUL;
	    o_lstring(string,51,FALSE);
	    break;
	case callTypeExternalOutbound:
	    o_char('G',0,FALSE);
	    o_lstring(normalizeNumber(toNumber.c_str()).c_str(),19,FALSE);
	    strncpy(string,normalizeNumber(callerNumber.c_str()).c_str(),
		    DIM(string));
	    string[4] = NUL;
	    o_lstring(string,33,FALSE);	/* Called extn */
	    o_lstring("01",56,FALSE);	/* dummy trunk ID */
	    break;
    };
    if (outputCRenabled) o_char('\r',0,FALSE);
    o_newline(0);
    return;
};

void CallInfo::o_reportMitelSX2000(int indent) const {
    long int amountOfTime;
    int hours;
    int minutes;
    int seconds;
    char string[30];
    char codeChar;
    struct tm *time_now;
    if (!toSet) return;
    amountOfTime = (hangupTime != 0) ? (hangupTime - answeringTime) : 0;
    codeChar = ' ';
    if (amountOfTime > 5*60) {
	codeChar = '-';
	if (amountOfTime > 10*60) {
	    codeChar = '%';
	    if (amountOfTime > 30*60) {
		codeChar = '+';
	    };
	};
    };
    o_char(codeChar,indent,FALSE);
    time_now = localtime(&answeringTime);
    strftime(string,DIM(string),"%m/%d %H:%M",time_now);
    o_lstring(string,indent+2,FALSE);	/* date and time */
    hours = (int)(amountOfTime / 3600);
    minutes = (int)((amountOfTime % 3600) / 60);
    seconds = (int)(amountOfTime % 60);
    if (hours > 99) {
	hours = 99;
	minutes = 59;
	seconds = 9;
    };
    sprintf(string,"%02d:%02d:%02d",hours,minutes,seconds);
    o_lstring(string,indent+15,FALSE);	/* duration */
    strncpy(string,normalizeNumber(callerNumber.c_str()).c_str(),DIM(string));
    string[3] = NUL;
    o_lstring(string,indent+24,FALSE);	/* calling party should be trunkID */
#if 1
    /* example always has a 9 at 30 */
    o_char('9',indent+30,FALSE);
#else
    if (answeringTime == 0) {
	o_lstring("***",indent+30,FALSE); /* time to answer */
    } else {
	time_t timeToAnswer = answeringTime - creationTime;
	sprintf(string,"%03d",(int)(timeToAnswer/60));
	string[3] = NUL;
	o_lstring(string,indent+30,FALSE); /* time to answer */
    };
#endif
    strncpy(string,dialedNumber.c_str(),DIM(string));
    string[20] = NUL;
    o_lstring(string,indent+34,FALSE);	/* actual digits outdialed */
#if 0
    //Actual example had no pulses or type flag.
    sprintf(string,"%05d",(int)amountOfTime);
    string[5] = NUL;
    o_lstring(string,indent+55,FALSE);	/* meter pulses (seconds) */
    switch (callType) {			/* call completion status */
	case callTypeUnknown:
	    break;
	case callTypeInternal:
	    o_char('I',indent+60,FALSE);
	    break;
	case callTypeExternal:
	    break;
	case callTypeExternalOutbound:
	    o_char('A',indent+60,FALSE);
	    break;
    };
#endif
    o_char('R',indent+61,FALSE);	/* S = speed dial or F = forward flag
					** R = seen in example (most cases)
					** C = seen in example (a few cases)
					*/
    //???should be trunkID or extn number MITEL uses all 4 for extn
    o_char('T',indent+62,FALSE);	/* all cases had Tnnn in example */
    strncpy(string,toNumber.c_str(),DIM(string));
    string[3] = NUL;
    o_lstring(string,indent+63,FALSE);
    o_char(' ',indent+66,FALSE);	/* T = supervised transfer
					** X = unsupervised transfer
					** C = 3 way or conference
					** U = path unavailable
					** I = interflow
					*/
    o_lstring("",indent+68,FALSE);	/* Third party (for us the called
					** or alerting number after a transfer
					** occured.
					*/
#if 0
    /* example had none of these */
    o_lstring("000",indent+86,FALSE);	/* system identifier */
#endif
    if (outputCRenabled) o_char('\r',0,FALSE);
    o_newline(0);
    return;
};

void CallInfo::o_report(int indent) {
    if (callType == callTypeUnknown) {
	/* here we try to establish what is going on by guessing */
	if ((callerNumber.length() == 3) && (callerNumber[0] == '9')) {
	    /* trunk with missing caller ID */
	    callType = callTypeExternal;
	} else if (callerNumber.contains('#')) {
	    /* trunk with missing caller ID */
	    callType = callTypeExternal;
	} else if ((toNumber.length() < 5) && (callerNumber.length() < 5)) {
	    /* internal call */
	    callType = callTypeInternal;
	} else if ((toNumber.length() < 5) && (callerNumber.length() > 4)) {
	    callType = callTypeExternal;
	};
    };
    switch (currentPBXformat) {
	default:
	case PBXtypeNetPhone:
	    this->o_reportNetPhone(indent);
	    break;
	case PBXtypeDefinityG3:
	    this->o_reportDefinityG3(indent);
	    break;
	case PBXtypeMitelSX2000:
	    this->o_reportMitelSX2000(indent);
	    break;
    };
    return;
}

BOOLEAN CallInfo::initiated(const long callID,char *numberCalling) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    BOOLEAN wasPresent = TRUE;
    if (info == NULL) {
	wasPresent = FALSE;
	if (!activeCalls.Add(CallInfo(callID))) return FALSE;
	info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    };
    info->setCallerNumber(numberCalling);
/*
 * add lastSeen information
 */
    info->lastSeen = time(NULL);
    return wasPresent;
}

BOOLEAN CallInfo::originated(const long callID,const char *numberCalling,
			     const char *numberDialed) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    BOOLEAN wasPresent = TRUE;
    if (info == NULL) {
	wasPresent = FALSE;
	if (!activeCalls.Add(CallInfo(callID))) return FALSE;
	info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    };
    info->ringingTime = time(NULL);
    info->setCallerNumber(numberCalling);
    info->setToNumber(numberDialed);
    info->setDialed(numberDialed);
/*
 * add lastSeen information
 */
    info->lastSeen = info->ringingTime;
    return wasPresent;
}

BOOLEAN CallInfo::alerting(const long callID,const char *numberCalling,
			   const char *numberCalled, const char *numberAlerting) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    BOOLEAN wasPresent = TRUE;
    if (info == NULL) {
	wasPresent = FALSE;
	if (!activeCalls.Add(CallInfo(callID))) return FALSE;
	info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    };
    info->ringingTime = time(NULL);
    info->setCallerNumber(numberCalling);
    info->setToNumber(numberAlerting);
    info->setDialed(numberCalled);
/*
 * add lastSeen information
 */
    info->lastSeen = info->ringingTime;
    return wasPresent;
}

BOOLEAN CallInfo::answered(const long callID,const char *numberCalling,
			   const char *numberCalled, const char *numberAnswering) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    BOOLEAN wasPresent = TRUE;
    if (info == NULL) {
/*
 * workaround for firmware bug that a CSTA Established event is sent after the call was cleared for voicemail Record and
 * Playback operations
 */
	if ( !compare_name((char *)numberAnswering, "PLAYING", 7)
	     || !compare_name((char *)numberAnswering, "RECORDING", 9) ) {
	    return (FALSE);
	}
	wasPresent = FALSE;
	if (!activeCalls.Add(CallInfo(callID))) return FALSE;
	info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    };
    if (info->answeringTime < info->ringingTime) {
	info->answeringTime = time(NULL);
	if (info->category == "U") info->category = "C";
    }
/*
 * set the answering number
 */    
    info->setAnswerNumber(numberAnswering);    
    info->setCallerNumber(numberCalling);
    info->setToNumber(numberAnswering);
    info->setDialed(numberCalled);
/*
 * add lastSeen information
 */
    info->lastSeen = info->answeringTime;
    return wasPresent;
}

BOOLEAN CallInfo::netReached(const long callID,const char *numberCalling,
			     const char *numberDialed, const char *trunkUsed) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    BOOLEAN wasPresent = TRUE;
    if (info == NULL) {
	wasPresent = FALSE;
	if (!activeCalls.Add(CallInfo(callID))) return FALSE;
	info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    };
    info->category = "C";
    info->callType = callTypeExternalOutbound;
    info->ringingTime = time(NULL);
    info->answeringTime = info->ringingTime;
    info->setCallerNumber(numberCalling);
    info->setToNumber(numberDialed);
    info->setDialed(numberDialed);
/*
 * add trunk used information
 */
    if ( (trunkUsed != NULL) && (strlen(trunkUsed) != 0) ) {
	info->trunkName = string(trunkUsed);
    }
/*
 * add lastSeen information
 */
    info->lastSeen = info->answeringTime;
    return wasPresent;
}

void CallInfo::callSnapshoted(const long callID) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    if (info == NULL) return;	/* not found in active calls list */
    info->lastSeen = time(NULL);
    lastSnapshotTime = 0;	/* ready to send next snapshot call request */
    return;
}
    
#ifdef WINDOWS
#define FILE_TYPE HFILE
#define SET_O_FILE set_windows_file
/*
 * Add WIN32 support
 */
#elif defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
#define FILE_TYPE HANDLE
#define SET_O_FILE set_win32_file
#else
#define FILE_TYPE FILE *
#define SET_O_FILE set_o_file
#endif
/*
 * Returns:
 *	FALSE if need more polling
 *	TRUE when all active calls have been  polled
 */
BOOLEAN CallInfo::pollNextSnapshot(void) {
    if (activeCalls.IsEmpty()) return (TRUE);
    time_t now = time(NULL);
    if ((now - lastSnapshotTime) < POLL_INTERVAL) return (FALSE);
    TSetIterator<CallInfo> Iter(activeCalls);
    while ( Iter != 0 ) {
	// CallInfo *info = (CallInfo *) activeCalls.Find(CallInfo(Iter.Current().callID));
	CallInfo *info = (CallInfo *) activeCalls.Find(Iter.Current());
	if (info == NULL) {	/* this should never happen */
	    return (TRUE);
	}
	if (!info->snapshotInitiated) {
	    info->snapshotInitiated = TRUE;
	    lastSnapshotTime = now;
	    doSnapshotActiveCall(info->callID);
	    return (FALSE);			/* poll the next call later */
	}
	if ((now - info->lastSeen) > TEN_MINUTES) {
	    log_error("?Logging an inactive call, callID=%lX callerNumber=%s toNumber=%s\n",
		      info->callID, info->callerNumber.c_str(), info->toNumber.c_str());
	    info->hangupTime = info->lastSeen;	/* mark the hangup time as time last seen */
/*
 * Add WIN32 support
 */
#if defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
	    if (callLogFile != INVALID_HANDLE_VALUE) {
#else		    
	    if (callLogFile != NULL) {
#endif		
		FILE_TYPE currentFile = SET_O_FILE(callLogFile);
		BOOLEAN oldTabs = set_tabs_allowed(FALSE);
		info->o_report(0);
		set_tabs_allowed(oldTabs);
		SET_O_FILE(currentFile);
	    };
	    CallInfo deadCall(info->callID);
	    activeCalls.Detach(deadCall);
	    return (FALSE);			/* get a fresh iterator next time ??? */
	}
	Iter++;
    }
    return (TRUE);				/* no more polling needed */
}

BOOLEAN CallInfo::ended(const long callID, BOOLEAN report) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    //@@@It seems there is a terrible efficiency problem here
    //	in that the CallInfo is constructed two extra times
    //	once to find it once to detach it. It need not be
    //  created at all. It should take the CALL_ID as a
    //  argument (or promote a long) and then take the pointer
    //	to the object to detach. Internal to the class it can
    //	remember the most reecent callID and have a very fast
    //	lookup.
    if (info != NULL) {
	FILE_TYPE currentFile;
	info->hangupTime = time(NULL);
/*
 * this is to take care of the problem that when PhoneMaster user accesses VoiceMail, we don't receive a private
 * broadcast message which means that VoiceMail will take over the call.  So in this case in order to flag the call as
 * VoiceMail call, we have to check the caller and called numbers and try to guess it
 */
	if ( (stricmp(info->toNumber.c_str(), "Playing") == 0 ) ||
	     (stricmp(info->toNumber.c_str(), "Recording") == 0) ) {
	    if (info->category != "X") info->category = "M";
	}
	if (info->lastHeld) info->totalTimeHeld += info->hangupTime - info->lastHeld;
/*
 * Add WIN32 support
 */
#if defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
	if (report && (callLogFile != INVALID_HANDLE_VALUE)) {
#else	 	
	if (report && (callLogFile != NULL)) {
#endif	    
	    BOOLEAN oldTabs;
	    currentFile = SET_O_FILE(callLogFile);
	    oldTabs = set_tabs_allowed(FALSE);
	    info->o_report(0);
	    set_tabs_allowed(oldTabs);
	    SET_O_FILE(currentFile);
	};
	activeCalls.Detach(CallInfo(callID));
	return TRUE;
    } else {
	return FALSE;
    };
}

FILE_TYPE CallInfo::setOutputFile(FILE_TYPE oFile) {
    FILE_TYPE temp;
    temp = callLogFile;
    callLogFile = oFile;
    return temp;
}
#undef FILE_TYPE
#undef SET_O_FILE

BOOLEAN CallInfo::setOutputCR(BOOLEAN outputCR) {
    BOOLEAN temp;
    temp = outputCRenabled;
    outputCRenabled = outputCR;
    return temp;
}

PBXtype CallInfo::setSMDRformat(PBXtype format) {
    PBXtype temp;
    temp = currentPBXformat;
    currentPBXformat = format;
    return temp;
}

void CallInfo::setCallerNumber(const char *number)
{
    if ( (number == NULL) || (strlen(number) == 0) ) return;
    if (NPOS != callerNumber.find_first_not_of("0123456789")) {
	callerSet = FALSE;
    };
    if (!callerSet) {
	callerSet = TRUE;
	callerNumber = string(number);
    };
    return;
}

/* Normalize a phone number. A 7 digit number gets the local area code
** added. An 1+10 digit number gets the leading 1 stripped off.
**
** Inputs:
**	number
** Returns:
**	normalized number, valid untill the next call to this routine
*/;
string CallInfo::normalizeNumber(const char *number)
{
    string normalized;
    if (number == NULL) return normalized;
    if (!isdigit(number[0])) {
	normalized = string(number);
	return normalized;
    };	
    int length = strlen(number);
    switch (length) {
	case 7:
	    normalized = localAreaCode + number;
	    break;
	case 11:
	    if (number[0] == '1') {
		normalized = string(&number[1]);
		break;
	    };
	default:
	    normalized = string(number);
	    break;
    };
    return normalized;
}


void CallInfo::setToNumber(const char *number)
{
    if ( (number == NULL) || (strlen(number) == 0) ) return;
    if (NPOS != toNumber.find_first_not_of("0123456789")) {
	toSet = FALSE;
    };
    if (!toSet) {
/*
 * if a call was directed to VoiceMail, we will keep getting CSTA Established event with different answeringDevice strings,
 * such as "Playing", "Recording", but sometimes the answering number is the same as the caller number, in this case, just
 * ignore it instead of logging it as a call from "callerNumber" to "callerNumber"
 */
	string newNumber = string(number);
	if (newNumber == callerNumber) return;
	toSet = TRUE;
	toNumber = newNumber;
    };
    return;
}

void CallInfo::setDialed(const char *number)
{
    if ( (number == NULL) || (strlen(number) == 0) ) return;
    if (NPOS != dialedNumber.find_first_not_of("0123456789")) {
	dialedSet = FALSE;
    };
    if (!dialedSet) {
	dialedSet = TRUE;
	dialedNumber = string(number);
    };
    return;
}

void CallInfo::setAnswerNumber(const char *number)
{
    if ( (number == NULL) || (strlen(number) == 0) ) return;
    string newNumber = string(number);
    if (NPOS != newNumber.find_first_not_of("0123456789")) {	/* e.g., "Playing" or "Recording" */
	if (NPOS == toNumber.find_first_not_of("0123456789"))
	    answerNumber = toNumber;
    } else {
	answerNumber = newNumber;
    }
    if (answerNumber == callerNumber) answerNumber = string("NONE");
    return;
}


string CallInfo::localAreaCode = string("999");

void CallInfo::setLocalAreaCode(const char *number)
{
    if (number != NULL) {
	localAreaCode = normalizeNumber(number);
    };
    return;
}

void CallInfo::initializeSnapshot(void) {
    if (activeCalls.IsEmpty()) return;
    TSetIterator<CallInfo> Iter(activeCalls);
    while ( Iter != 0 ) {
	// CallInfo *info = (CallInfo *) activeCalls.Find(CallInfo(Iter.Current().callID));
	CallInfo *info = (CallInfo *) activeCalls.Find(Iter.Current());
	if (info != NULL) {
	    info->snapshotInitiated = FALSE;
	}
	Iter++;
    }
    lastSnapshotTime = 0;
}

void CallInfo::held(const long callID) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    if (info != NULL) info->lastHeld = time(NULL);
    return;
}

void CallInfo::retrieved(const long callID) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    if ( (info != NULL) && info->lastHeld ) {
	info->totalTimeHeld += time(NULL) - info->lastHeld;
	info->lastHeld = 0;
    }
    return;
}

void CallInfo::transferred(const long newCallId, const long primaryId, const long secondaryId)
{
    CallInfo *primaryOldCall = (CallInfo *) activeCalls.Find(CallInfo(primaryId));
    CallInfo *secondaryOldCall = (CallInfo *) activeCalls.Find(CallInfo(secondaryId));
    if ( (primaryOldCall != NULL) && (secondaryOldCall != NULL) ) {
	CallInfo *newCall;
	primaryOldCall->category = "X";
	primaryOldCall->xferInfo = secondaryOldCall->toNumber;
	string originalCaller = primaryOldCall->callerNumber;
	if (originalCaller == secondaryOldCall->callerNumber) {
	    originalCaller = primaryOldCall->toNumber;
	}
	if (newCallId == primaryId) {
	    /* primaryOldCall->xferInfo is already setup */
	}
	else if (newCallId == secondaryId) {
	    newCall = secondaryOldCall;
	    newCall->category = "X";
	    newCall->xferInfo = originalCaller;
	} else {
	    if (!activeCalls.Add(CallInfo(newCallId))) return;		/* add the new call */
	    newCall = (CallInfo *)activeCalls.Find(CallInfo(newCallId));
	    if (newCall != NULL) {
		newCall->setCallerNumber(secondaryOldCall->callerNumber.c_str());
		newCall->setToNumber(secondaryOldCall->toNumber.c_str());
		newCall->category = "X";
		newCall->ringingTime = newCall->answeringTime = time(NULL);
		newCall->xferInfo = originalCaller;
	    }
	}
    }
    return;
}

void CallInfo::conferenced(const long newCallId, const long primaryId, const long secondaryId)
{
    CallInfo *primaryOldCall = (CallInfo *) activeCalls.Find(CallInfo(primaryId));
    CallInfo *secondaryOldCall = (CallInfo *) activeCalls.Find(CallInfo(secondaryId));
    if ( (primaryOldCall != NULL) && (secondaryOldCall != NULL) ) {
	primaryOldCall->category = "F";
	secondaryOldCall->category = "F";
	if ( (newCallId != primaryId) && (newCallId != secondaryId) ) {	/* add the new call */
	    if (!activeCalls.Add(CallInfo(newCallId))) return;
	    CallInfo *newCall = (CallInfo *)activeCalls.Find(CallInfo(newCallId));
	    if (newCall != NULL) {
		newCall->setCallerNumber(secondaryOldCall->callerNumber.c_str());
		newCall->setToNumber(secondaryOldCall->toNumber.c_str());
		newCall->category = "F";
		newCall->ringingTime = newCall->answeringTime = time(NULL);
	    }
	}
    }
    return;
}

BOOLEAN CallInfo::diverted(const long callID, const char *newDestination) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    BOOLEAN wasPresent = TRUE;
    if (info == NULL) {
	wasPresent = FALSE;
	if (!activeCalls.Add(CallInfo(callID))) return FALSE;
	info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    };
    if (newDestination != NULL)
    {
	info->setDialed(info->toNumber.c_str());
	info->setToNumber(newDestination);
    }
/*
 * add lastSeen information
 */
    info->lastSeen = time(NULL);
    return wasPresent;
}

BOOLEAN CallInfo::voicemailCall(const long callID) {
    CallInfo *info = (CallInfo *)activeCalls.Find(CallInfo(callID));
    BOOLEAN wasPresent = FALSE;
    if (info != NULL) {
	wasPresent = TRUE;
	if ( (info->category == "U") || (info->category == "C") )
	    info->category = "M";
/*
 * add lastSeen information
 */	
	info->lastSeen = time(NULL);
    };
    return wasPresent;
}