/* This module provides a call logging service.
**
** Copyright (c) by NetPhone Inc. 1997 All rights reserved.
**
**
** Modification History:
**  18Mar97 MJNewman 2: Windows style file output.
**   7Feb97 MJNewman 1: Created.
*/

#ifndef __calllog_H__
#define __calllog_H__

#define calllogHedit 2

#ifdef WINDOWS
#include <windows.h>
#endif

#include <cstring.h>
#include <time.h>
#include <cstring.h>
#include <classlib\sets.h>

#include "basetype.h"

#include "iosubs.h"
#include "iosubs.xh"
#include "acs.h"
#include "csta.h"
#include "query.xh"

enum PBXtype {
    PBXtypeUnknown,
    PBXtypeNetPhone,
    PBXtypeDefinityG3,
    PBXtypeMitelSX2000,
};

class CallInfo {
    private:
	enum { POLL_INTERVAL = 5 /* seconds */};
	enum { TEN_MINUTES = 600 /* seconds */};
/*
 * new members for checking active calls
 */
	BOOLEAN snapshotInitiated;
	time_t lastSeen;
	static time_t lastSnapshotTime;
/*
 * add call hold time information
 */
	time_t lastHeld;
	long totalTimeHeld;
/*
 * add call category string, i.e., "$CALL", "UCALL", "XCALL", "FCALL", "VCALL", etc
 */
	string category;
	static BOOLEAN outputCRenabled;
	static BOOLEAN reportUnansweredCalls;
	static PBXtype currentPBXformat;
	static string localAreaCode;
	long callID;
	time_t creationTime;
	time_t ringingTime;
	time_t answeringTime;
	time_t hangupTime;
	enum {
	    callTypeUnknown,
	    callTypeInternal,
	    callTypeExternal,
	    callTypeExternalOutbound,
	} callType;
	BOOLEAN callerSet;		/* TRUE when caller is set */
	string callerNumber;
	BOOLEAN toSet;			/* TRUE when toNumber is set */
	string toNumber;
	string trunkName;
	BOOLEAN dialedSet;
	string dialedNumber;		/* fetch the real dialed digits! */
/*
 * add number which answers the call
 */
	string answerNumber;
/*
 * add call transfer information
 */
	string xferInfo;		/* extra information about transferred call */

	/* Output a one line report about the call which tells a logging program
	 ** everything there is to know about this call. The iosubs subroutines are used
	 ** to emit the report.
	 **
	 ** When the PBX is NETphone a line of one of the following forms is generated:
	 **
	 ** If callType is unknown defines it by taking an educated guess.
	 **
	 **			start
	 **			time	duration	caller		called
	 **
	 **    $CALL,	13Jan97 9:03:07, 5:01,		16175551212,	203
	 **    $CALLU,	13Jan97 9:03:07, 5:01,		16175551212,	203
	 **
	 ** Inputs:
	 **	indent		indentation for each line
	 */
	void o_report(int indent = 0);

    public:
	CallInfo(const long id = 0);
/*
 * add the assignment operator
 */
	CallInfo& CallInfo::operator = (const CallInfo &);

	int operator== (const CallInfo& call) const {
	    return (callID == call.callID);
	};

	/* Record the fact that the call was started (dial tone)
	 **
	 ** Inputs:
	 **	callID		TSAPI call ID
	 **	numberCalling
	 ** Returns:
	 **	TRUE if call was present
	 */
	static BOOLEAN initiated(const long callID,char *numberCalling);


	/* Record the fact that the call started ringing outbound.
	 **
	 ** Inputs:
	 **	callID		TSAPI call ID
	 **	numberCalling	number placing the call
	 **	numberDialed	number being called
	 ** Returns:
	 **	TRUE if call was present
	 */
	static BOOLEAN originated(const long callID,
				  const char *numberCalling,
				  const char *numberDialed);
			       

	/* Record the fact that the call started ringing at the destination
	 **
	 ** Inputs:
	 **	callID		TSAPI call ID
	 **	numberCalling
	 **	numberRinging
	 ** Returns:
	 **	TRUE if call was present
	 */
	static BOOLEAN alerting(const long callID,const char *numberCalling,
				const char *numberCalled, const char *numberAlerting);
				

	/* Record the fact that the call was answered
	 **
	 ** Inputs:
	 **	callID		TSAPI call ID
	 **	numberCalling
	 **	numberAnswering
	 ** Returns:
	 **	TRUE if call was present
	 */
	static BOOLEAN answered(const long callID,const char *numberCalling, const char*numberCalled,
				const char *numberAnswering);
				

	/* Record the fact that the call reached the public network
	 **
	 ** Inputs:
	 **	callID		TSAPI call ID
	 **	numberCalling
	 **	numberDialed
	 **	trunkUsed
	 ** Returns:
	 **	TRUE if call was present
	 */
	static BOOLEAN netReached(const long callID,const char *numberCalling,
				  const char *numberDialed, const char *trunkUsed);


	/* When a call ends record the times and flush it from the call list
	 **
	 ** Inputs:
	 **	callID:		TSAPI call ID
	 **	report:		flag indicating  whether to report this call(TRUE by default)
	 ** Returns:
	 **	TRUE if call was present
	 */
	static BOOLEAN ended(const long callID, BOOLEAN report = TRUE);

	
	/* When a snapshot call request is confirmed, mark the corresponding call as normal
	 **
	 ** Inputs:
	 **	callID		TSAPI call ID
	 */
	static void callSnapshoted(const long callID);
	static void initializeSnapshot(void);
	static BOOLEAN pollNextSnapshot(void);

	/*
	 * When a Call Held event or Call Retrieved event is received, get the call hold time information
	 */
	static void held(const long);
	static void retrieved(const long);

	/*
	 * Handle Transferred and Conferenced events
	 */
	static void transferred(const long, const long, const long);
	static void conferenced(const long, const long, const long);

	/*
	 * Add new function for CSTA_DIVERTED event instead of using the same handler for CSTA_DELIVERED event
	 */
	static BOOLEAN diverted(const long, const char *);

	/*
	 * Mark a call as voicemail call, i.e., "$VCALL"
	 */
	static BOOLEAN voicemailCall(const long);
	
	/* Direct call log output to a file
	**
	** Inputs:
	**	ofile	new output file to use
	** Returns:
	**	previous output file
	*/
#ifdef WINDOWS
	static HFILE setOutputFile(HFILE oFile);
/*
 * Add WIN32 support
 */
#elif defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
	static HANDLE setOutputFile(HANDLE oFile);	
#else
	static FILE *setOutputFile(FILE *oFile);
#endif
	
	/* When set to TRUE a CR character is added prior to each newline
	** Typically used when output is sent to a COM port.
	**
	** Inputs:
	**	outputCR	TRUE to output <CR> on each line
	*/
	static BOOLEAN setOutputCR(BOOLEAN outputCR);

	/* Tell the logging system which output format to use
	**
	** Inputs:
	**	format		type of report to generate
	** Returns:
	**	output format previously in use
	*/
	static PBXtype setSMDRformat(PBXtype format);

	/* Tell the logging system the local area code which is used
	** when a 7 digit number is dialed outbound.
	**
	** Inputs:
	**	areaCode
	*/
	static void setLocalAreaCode(const char *areaCode);

    private:
/*
 * add setAnswerNumber()
 */
	void setAnswerNumber(const char *);
	void setCallerNumber(const char *number);
	void setToNumber(const char *number);
	void setDialed(const char *number);
	static string normalizeNumber(const char *number);
	void o_reportNetPhone(int indent) const;
	void o_reportDefinityG3(int indent) const;
	void o_reportMitelSX2000(int indent) const;

};

#endif /* __calllog_H__ */
