/* This module provides formatted file output services and command parsing
** services. The public interface routines are flagged with EXTERNDEF
** declarations. Public interface variables are flagged with EXTERNVAR.
**
** Copyright (c) by Michael Newman 1997 All rights reserved.
** Copyright (c) by Michael Newman 1996 All rights reserved.
** Copyright (c) by Michael Newman 1995 All rights reserved.
** Copyright (c) by Michael Newman 1994 All rights reserved.
** Copyright (c) by Michael Newman 1993 All rights reserved.
**
** MAKEDOC:EMIT_TO_HEADER_ON
#define iosubsEdit 23
** MAKEDOC:EMIT_TO_HEADER_OFF
**
** Modification History:
**   9May97 MJNewman 23: Handle NULL filenames.
**  13Mar97 MJNewman 21: Allow COMx: as filenames.
**  26Feb97 MJNewman 21: Allow COMx: as filenames.
**  13Feb97 MJNewman 19: add get_boolean
**  12Feb97 MJNewman 18: reformat time stamp.
**  12Feb97 MJNewman 18: Remove edit global symbol.
**  10Feb97 VABerry  17: Fixed problem where lists could run into each other.
**   7Feb97 MJNewman 16: Reformat dates output to ddMMMyy.
**		Change some arguments to const char * @@@need to do all.
**   6Feb97 VABerry  16: Added support for Label and goto in indirect files.
**		Added support for Q to quit, and P to pause indirect files.
**  22Jan97 VABerry  15: Added support for Emacs Style Command History
**  20Dec96 VABerry  14: created o_print_list function to print arrays of
		strings that are column sorted. get_list now uses this.
**  18Dec96 VABerry  13: Made list output (get_list) three columns.
**   6Jun96 MJNewman 12: Add logging output to both output and log file.
**   8May96 MJNewman 11: Add getting a number in a range.
**   5May96 MJNewman 10: Include iosubs.h. Add call when idle service.
**  13Apr95 MJNewman:	add control variables that were in other modules.
**  20Sep94 MJNewman:	Ensure that all errfree calls cast to (VOID *) for
**		DG platform where char * pointers can be different.
**   2Aug94 MJNewman:	Data type of set_max_malloc. Data type of max_malloc
**		becomes unsigned. Fix missing restore of old max value in
**		alloc_file case. Limit error printouts. Allow errmalloc to
**		return in no memory case for later failures. Add log_info.
**  25Jan94 MJNewman:	Fail if user input failure.
**   3Jan94 MJNewman:	Fix loss of significant digits warnings.
**   7Nov93 MJNewman:	Extend i_makeField to split at any of a set of chars.
**		Add mechanism to parse multiple fields on a line.
**   7Nov93 MJNewman:	Add get from indirect file mechanism.
**   8Sep93 MJNewman:	Add private DG to IEEE conversions.
**  10Aug93 MJNewman:	Add o_unixTime.
**   8Jul93 MJNewman:	Move include of time.h to central includes.
**  15Jun93 MJNewman:	Give MSDOS, UNIX and AOSVS default values.
**   8Apr93 MJNewman:	Add edit number.
**  24Feb93 MJNewman:	Use AOSVS to ID DG environment.
**  22Feb93 MJNewman:	On AOSVS allow file of -1 to put to tcpmsg(0,0,string)
**   2Feb93 MJNewman:	Comments on o_time
**   5Jan93 MJNewman:	Created from GASSER file services.
*/

#include <windows.h>

#include "basetype.h"
#include "iosubs.h"
#include "iosubs.xh"
#include "gets.h"
#include "gets.xh"

#include <time.h>

#if (!ANSI_C)
#include <varargs.h>
#endif

#if MSDOS
#include <conio.h>
#endif

/* TRUE to produce detailed debugging messages */
EXTERNVAR BOOLEAN debugPrint = FALSE;
EXTERNVAR BOOLEAN debugDetailed = FALSE;

/* after this many errors are reported the error output facility shuts up */
EXTERNVAR int maxErrorPrintouts = 200;

/* Stuff needed for labels */
EXTERNVAR LABELLIST labels;

/* turns on subroutines that prompt with printf and get things from user
** typein.
*/
#ifndef GET_USER_ENABLED
#define GET_USER_ENABLED 1
#endif

static char cur_line[STRING_LENGTH+1];		/* current output line */
static long int cur_lcount = 0;

LOCALVAR FILE *cur_file; /* current output file, NULL means the bit bucket */

#ifdef WINDOWS
/* When not HFILE_ERROR this is output destination takes the place of
** the cur_file destination. Nothing is output to cur_file
** while this happens.
*/
LOCALVAR HFILE windowsFile = HFILE_ERROR;
/*
 * Add WIN32 support
 */
#elif defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
LOCALVAR HANDLE win32File = INVALID_HANDLE_VALUE;
#endif

/* when a log file is enabled is the file pointer and name
*/
LOCALVAR struct {
    char *name;
    FILE *file;
} logFile = {NULL,NULL};
LOCALVAR BOOLEAN logOutputAlso = FALSE;

/* Control the combination of output to the current file and output to the
** log file. When TRUE the current output is delivered to the log file as well
** as the output file.
**
** Inputs:
**	outputAlso  TRUE to log output as well as commands
** Returns:
**	old state of this flag
*/
EXTERNDEF BOOLEAN io_setLogOutputAlso(BOOLEAN outputAlso)
{
    BOOLEAN temp = logOutputAlso;
    logOutputAlso = outputAlso;
    return temp;
}

/* Open a new log file. The current log file is closed and a new one is opened.
** Can be called with a NULL name to simply close the current log file.
** The log file is opened for append access, old data is not destroyed.
** The caller is responsible for closing the log file upon program termination.
**
** Inputs:
**	name	File name to open
** Returns:
**	TRUE if file is now being used for a log
**	FALSE if file can not be opened
*/
EXTERNDEF BOOLEAN establishLogFile(char *name)
{
    if (logFile.file != NULL) {
	fclose(logFile.file);
	logFile.file = NULL;
    };
    if (logFile.name != NULL) {
	printf("Closing log file %s\n",logFile.name);
	errfree((VOID *)logFile.name);
	logFile.name = NULL;
    };
    if (name == NULL) return FALSE;
    logFile.file = fopen(name,"a");
    if (logFile.file == NULL) return FALSE;
    logFile.name = alloc_str(name);
    printf("Logging commands in %s\n",logFile.name);
    return TRUE;
}

/* the input subroutines are capable of reading from indirect files. This list
** is the current set of open files.
*/
EXTERNVAR IFILES indirectFiles[10];
EXTERNVAR int nextIndirectFile = 0;		/* none open yet */

/* Adds a new indirect file to the stack of open indirect files. Input then
** comes from this file untill the file ends. It is the callers responsibility
** call the closeAllIndirectFiles() function upon program exit or abnormal
** termination of an indirect file. Also available is the
** closeCurrentIndirectFile() function to terminate the current indirect file
** and continue to read from any otheres.
**
** Inputs:
**	name	File name to open
** Returns:
**	TRUE if file is now being read for commands
**	FALSE if file can not be opened
*/
EXTERNDEF BOOLEAN addIndirectFile(char *name)
{
    if (name == NULL) return FALSE;
    if (nextIndirectFile >= DIM(indirectFiles)) {
	log_error("? Too many indirect files open at once. Trying to open %s\n",
		  name);
	return FALSE;
    };
    indirectFiles[nextIndirectFile].inputFile = fopen(name,"r");
    if (indirectFiles[nextIndirectFile].inputFile == NULL) {
	log_error("? Unable to open indirect file %s\n",name);
	return FALSE;
    };
    indirectFiles[nextIndirectFile].name = alloc_str(name);
    nextIndirectFile += 1;
    /* If nextIndirectFile goes to 1 we need to clean out the label struct */
    if (nextIndirectFile == 1) labels.count = 0;
    return TRUE;
}

/* Closes any outstanding indirect files.
*/
EXTERNDEF void closeAllIndirectFiles(void)
{
    int i;
    if (nextIndirectFile <= 0) {
	nextIndirectFile = 0;
	return;
    };
    for (i = 0;i < nextIndirectFile;i += 1) {
	fclose(indirectFiles[i].inputFile);
	errfree((VOID *)indirectFiles[i].name);
    };
    nextIndirectFile = 0;
    return;
}

/* Closes the current indirect file. Later user input will come from
** other open indirect files, if any, the console if not.
*/
EXTERNDEF void closeCurrentIndirectFile(void)
{
    if (nextIndirectFile <= 0) {
	nextIndirectFile = 0;
	return;
    };
    fclose(indirectFiles[nextIndirectFile-1].inputFile);
    errfree((VOID *)indirectFiles[nextIndirectFile-1].name);
    nextIndirectFile -= 1;
    return;
}

/* The o_xxx subroutines can use a mix of tabs and spaces or only spaces
** to get to the required column. When this is TRUE tabs can be used
** FALSE only spaces will be used. The space only feature is be some
** files delineated fields by tabs.
*/
LOCALVAR BOOLEAN tabs_allowed = TRUE;

/* control if filling to a column allows tabs. The default on opening
** a file is to allow tabs. When tabs are used in the file to delineate
** fields set this to FALSE.
**
** Inputs:
**	allowed		TRUE to allow tabs
** Returns:
**	previous state of flag
*/
EXTERNDEF BOOLEAN set_tabs_allowed(BOOLEAN allowed)
{
   BOOLEAN temp;

   temp = tabs_allowed;
   tabs_allowed = allowed;
   return temp;
}

 
/* function to call when waiting for a character
*/
LOCALVAR pFVOID callWhileIdle = NULL;

/* set function to call periodically while waiting for an input character.
**
** Inputs:
**	callee	    function to call
** Returns:
**	old function to call when idle
*/
EXTERNDEF pFVOID setCallWhileIdle(pFVOID callee)
{
    pFVOID temp;
    temp = callWhileIdle;
    callWhileIdle = callee;
    return temp;
}

/* GENERAL SUBROUTINES */

/* Months of the year */
EXTERNVAR char *months[] = {
    "January","February","March","April",
    "May","June","July","August",
    "September","October","November","December",
    NULL
};
EXTERNVAR char *ucmonths[] = {
    "JANUARY","FEBRUARY","MARCH","APRIL",
    "MAY","JUNE","JULY","AUGUST",
    "SEPTEMBER","OCTOBER","NOVEMBER","DECEMBER",
    NULL
};

static long int blks_out = 0;			/* count of mallocs */
static long int blks_in = 0;			/* count of freed */

#define MAX_BASIC_MALLOC 500
LOCALVAR unsigned int max_malloc = MAX_BASIC_MALLOC;

/* set the max size of a malloc before bug report
**
** Inputs:
**	size	new max number of bytes
** Returns:
**	previous setting of max malloc
*/
EXTERNDEF unsigned int set_max_malloc(unsigned size)
{
    unsigned int temp;
    temp = max_malloc;
    max_malloc = size;
    return temp;
}

/* If this module has a fatal error it will call this routine */
LOCALVAR pFVOID fatalCallback = NULL;

/* set a routine to call if a fatal error occurs
**
** Inputs:
**	callback    address of routine to call
** Returns:
**	previous callback
*/
EXTERNDEF pFVOID io_setFatalCallback(pFVOID callback)
{
    pFVOID temp;

    temp = fatalCallback;
    fatalCallback = callback;
    return temp;
}


/* Substitutes for malloc but prints an error if no space.
 */
EXTERNDEF char *errmalloc(unsigned size)
{
    char *i;
    unsigned long int j;

    j = size;
    if (j > max_malloc) {
	if (j > MAX_BASIC_MALLOC) {
	    log_error("? Allocating %ld bytes, probably a bug.\n",j);
	} else {
	    static BOOLEAN smallReport = FALSE;
	    if (!smallReport) {
		log_error("? Allocating %ld bytes probably a bug calculating block size\n",j);
		smallReport = TRUE;	/* only report this once */
	    };
	};
    };
    i = (char *)malloc(size);
    if (i == NULL) {
	log_error("? Out of memory allocating %d bytes. Unable to continue.\n",
		  size);
	cur_file = stdout;
	o_space();
	if (fatalCallback != NULL) {
	    (fatalCallback)();
	};
	exit(0);
    };
    blks_out += 1;
    return i;
}

/* Substitutes for free, prints an error if an error occured
 */
EXTERNDEF int errfree(VOID *block)
{
    int i;

    if (block == NULL) return 0;
    i = 0;
    free(block);
#if 0
    i = free(block);
    if (i != 0) {
	log_error("? free() of memory failed. The program is likely broken\n");
    };
#endif
    blks_in += 1;
    return i;
}

/* report number of free storage blocks outstanding
**
** Returns:
**	number of blocks outstanding
*/
EXTERNDEF long int number_blocks_out(void)
{
    return blks_out - blks_in;
}

/* Print a summary of blocks that have been allocated and freed to the current
 * file.
 */
EXTERNDEF void o_space(void)
{
#define BLOCK_COLUMN 25
    o_lstring("Blocks allocated",0,FALSE);
    o_rnumber(blks_out,BLOCK_COLUMN,TRUE,10);
    o_newline(0);
    o_lstring("Blocks freed",0,FALSE);
    o_rnumber(blks_in,BLOCK_COLUMN,TRUE,10);
    o_newline(0);
    o_lstring("Net blocks out",0,FALSE);
    o_rnumber(blks_out-blks_in,BLOCK_COLUMN,TRUE,10);
    o_newline(0);
    return;
    }

/* Close a file printing a warning if there is trouble
 *
 * Inputs:
 *	file		file to close
 *	filename	name of file being closed
 */
EXTERNDEF void file_close(FILE *file,char *filename)
{
    if (EOF == fclose(file)) {
	log_error("? Unable to close file \"%s\"\n",filename);
	};
    return;
    }

/* Not implemented yet
 */
EXTERNDEF void notyet(char *name)
{
    log_error("? Not implemented yet: \"%s\"\n",name);
    return;
}

/* Append one file to another. (Can not use "a" from C library because
** some DOS editors leave a spurious ^Z in the file and appending to those
** files gives a file that has the data but with data that can not be seen
** after the ^Z.)
**
** Inputs:
**	first		file appended to
**	second		file put on at the end of above
*/
EXTERNDEF void append(char *first,char *second)
{
    FILE *ofile;

#define APPEND_FILE  "APPEND.TMP"
    ofile = fopen(APPEND_FILE,"w");
    if (ofile == NULL) {
	log_error("? Unable to open \"%s\" for append.\n",APPEND_FILE);
	return;
	};
    copyfile(first,ofile);
    copyfile(second,ofile);
    file_close(ofile,APPEND_FILE);
    ofile = fopen(first,"w");
    if (ofile == NULL) {
	log_error("? Unable to open \"%s\" for append.\n",first);
	return;
	};
    copyfile(APPEND_FILE,ofile);
    file_close(ofile,first);
    return;
}

/* copy one file to another.
**
** Inputs:
**	first		file appended to
**	second		file put on at the end of above
*/
EXTERNDEF void copyfile_by_name(char *first,char *second)
{
    FILE *ofile;

    if (first == NULL) return;
    if (second == NULL) return;
    ofile = fopen(second,"w");
    if (ofile == NULL) {
	log_error("? Unable to open \"%s\" for copy.\n",second);
	return;
    };
    copyfile(first,ofile);
    file_close(ofile,second);
    return;
}

/* Puts a string to the current file, all output funnels through here.
** On the DG platform worry about the "console" file handle of -1.
**
** Inputs:
**	string
*/
LOCALDEF void put_string(const char *string)
{
    if (logOutputAlso && (logFile.file != NULL)) {
	fputs(string,logFile.file);
    };
#ifdef WINDOWS
    if (windowsFile != HFILE_ERROR) {
	_lwrite(windowsFile,string,strlen(string));
	return;
    };
/*
 * Add WIN32 support
 */
#elif defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
    if (win32File != INVALID_HANDLE_VALUE) {
	DWORD bytesWriten;
	WriteFile(win32File, string, strlen(string), &bytesWriten, NULL);
	return;
    };    
#endif
    if (cur_file == NULL) return;
#if AOSVS
	if (cur_file == (FILE *)-1) {
	    tcpmsg(0,0,string);
	} else {
#endif
	    fputs(string,cur_file);
#if AOSVS
	};
#endif
	return;
}

/* Formfeed
 */
EXTERNDEF void o_formfeed(void)
{
    put_string(cur_line);
    put_string("\n\f");
    cur_line[0] = NUL;
    cur_lcount = 1;
    return;
}

/* Newline
 */
EXTERNDEF void o_newline(int line)
{
    int i;
    while (cur_lcount < line) {
	cur_lcount += 1;
	put_string("\n");
    };
    cur_lcount += 1;
    put_string(cur_line);
	put_string("\n");
    i = 0;
    while (cur_line[i] != NUL) {
	if (cur_line[i] == '\n') cur_lcount += 1;
	i += 1;
    };
    cur_line[0] = NUL;
    return;
}

/* Accumulate a string to be written.
 *
 * Inputs:
 *	msg		ASCIZ string that is the message fragment
 */
LOCALDEF void write_string(const char *msg)
{
    if (msg == NULL) return;
    if ((strlen(cur_line) + strlen(msg)) < STRING_LENGTH) {
	strcat(cur_line,msg);
    } else {
	int i;
	put_string(cur_line);
	for (i = 0;cur_line[i] != NUL;i += 1) {
	    if (cur_line[i] == '\n') cur_lcount += 1;
	};
	cur_line[0] = NUL;
	put_string(msg);
	for (i = 0;msg[i] != NUL;i += 1) {
	    if (msg[i] == '\n') cur_lcount += 1;
	};
    };
    return;
}

/* compute current column number. Handles tabs and newlines.
 *
 * Outputs:
 *	prev		last character of string
 * Returns:
 *	current column number
 */
LOCALDEF int cur_column(char *prev)
{
    char chr;
    int tindex;
    int current;
    
    current = 0;
    tindex = 0;
    while (TRUE) {
	chr = cur_line[tindex++];
	switch (chr) {
	    case NUL: {			/* End of string */
		if (tindex > 1) *prev = cur_line[tindex-2];
		else *prev = '\n';
		return current;
		};
	    case '\t': {			/* TAB */
		current = ((current / 8) * 8) + 8; /* next tab stop */
		break;
	        };
	    case '\n': current = 0; break;	/* newline reset count */
	    default: current += 1; break;
	    };
	};
    }

/* Write a character to the current output file
 *
 * Inputs:
 *	chr		character
 */
LOCALDEF void write_char(char chr)
{
    char string[2];

    string[0] = chr;
    string[1] = NUL;
    write_string(string);
    return;
    }

/* Check to see if character is whitespace put out a space if needed.
 *
 * Inputs:
 *	last		character to check
 */
LOCALDEF void force_whitespace(char last)
{
    if (last == ' ') return;
    if (last == '\n') return;
    if (last == '\t') return;
    write_char(' ');
    return;
    }


/* Move to a particular column. If that column is already occupied no
 * whitespace is produced.
 *
 * Inputs:
 *	column		position of next character to print
 *			1 for first character of line, 9 for first tab etc.
 * Returns:
 *	last character of the string or 0 if unknown
 */
LOCALDEF char move_to_column(int column)
{
    int current;
    char prev;

    current = cur_column(&prev);
    column -= 1;				/* make it 0 based */
    while (column > current) {
	int next_tab;
	next_tab = ((current / 8) * 8) + 8;
	if (tabs_allowed && (next_tab <= column)) {
	    write_char('\t');			/* tab to next stop */
	    prev = '\t';
	    current = next_tab;			/* and set count */
	    }
	else {
	    write_char(' ');		/* fill with spaces */
	    prev = ' ';
	    current += 1;
	    };
	};
    return prev;
    }

/* Write a character.
 *
 * Inputs:
 *	chr		character to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed character
 */
EXTERNDEF void o_char(char chr,int column,BOOLEAN flag)
{
    char last;

    last = move_to_column(column);
    if (flag) force_whitespace(last);
    write_char(chr);
    return;
    }

/* Write a left aligned field starting at a particular column.
 *
 * Inputs:
 *	string		asciz string to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 */
EXTERNDEF void o_lstring (const char *string,int column,BOOLEAN flag)
{
    char last;

    last = move_to_column(column);
    if (flag) force_whitespace(last);
    write_string(string);
    return;
}

/* Write a left aligned field starting at a particular column which may
** not be terminated by a null.
 *
 * Inputs:
 *	string		asciz string to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 *	maxLength	maximum number of characters to output
 */
EXTERNDEF void o_maxLstring(char *string,int column,BOOLEAN flag,int maxLength)
{
    int i;
    if (string == NULL) return;
    if (string[0] == NUL) return;
    o_char(string[0],column,flag);
    for (i = 1;i < maxLength;i += 1) {
	if (string[i] == NUL) return;
	o_char(string[i],0,FALSE);
    };
    return;
}
			    
/* using printf style arguments write a left aligned field starting at
 * a particular column. The caller must ensure that the string will be
 * no longer than 3 * STRING_LENGTH characters.
 *
 * Inputs:
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 *	format		printf pattern followed by variables
 */
#if ANSI_C
EXTERNDEF void o_lprintf(int column,BOOLEAN flag,const char *format,...)
{
    va_list va_ptr;
    char last;
    char string[3*STRING_LENGTH+1];

    va_start(va_ptr,format);
    vsprintf(string,format,va_ptr);
    va_end(va_ptr);
#else
void o_lprintf(column,flag,format,va_alist)
int column;
BOOLEAN flag;
char *format;
int va_alist;
{
    va_list va_ptr;
    char last;
    char string[3*STRING_LENGTH+1];
    
    va_start(va_ptr);
    vsprintf(string,format,va_ptr);
    va_end(va_ptr);
#endif
    last = move_to_column(column);
    if (flag) force_whitespace(last);
    write_string(string);
    return;
}

/* Write a left aligned quoted string starting at a particular column.
 * The string can not contain any quote characters.
 *
 * Inputs:
 *	string		asciz string to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 */
EXTERNDEF void o_quoted_lstring (char *string,int column,BOOLEAN flag)
{
    o_char('"',column,flag);
    o_lstring(string,0,FALSE);
    o_char('"',0,FALSE);
    return;
}

/* Write a right aligned field ending at a particular column.
 * If writing the string would pass the column no leading whitespace is written.
 *
 * Inputs:
 *	string		asciz string to write
 *	column		column for last character
 *	flag		TRUE if whitespace must preceed string
 */
EXTERNDEF void o_rstring(char *string,int column,BOOLEAN flag)
{
    int length;
    char chr;

    cur_column(&chr);
    length = strlen(string);
    chr = move_to_column(column-length+1);
    if (flag) force_whitespace(chr);
    write_string(string);
    return;
    }

/* Convert a number to a string in a buffer. Handles radix up to 36
 * using 0-9 and uppercase A-Z
 *
 * Inputs:
 *	num		integer to output
 *	radix		Output radix
 *	buffer		buffer to put string into
 * Returns:
 *	buffer cell for next character of number
 */
LOCALDEF char *number_to_string(long int num,int radix,char *buffer)
{
    char digit;
    if (num == 0) {			/* exactly 0 done */
	*buffer++ = '0';
    	*buffer = NUL;
	return buffer;
	};	    
    if (num < 0) {			/* start with minus sign */
	*buffer++ = '-';
	num = -num;
	};
    digit = (char)(num % radix);
    num  = num / radix;
    if (num > 0) buffer = number_to_string(num,radix,buffer);
    *buffer++ = (digit <= 9) ? (char)(digit + '0') : (char)(digit - (char)10 + 'A');
    *buffer = NUL;
    return buffer;
    }

/* Write a left aligned numeric field starting at a particular column.
 *
 * Inputs:
 *	number		number to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 *	radix		radix used for output 2-36 are legal (default 10)
 */
EXTERNDEF void o_lnumber(long int number,int column,BOOLEAN flag,int radix)
{
    char buffer[66];			/* worst case is binary with sign */

    if ((radix < 2) || (radix > 36)) {
	log_error("? Bad radix %d.\n",radix);
	radix = 10;
	};
    number_to_string(number,radix,buffer);
    o_lstring(buffer,column,flag);
    return;
    }

/* Write a left aligned floating numeric field starting at a particular column.
 *
 * Inputs:
 *	number		number to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 */
EXTERNDEF void o_lfloat(float number,int column,BOOLEAN flag)
{
    char buffer[66];

    sprintf(buffer,"%f",number);
    o_lstring(buffer,column,flag);
    return;
}

/* Write a right aligned floating numeric field starting at a particular column
 *
 * Inputs:
 *	number		number to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 */
EXTERNDEF void o_rfloat(float number,int column,BOOLEAN flag)
{
    char buffer[66];

    sprintf(buffer,"%f",number);
    o_rstring(buffer,column,flag);
    return;
}

/* Write a left aligned fixed point number starting at a particular column.
 *
 * Inputs:
 *	number		number to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 *	radix		radix used for output 2-36 are legal (default 10)
 *	places		number of decimal places
 */
EXTERNDEF void o_fixed_number(long int number,int column,BOOLEAN flag,
    int radix,unsigned int places)
{
    long int value;
    char fraction[67];			/* worst case is binary */
    unsigned int i;
    char last;

    if ((radix < 2) || (radix > 36)) {
	log_error("? Bad radix %d.\n",radix);
	radix = 10;
    };
    last = move_to_column(column);
    if (flag) force_whitespace(last);
    if (number < 0) {
	o_char('-',0,FALSE);
    };
    value = (number >= 0) ? number : -number;
    fraction[places] = NUL;
    for (i = 0;i < places;i += 1) {
	char digit;
	digit = (char)(value % radix);
	value /= radix;
	fraction[places-i-1] = 
	    (digit <= 9) ? (char)(digit + '0') : (char)(digit - (char)10 + 'A');
    };
    o_lnumber(value,0,FALSE,radix);
    o_char('.',0,FALSE);
    o_lstring(fraction,0,FALSE);
    return;
}

/* Write a right aligned numeric field starting at a particular column.
 *
 * Inputs:
 *	number		number to write
 *	column		column for first character
 *	flag		TRUE if whitespace must preceed string
 *	radix		radix used for output 2-36 are legal (default 10)
 */
EXTERNDEF void o_rnumber(long int number,int column,BOOLEAN flag,int radix)
{
    char buffer[66];			/* worst case is binary with sign */

    if ((radix < 2) || (radix > 36)) {
	log_error("? Bad radix %d.\n",radix);
	radix = 10;
    };
    number_to_string(number,radix,buffer);
    o_rstring(buffer,column,flag);
    return;
}

/* output a date in dd-month-year format
**
** Inputs:
**	day
**	month
**	year
**	column		column for first character (0 if at present col)
**	flag		TRUE if whitespace must preceed string
*/
EXTERNDEF void o_ldate(INT8 day,INT8 month,INT16 year,int column,BOOLEAN flag)
{
    o_lnumber(day,column,flag,10);
    while (month < 1) {
	month += (INT8)12;
	year -= (INT16)1;
    };
    while (month > 12) {
	month -= (INT8)12;
	year += (INT16)1;
    };
    o_char('-',0,FALSE);
    o_lstring(months[month-1],0,FALSE);
    o_char('-',0,FALSE);
    o_lnumber(year,column,flag,10);
    return;
}

/* output a unix time of year
**
** Inputs:
**	unixTime    time to output
**	column		column for first character (0 if at present col)
**	whitespace	TRUE if whitespace must preceed string
*/
EXTERNDEF void o_unixTime(time_t unixTime,int column, BOOLEAN whitespace)
{
    char time_string[30];
    struct tm *time_now;

#if (!AOSVS)
#ifndef __STDC__
    tzset();
#endif
#endif
    time_now = localtime(&unixTime);
    strftime(time_string,DIM(time_string),"%d%b%Y %H:%M:%S",time_now);
    o_lstring(time_string,column,whitespace);
    return;
}

/* output a unix time of year n the mm/dd/yy format
**
** Inputs:
**	unixTime    time to output
**	column		column for first character (0 if at present col)
**	whitespace	TRUE if whitespace must preceed string
*/
EXTERNDEF void o_unixSlashTime(time_t unixTime,int column, BOOLEAN whitespace)
{
    char time_string[30];
    struct tm *time_now;

#if (!AOSVS)
#ifndef __STDC__
    tzset();
#endif
#endif
    time_now = localtime(&unixTime);
    strftime(time_string,DIM(time_string),"%m/%d/%y %H:%M:%S",time_now);
    o_lstring(time_string,column,whitespace);
    return;
}


/* Write a time of day field starting at a particular column.
 *
 * Inputs:
 *	time		time of day to write (in minutes)
 *	column		column for first character (0 if at present col)
 *	whitespace	TRUE if whitespace must preceed string
 */
EXTERNDEF void o_time_of_day(long int time,int column, BOOLEAN whitespace)
{
    char string[STRING_LENGTH+1];
    sprintf(string,"%ld:%02ld",time/60,time%60);
    o_lstring(string,column,whitespace);
    return;
}

/* Write a right aligned time of day field ending at a particular column.
 *
 * Inputs:
 *	number		time of day to write
 *	column		column for last character (0 if at present col)
 *	whitespace	TRUE if whitespace must preceed string
 */
EXTERNDEF void o_rtime_of_day(long int time,int column, BOOLEAN whitespace)
{
    char string[STRING_LENGTH+1];
    sprintf(string,"%ld:%02ld",time/60,time%60);
    o_rstring(string,column,whitespace);
    return;
}

/* Write an ammount of time field starting at a particular column.
 *
 * Inputs:
 *	number		ammount of time to write
 *	column		column for first character (0 if at present col)
 *	whitespace	TRUE if whitespace must preceed string
 */
EXTERNDEF void o_amount_of_time(long int time,int column, BOOLEAN whitespace)
{
    char string[STRING_LENGTH+1];
    sprintf(string,"%ld:%02ld",time/60,time%60);
    o_lstring(string,column,whitespace);
    return;
}

/* Write a right aligned time of day field ending at a particular column.
 *
 * Inputs:
 *	number		time of day to write
 *	column		column for last character (0 if at present col)
 *	whitespace	TRUE if whitespace must preceed string
 */
EXTERNDEF void o_ramount_of_time(long int time,int column, BOOLEAN whitespace)
{
    char string[STRING_LENGTH+1];
    sprintf(string,"%ld:%02ld",time/60,time%60);
    o_rstring(string,column,whitespace);
    return;
}

/* Write a tow height field starting at a particular column.
 *
 * Inputs:
 *	height		height of the tow
 *	column		column for first character (0 if at present col)
 *	whitespace	TRUE if whitespace must preceed string
 */
EXTERNDEF void o_tow_height(long int height,int column, BOOLEAN whitespace)
{
    char string[STRING_LENGTH+1];
    sprintf(string,"%ld.%1ldK",height/10,height%10);
    o_lstring(string,column,whitespace);
    return;
}

/* Write a right aligned tow height field ending at a particular column.
 *
 * Inputs:
 *	height		height of the tow
 *	column		column for last character (0 if at present col)
 *	whitespace	TRUE if whitespace must preceed string
 */
EXTERNDEF void o_rtow_height(long int height,int column, BOOLEAN whitespace)
{
    char string[STRING_LENGTH+1];
    sprintf(string,"%ld.%1ldK",height/10,height%10);
    o_rstring(string,column,whitespace);
    return;
}

/* Write a right aligned ammout of money field starting at a particular column.
 *
 * Inputs:
 *	number		cents to write
 *	column		column for last character (0 if at present col)
 *	flag		TRUE if whitespace must preceed string
 */
EXTERNDEF void o_rmoney(long int number,int column,BOOLEAN flag)
{
    char buffer[STRING_LENGTH+1];

    if (number < 0) {
	number = (-number);
	sprintf(buffer,"$(%ld.%02ld)",number/100,number%100);
	if (column != 0) column += 1;
    } else {
	sprintf(buffer,"$%ld.%02ld",number/100,number%100);
    };
    o_rstring(buffer,column,flag);
    return;
}

/* malloc space for a string and copy the value into the string.
 *
 * Inputs:
 *	string		string to get a free storage space for
 * Returns:
 *	newly allocated string
 */
EXTERNDEF char *alloc_str(char *string)
{   
    char *new_string;

    if (string == NULL) return alloc_str("");
    new_string = errmalloc((1+strlen(string)));
    strcpy(new_string,string);
    return new_string;
    }

/* Read a file into a free storage string.
 * intended only for small files.
 *
 * Inputs:
 *	name		name of file to read
 *	message		fragment of error message when read fails
 *	open_msg	TRUE to type message on opening file
 * Returns:
 *	contents of file as a free storage string
 */
EXTERNDEF char *alloc_file(char *name,char *message,BOOLEAN open_msg)
{
    char line[STRING_LENGTH+1];
    FILE *ifile;
    char *string;
    int old_max_malloc;

    old_max_malloc = set_max_malloc(10000);
    ifile = fopen(name,"r");
    if (ifile == NULL) {
	log_error("? Unable to read %s file \"%s\"\n",message,name);
	set_max_malloc(old_max_malloc);
	return NULL;
	};
    if (open_msg) {
	printf("Reading %s file \"%s\"\n",message,name);
    };
    string = alloc_str(NULL);
    line[STRING_LENGTH] = NUL;
    while (NULL != fgets(line,STRING_LENGTH,ifile)) {
	char *temp;
	temp = errmalloc(strlen(string) + strlen(line) + 1);
	strcpy(temp,string);
	strcat(temp,line);
	errfree((VOID *)string);
	string = temp;
	};
    if (EOF == fclose(ifile)) {
	log_error("? Unable to close %s file \"%s\".\n",message,name);
	};
    set_max_malloc(old_max_malloc);
    return string;
    }


/* Convert a string to uppercase. It is converted in place.
 * 
 * Inputs:
 *	string		string to convert
 */
EXTERNDEF char *uc_string(char *string)
{
    int i;

    if (string == NULL) return NULL;
    i = 0;
    while (string[i] != NUL) {
	string[i] = (char)toupper(string[i]);
	i += 1;
    };
    return string;
}


/* Decide if a string is completely blank
 *
 *Inputs:
 *	string		string to test
 * Returns:
 *	TRUE if the string is blank
 */
EXTERNDEF BOOLEAN is_blank(char *string)
{
    int i;
    i = 0;
    if (string == NULL) return TRUE;
    while (string[i] != NUL) {
	switch (string[i]) {
	    case '\n':
	    case ' ':
	    case '\t':
		break;
	    default:
		return FALSE;
	    };
	i += 1;
	};
    return TRUE;
    }

/* Decide if a string contains a number possibly preceeded or followed by
 * blanks
 *
 *Inputs:
 *	string		string to test
 * Returns:
 *	TRUE if the string is a number or empty
 */
EXTERNDEF BOOLEAN is_number(char *string)
{
    int i;
    BOOLEAN anydigit;
    BOOLEAN midblank;	

    anydigit = FALSE;
    midblank = FALSE;	
    for (i = 0;string[i] != NUL;i += 1) {
	switch (string[i]) {
	    case '\n':
	    case ' ':
	    case '\t':
		if (anydigit) midblank = TRUE;
		break;
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
		if (midblank) return FALSE;	/* digit(s) blank(s) digit */
		anydigit = TRUE;
	        break;
	    default:
		return FALSE;
	};
    };
    return TRUE;
}

/* When a line is used for input the rest of the line can be
** given back for later use. If any was given back it sits here untill used.
** Prompts are suppressed when input comes from this source.
*/
LOCALVAR char *restOfLastLine = NULL;

/* Returns a pointer to the rest of the current input line valid untill the
** next call to an input routine.
**
** Returns:
**	pointer to an asciz string (may be blank)
*/
EXTERNDEF const char *i_restOfLine(void)
{
    if (restOfLastLine == NULL) return "";
    return restOfLastLine;
}

/* No longer need rest of line
**
** Inputs:
**	moreOK	    TRUE if discarding input is done quietly
**		    FALSE to report an error if there is more input
*/
EXTERNDEF void i_ignoreRestOfLine(BOOLEAN moreOK)
{
    if (!moreOK && (restOfLastLine != NULL)) {
	log_error("? Discarding rest of line: %s\n",restOfLastLine);
    };
    errfree((VOID *)restOfLastLine);
    restOfLastLine = NULL;
    return;
}

/* Take back the remainder of a line. This will be used the next time
** something is requested from the user. The string must have been allocated
** with alloc_str() or equivalent. It will be freed after use
** 
** Inputs:
**	string	line to use later
*/
EXTERNDEF void i_giveBackLine(char *string)
{
    if (string == NULL) return;
    if (is_blank(string)) {
	errfree((VOID *)string);
	return;
    };
    errfree((VOID *)restOfLastLine);
    restOfLastLine = string;
    return;
}

/* Clean up after user input. Returns any free storage. Closes indirect files
** and prepares for program termination. The call is safe in the sense that
** it can be made and user input can be continued if desired.
*/
EXTERNDEF void i_endUserInput(void)
{
    i_ignoreRestOfLine(TRUE);
    closeAllIndirectFiles();
    return;
}

/* When reading from the user we might be reading from an indirect file.
** this is the central point that gets a line from the logical user
** (indirect file if appropriate) and returns the data.
** Also serviced here is the cannonical conversion of the string to
** a standard form. Leading and trailing blanks are removed, blank strings are
** ignored etc.
**
** Inputs:
**	string		buffer to put answer into
**	length		max length of buffer
**	defaultValue	default value if no string found
** Returns:
**	pointer to string fetched, this is in the users buffer
**	but may not start at the beginning because we ignore leading blanks
**	NULL on failures
*/
LOCALDEF char *getLineFromLogicalUser(char *string,int length,
				     char *defaultValue)
{
    char *inputLine;
    int index;
    FILE *inputFile;
    
    if (string == NULL) return NULL;
    if (length < 0) length = 0;
    while (TRUE) {
	if (restOfLastLine != NULL) {
	    strncpy(string,restOfLastLine,length);
	    errfree((VOID *)restOfLastLine);
	    restOfLastLine = NULL;
	    inputLine = string;
	} else {
	    inputFile = stdin;
	    if (nextIndirectFile > 0) {
		inputFile = indirectFiles[nextIndirectFile-1].inputFile;
	    };
	    if ((inputFile == stdin) && (callWhileIdle != NULL)) {
		while (TRUE) {
		    if (kbhit()) {
			break;
		    };
		    (callWhileIdle)();
		};
	    };
#ifdef USE_GETS
	    if (inputFile == stdin) {
		string[0] = NUL;
		echo_gets(-1, -1, string);
		inputLine = string;
		if (inputLine != NULL) printf("\n");
	    }
	    else {
#endif
		inputLine = fgets(string,length,inputFile);
#if MSDOS
		if (kbhit()) {
		    int c;
		    c = getch();
		    switch(c) {
			case 'q':
			case 'Q':
			    printf("Indirect File Processing Terminated\n");
			    closeAllIndirectFiles();
			    break;

			case 'p':
			case 'P':
			    printf("Indirect File Processing Paused,\n");
			    printf("Type any Key to continue...");
			    while(!kbhit()) ;
			    printf("\r\n");
			    break;

			default:
			    break;
		    };
		};
#endif
#ifdef USE_GETS
	    };
#endif
	};
	if (inputLine != NULL) break;	/* good data found */
	if (nextIndirectFile <= 0) {
	    log_error("? Could not get line from user\n");
	    return NULL;
	};
	/* close indirect file pop to last indirect file and try again */
	fclose(indirectFiles[nextIndirectFile-1].inputFile);
	errfree((VOID *)indirectFiles[nextIndirectFile-1].name);
	indirectFiles[nextIndirectFile-1].name = NULL;
	labels.count=0;			/* New file no labels */
	nextIndirectFile -= 1;
    };
    inputLine[length] = NUL;		/* make sure terminated */
    index = strlen(inputLine) - 1;
    if (inputLine[index] == '\n') {
	inputLine[index] = NUL;		/* remove EOL char */
	index -= 1;
    };
    if ((defaultValue != NULL) && is_blank(inputLine)) {
	strncpy(inputLine,defaultValue,length); /* blank use default */
	inputLine[length] = NUL;		/* make sure terminated */
    };
    while (isspace(inputLine[index]) && (index > 0)) {
	inputLine[index] = NUL;/* remove trailing blanks */
	index -= 1;
    };
    index = 0;
    while (isspace(inputLine[index])) {
	index += 1;			/* ignore leading blanks */
    };
    if (logFile.file != NULL) {
	fputs(&inputLine[index],logFile.file);
	fputc('\n',logFile.file);
    };
    return &inputLine[index];
}

/* Read an ASCIZ string from the user. Storage for the string is allocated with
 * malloc.
 *
 * Inputs:
 *	prompt		ASCIZ prompt string
 *	previous	ASCIZ default value if the user types <CR>
 *	upflag		TRUE if string is to be converted to upper case
 * Returns:
 *	pointer to string just created
 */
LOCALDEF char *getuser(char *prompt,char *previous,BOOLEAN upflag)
{
    char string[STRING_LENGTH + 1];
    char *userString;

    if (previous == NULL) previous = "";
    while (TRUE) {
	if ((restOfLastLine == NULL) && (nextIndirectFile == 0)) {
	    if (0 == strcmp(previous, "")) printf("%s: ", prompt);
	    else printf("%s [%s]: ", prompt, previous);
	};
	userString = getLineFromLogicalUser(string,STRING_LENGTH,previous);
	if (upflag) {
	    uc_string(userString);
	};
	return alloc_str(userString);
    };
}

#if GET_USER_ENABLED
/* Get a word from the user
**
** Inputs:
**	prompt		string to prompt with
**	previous	default
**	case_flag	TRUE converts to upper
** Returns:
**	newly allocated string the word typed by the user
*/
EXTERNDEF char *get_word(char *prompt,char *previous,BOOLEAN case_flag)
{
    char *word;
    BOOLEAN prev_blank;
    if (previous == NULL) previous = "";    
    prev_blank = is_blank(previous);
    while (TRUE) {
	char *line;
	line = getuser(prompt,previous,case_flag);
	i_giveBackLine(i_afterField(" \t",line,1,TRUE));
	word = alloc_str(i_getField(" \t",line,1,TRUE));
	errfree((VOID *)line);
	if (0 == strcmp(word,"?")) {
	    printf("Enter a word (a string containing no spaces).\n");
	    if (!prev_blank) {
		printf("Enter tilde ~ to leave the word blank.\n");
		};
	    errfree((VOID *)word);
	    continue;
	};
	if ((0 == strcmp(word,"~")) && (!prev_blank)) {
	    errfree((VOID *)word);
	    word = alloc_str(NULL);
	};
	return word;
    };
}
#endif

#if GET_USER_ENABLED
/* Similar to get_word but special characters (~ ? etc.) are handled by
** the caller allowing for messages more specific to what is being fetched.
**
** Inputs:
**	prompt		string to prompt with
**	previous	default
**	case_flag	TRUE converts to upper
** Returns:
**	newly allocated string the word typed by the user
*/
EXTERNDEF char *get_commandWord(char *prompt,char *previous,BOOLEAN case_flag)
{
    char *word;

    if (previous == NULL) previous = "";    
    while (TRUE) {
	char *line;
	line = getuser(prompt,previous,case_flag);
	i_giveBackLine(i_afterField(" \t",line,1,TRUE));
	word = alloc_str(i_getField(" \t",line,1,TRUE));
	errfree((VOID *)line);
	return word;
    };
}
#endif

#if GET_USER_ENABLED
/* get a name from the user. Do not allow certain characters
 *
 * Inputs:
 *	prompt
 *	previous
 * Returns:
 *	name the user typed allocated with errmalloc
 */
EXTERNDEF char *i_getName(char *prompt,char *previous)
{
    char *name;
    while (TRUE) {
	int i;
	name = get_word(prompt,previous,FALSE);
	for (i = 0;name[i] != '\0';i += 1) {
	    if (isspace(name[i])) name[i] = ' ';
	    if (name[i] == ',') break;
	    if (name[i] == '?') break;
        };
	if (name[i] == ',') {
	    printf("? Names can not contain the character %c\n",name[i]);
	    errfree((VOID *)name);
	    continue;
	};
	break;
    };
    return name;
}
#endif

#if GET_USER_ENABLED
/* Get a line from the user
**
** Inputs:
**	prompt		string to prompt with
**	previous	default
**	case_flag	TRUE converts to upper
** Returns:
**	newly allocated line fetched
*/
EXTERNDEF char *get_line(char *prompt,char *previous,BOOLEAN case_flag)
{
    char *line;
    BOOLEAN prev_blank;
    if (previous == NULL) previous = "";    
    prev_blank = is_blank(previous);
    while (TRUE) {
	line = getuser(prompt,previous,case_flag);
	if (0 == strcmp(line,"?")) {
	    printf("Enter a line of text.\n");
	    if (!prev_blank) {
		printf("Enter tilde ~ to leave the line blank.\n");
		};
	    errfree((VOID *)line);
	    continue;
	    };
	if ((0 == strcmp(line,"~")) && (!prev_blank)) {
	    errfree((VOID *)line);
	    line = alloc_str(NULL);
	    };
	return line;
	};
    }
#endif

/* Check a phone number for legit area exchange and number
 *
 * Inputs:
 *	area		area code 0 if none
 *	exchange	3 digit exchange
 *	number		4 digit number
 *      extention	up to 5 digit number
 * Returns:
 *	TRUE if a US style phone number with area of 0 allowed
 */
LOCALDEF BOOLEAN check_phone(int area,int exchange,int number,
	long int extention)
{
    if ((area<0) || (area>999)) return FALSE;
    if ((exchange<=0) || (exchange>999)) return FALSE;
    if ((number<0) || (number>9999)) return FALSE;
    if ((extention<0) || (extention>99999L)) return FALSE;
    return TRUE;
}

/* Parse a phone number. A large variety of formats are permitted:
 *	(aaa) eee-nnnn X xxxxx
 *	aaa eee-nnnn X xxxxx
 *	aaa eee nnnn X xxxxx
 *	aaa eeennnn X xxxxx
 *	eee-nnnn X xxxxx
 *	eee-nnnn X xxxxx
 *	eee nnnn X xxxxx
 *	eeennnn X xxxxx
 * The extention is optional, if there the X is required. The extention may
 * have any number of digits.
 *
 * Inputs:
 *	string		number to parse
 * Outputs:
 *	area		area code 0 if none
 *	exchange	3 digit exchange
 *	number		4 digit number
 *	extention	up to 5 digit number
 * Returns:
 *	TRUE if a US style phone number was parsed with integers assigned
 *	to the outputs
 */
LOCALDEF BOOLEAN parse_phone(char *string,int *area,int *exchange,
	int *number,long int *extention)
{
    char rest[STRING_LENGTH+1];
    int matches;
/* with area code and extention */
    matches = sscanf(string," ( %3d ) %3d - %4d X %d %s",
	    area,exchange,number,extention,rest);
    if (matches == 4) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," ( %3d ) %3d- %4d X %d %s",
	    area,exchange,number,extention,rest);
    if (matches == 4) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," ( %3d ) %3d-%4d X %d %s",
	    area,exchange,number,extention,rest);
    if (matches == 4) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," ( %3d ) %3d %4d X %d %s",
	    area,exchange,number,extention,rest);
    if (matches == 4) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," %3d %3d - %4d X %d %s",
	    area,exchange,number,extention,rest);
    if (matches == 4) return check_phone(*area,*exchange,*number,*extention);
/* with area code and no extention */
    *extention = 0;
    matches = sscanf(string," ( %3d ) %3d - %4d %s",area,exchange,number,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," ( %3d ) %3d- %4d %s",area,exchange,number,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," ( %3d ) %3d-%4d %s",area,exchange,number,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," ( %3d ) %3d %4d %s",area,exchange,number,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," %3d %3d - %4d %s",area,exchange,number,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    *area = 0;
/* without area code and with extention */
    matches = sscanf(string,"%3d - %4d X %d %s",exchange,number,extention,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," %3d %4d X %d %s",exchange,number,extention,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," %3d %3d %4d X %d %s",
	    area,exchange,number,extention,rest);
    if (matches == 4) return check_phone(*area,*exchange,*number,*extention);
/* without area code and without extention */
    *extention = 0;
    matches = sscanf(string,"%3d - %4d %s",exchange,number,rest);
    if (matches == 2) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," %3d %4d %s",exchange,number,rest);
    if (matches == 2) return check_phone(*area,*exchange,*number,*extention);
    matches = sscanf(string," %3d %3d %4d %s",area,exchange,number,rest);
    if (matches == 3) return check_phone(*area,*exchange,*number,*extention);
    *area = 0;
    *exchange = 0;
    *number = 0;
    *extention = 0;
    return FALSE;
}

#if GET_USER_ENABLED
/* parse a phone number
** allocate a new string with the number in normalized form
**
** Inputs:
**	phone	    string to parse
** Returns:
**	newly allocated string with the normalized phone number
*/
LOCALDEF char *normalize_phone(char *phone)
{
    int area,exchange,number;
    long int extention;

    if (parse_phone(phone,&area,&exchange,&number,&extention)) {
	char ph[20];
	char ph2[30];
	if (area == 0) {
	    sprintf(ph,"%03d-%04d",exchange,number);
	} else {
	    sprintf(ph,"(%03d)%03d-%04d",area,exchange,number);
	};
	if (extention != 0) {
	    sprintf(ph2,"%s X%d",ph,extention);
	} else {
	    return alloc_str(ph);
	};
	return alloc_str(ph2);
    };
    return alloc_str(phone);
}
#endif

/* Determine if a string is a valid phone number.
 * A large variety of formats are permitted:
 *	011 <any-string>
 *	(aaa) eee-nnnn X xxxxx
 *	aaa eee-nnnn X xxxxx
 *	aaa eee nnnn X xxxxx
 *	aaa eeennnn X xxxxx
 *	eee-nnnn X xxxxx
 *	eee-nnnn X xxxxx
 *	eee nnnn X xxxxx
 *	eeennnn X xxxxx
 * The extention is optional, if there the X is required. The extention may
 * have any number of digits.
 *
 * Inputs:
 *	string		number to parse
 * Returns:
 *	TRUE if a valid phone number was parsed
 */
EXTERNDEF BOOLEAN is_valid_phone(char *string)
{
    char prefix[STRING_LENGTH+1];
    int matches;
    int area,exchange,number;
    long int extention;

    matches = sscanf(string," %s ",prefix);
    if (matches >= 1) {
	if (0 == strcmp(prefix,"011")) {
	    return TRUE;  /* international number */
	};
    };
    if (parse_phone(string,&area,&exchange,&number,&extention)) {
	return TRUE;
    };
    return FALSE;
}

#if GET_USER_ENABLED
/* Get a phone number from the user. See parse_phone for legal formats.
 *
 * Inputs:
 *	prompt		string to prompt with
 *	previous	default
 */
EXTERNDEF char *get_phone(char *prompt,char *previous)
{
    char *phone;

    while (TRUE) {
	phone = getuser(prompt,previous,TRUE); 
	if (0 == strcmp(phone,"~")) {
	    errfree((VOID *)phone);
	    return alloc_str(NULL);
	};
	if (0 == strcmp(phone,"?")) {
	    printf("Enter a phone number in the form (aaa)eee-nnnn\n");
	    printf("Followed by an optional extention Xnnnnn\n");
	    printf("Enter \"~\" if no phone number is desired.\n");
	    errfree((VOID *)phone);
	    continue;
	};
	if (is_valid_phone(phone)) {
	    char *temp;
	    temp = normalize_phone(phone);
	    errfree((VOID *)phone);
	    return temp;
	};
	if (0 == strcmp(phone,"")) {
	    return phone;
	};
	printf("Phone number \"%s\" is not a US style phone number.\n",phone);
	if (get_yesno("Use number anyway",FALSE)) return phone;
	errfree((VOID *)phone);
    };
}
#endif

#if GET_USER_ENABLED
/* Get an area code number from the user.
 *
 * Inputs:
 *	prompt		string to prompt with
 *	previous	default
 */
EXTERNDEF char *get_areaCode(char *prompt,char *previous)
{
    char *phone;
    int area;
    int matches;
    char dummy;

    while (TRUE) {
	phone = getuser(prompt,previous,TRUE); 
	if (0 == strcmp(phone,"~")) {
	    errfree((VOID *)phone);
	    return alloc_str(NULL);
	};
	if (0 == strcmp(phone,"?")) {
	    printf("Enter an area code the form aaa\n");
	    printf("Enter \"~\" if no number is desired.\n");
	    errfree((VOID *)phone);
	    continue;
	};
	matches = sscanf(phone," %3d %c",&area,&dummy);
	if (matches != 1) {
	    matches = sscanf(phone," ( %3d ) %c",&area,&dummy);
	};
	if ((matches == 1) && (area >= 0)) {
	    char string[10];
	    errfree((VOID *)phone);
	    sprintf(string,"%03d",area);
	    return alloc_str(string);
	};
	if (0 == strcmp(phone,"")) {
	    return phone;
	};
	printf("Invalid area code \"%s\"\n",phone);
	errfree((VOID *)phone);
    };
}
#endif

/* Check if a string is exactly a word
 *
 * Inputs:
 *	string		string to check
 * Returns:
 *	TRUE if string contains no blanks, FALSE otherwise
 */
EXTERNDEF BOOLEAN is_word(char *string)
{
    int i;
    i = 0;
    while (string[i] != NUL) {
	if (string[i] == ' ') return FALSE;
	if (string[i] == '\t') return FALSE;
	i += 1;
	};
    return TRUE;
    }


/* Parse a filename into name and type fields
 *
 * Inputs:
 *	fname		string to parse
 *	name		buffer for name
 *	type		buffer for type
 * Returns:
 *	TRUE if parse was successful FALSE otherwise
 */
EXTERNDEF BOOLEAN parse_filename(char *fname,char *name,char *type)
{
    int fields;
    char *fcopy;

    name[0] = NUL;
    type[0] = NUL;
    if (fname == NULL) return FALSE;
    fields = sscanf(fname," .%s",type);
    if (fields == 1) return is_word(type);	/* just file type */
    fcopy = alloc_str(fname);
    for(fields = 0;fcopy[fields];fields += 1) {
	if (fcopy[fields] == '.') fcopy[fields] = ' ';
	};
    fields = sscanf(fcopy," %s %s ",name,type);
    errfree((VOID *)fcopy);
    if (fields == 1) return is_word(name);
    if (fields == 2) {
    	if (!is_word(name)) return FALSE;
	return is_word(type);
	};
    return FALSE;
    }

#if GET_USER_ENABLED
/* Get a filename from the user.
 *
 * Inputs:
 *	prompt			prompt string
 *	previous		default filename
 * Returns:
 *	errmalloc of a filename
 */
EXTERNDEF char *get_filename(char *prompt,char *previous)
{
    char *fname;
    char name_default[STRING_LENGTH+1];
    char type_default[STRING_LENGTH+1];
    char first[STRING_LENGTH+1];
    char second[2*STRING_LENGTH+4];

    while (TRUE) {
	int matches;
	int port;
	char dummy;
	parse_filename(previous,name_default,type_default);
	strcpy(first,name_default);
	strcat(first,".");
	strcat(first,type_default);
	strcpy(second,prompt);
	strcat(second," [");
	strcat(second,first);
	strcat(second,"]");
	fname = get_word(second,"",FALSE);
	if (fname[0] == NUL) {
	    errfree((VOID *)fname);
	    return alloc_str(previous);
	};	
	matches = sscanf(fname," COM%d %c",&port,&dummy);
	if (matches == 1) {
	    return fname;
	};
	matches = sscanf(fname," LPT%d %c",&port,&dummy);
	if (matches == 1) {
	    return fname;
	};
	if (0 == strcmp(fname,"~")) {
	    errfree((VOID *)fname);
	    return alloc_str(NULL);
	    };
	if (0 == strcmp(fname,"?")) {
	    printf("Enter the name of a file.\n");
	    printf("Enter \"~\" if no name is desired.\n");
	    errfree((VOID *)fname);
	    continue;
	    };
	if (parse_filename(fname,first,second)) {
	    if (first[0] == NUL) strcpy(first,name_default);
	    if (first[strlen(first)-1] != ':') {
		strcat(first,".");
		if (second[0] == NUL) strcat(first,type_default);
		else strcat(first,second);
	    };
	    errfree((VOID *)fname);
	    return alloc_str(first);
	    };
	printf("? Bad format for filename \"%s\"\n",fname);
	errfree((VOID *)fname);
	};
    }
#endif

/* Find a string in a NULL terminated list of strings. Least common
 * substring recognition is performed.
 *
 * Inputs:
 *	command		string to look for
 *	list		list of strings to look in
 * Returns:
 *	-1	ambiguous match
 *	-2	no match
 *	>=0	index into list of entry that matched
 */
EXTERNDEF int find_command(char *command,char *list[])
{
    int i;
    int num_matches = 0;
    int isave;

    if (command == NULL) return -1;
    i = -1;
    isave = -2;					/* no match yet */
    while (list[++i] != NULL) {
	int commandLength;
	int listLength;
	int ret_code;
	commandLength = strlen(command);
	listLength = strlen(list[i]);
	ret_code = compare_name(command,list[i],MAX(commandLength,listLength));
	if (ret_code == 0) {
	    num_matches = 1;
	    isave = i;
	    break;
	} else {
	     ret_code = compare_name(command,list[i],commandLength);
	};
	if (ret_code == 0) {
	    if (num_matches++ == 0) isave = i;
	};
    };
    if (num_matches > 1) return -1;		/* many */
    return isave;				/* none or one */
}

/* Print a list of character strings in a column sorted format.
 *
 * Allows options to define printing.  If the number of columns is 0
 * function determines the maximum columns, based on lines.  If lines
 * is more than determined by calculation of columns then that many lines
 * will be used.
 *
 * Inputs:
 *	list		List of character strings.
 *	lines		Number of lines to print the output in.  If this
 *			is 0, the best fit based on string size is performed.
 *	ScreenWidth	With of the screen (Used to calculate columns) max is
 *			128, will truncate if screen width is less then length.
 *	promptAfter	If prompt is not NULL display it after this many lines.
 *	prompt		Prompt that will be displayed after promptAfter lines.
 * 
 * Returns:
 *	nothing
 */
EXTERNDEF void o_print_list(char *list[], int lines, int screenWidth,
			   int promptAfter, char *prompt)
{
    o_print_slist(list, sizeof(list[0]), 0, lines, screenWidth,
		  promptAfter, prompt);
}

/* Print a list of character strings, from within structures,
 * in a column sorted format.
 *
 * Allows options to define printing.  If the number of columns is 0
 * function determines the maximum columns, based on lines.  If lines
 * is more than determined by calculation of columns then that many lines
 * will be used.
 *
 * Inputs:
 *	list		List of structures
 *      size		Size of one Structure
 *	offset		Offset of string within structure
 *	lines		Number of lines to print the output in.  If this
 *			is 0, the best fit based on string size is performed.
 *	ScreenWidth	With of the screen (Used to calculate columns) max is
 *			128
 *	promptAfter	If prompt is not NULL display it after this many lines.
 *	prompt		Prompt that will be displayed after promptAfter lines.
 * 
 * Returns:
 *	nothing
 */
EXTERNDEF void o_print_slist(void *list, int size, int offset, int lines,
			     int screenWidth, int promptAfter, char *prompt)
{
    const int maxScreen = 128;
    int i, j, k;
    size_t maxLength;
    int columns;
    int howMany;
    int maxInCol;
    int index;
    int linesPrinted;			/* Used for prompting */
    char *ptr;

    /* Don't exceed maximum screen width */
    if (screenWidth > maxScreen) screenWidth = maxScreen; /* Max screen width is 128 */
    /* First find out how many elements and how big */
    maxLength = 0;
    for (i = 0;i < maxScreen*8;i += 1) { /* 8 cols just to prevent runaways */
	size_t length;
	ptr = *(char **) ((char *)list + size*i + offset);
	if (ptr == NULL) break;
	length = strlen(ptr);
	if (length > maxLength) {
	    maxLength = length;
	};
    };
    howMany = i;
    if (howMany == 0) return;		/* You have to have something to print */
    /* Compute columns */
    columns = screenWidth / (maxLength+1);
    if (columns <= 0) columns = 1; /* Do not allow 0 columns */
    maxInCol = howMany / columns;
    if (lines > maxInCol) {		/* They want more lines */
	maxLength = screenWidth / ((howMany / lines) + 1);
	columns = screenWidth / (maxLength+1);
	maxInCol = lines;
    };
    k = 0;				/* Will be used to compute totals */
    linesPrinted = 0;			/* Haven't printed any yet. */
    for(i = 0; i < maxInCol; i++) {
	for (j = 0; j < columns; j++) {
	    index = k + (j * maxInCol);
	    ptr = *(char **) ((char *)list + size*index + offset);
	    if ((index < howMany) && (ptr != NULL)) {
		o_lstring(ptr,j*(maxLength+1),(i != 0));
	    };
	};
	k += 1;
	if (k > howMany) {
	    break;
	};
	o_newline(0);
	linesPrinted += 1;
	/* Prompt if necessary */
	if ((linesPrinted >= promptAfter) && (prompt != NULL)) {
	    char *p;
	    p = get_line(prompt, NULL, TRUE);
	    linesPrinted = 0;
	    errfree(p);
	};
	    
    };
    o_newline(0);
}

#if GET_USER_ENABLED
/* Get one of a list of items. The item names must contain no spaces.
** If the user types exactly ~ the string is taken as the empty string.
**
** Inputs:
**	prompt		string to prompt with
**	previous	default string
**	list		list to select entry from
**	nullok		OK to return no item in the list as -1
** Returns:
**	index into list for matching item
**	-1 if not in list (~ typed)
*/
EXTERNDEF int get_list(char *prompt,char *previous,char *list[],
	BOOLEAN nullok)
{
    char *string;
    int command;

    while (TRUE) {
	string = get_commandWord(prompt,previous,TRUE);
	if (0 == strcmp("~",string)) {
	    errfree((VOID *)string);
 	    if (nullok) return -1;
	    string = alloc_str(NULL);
	};
	command = find_command(string,list);
	if (command >= 0) {
	    errfree((VOID *)string);
	    return command;
	};
	if (0 == strcmp("?",string)) {
	    o_print_list(list, 0, 80, 0, NULL);
	} else if (command == -1) {
	    printf("? Ambiguous - please type a longer string - \"%s\"\n",
		   string);
	} else {
	    printf("? Unrecognized \"%s\"\n",string);
	};
	errfree((VOID *)string);
    };
}
#endif

#if GET_USER_ENABLED
/* Count the members of a namelist
 *
 * Inputs:
 *	nam	    namelist to count
 */
EXTERNDEF int sizeof_namelist(struct namelist *nam)
{
    int i = 0;
    while (nam != NULL) {
	i += 1;
	nam = nam->next;
    };
    return i;
}

/* Same as get_list except the list is a struct namelist.
 *
 * Inputs:
 *	prompt		string to prompt with
 *	previous	default string
 *	list		list to select entry from
 *	nullok		OK to return no item in the list as -1
 * Returns:
 *	string of matching item
 *	NULL if not in list (~ typed)
 */
EXTERNDEF char *get_namelist(char *prompt,char *previous,struct namelist *list,
	BOOLEAN nullok)
{
    char **converted_list;
    char *match;
    int i;
    
    i = sizeof_namelist(list);
    converted_list = (char **)errmalloc(sizeof(char *) * (i + 1));
    converted_list[0] = NULL;
    i = 0;
    while (list != NULL) {
	converted_list[i] = list->name;
	converted_list[i+1] = NULL;
	i += 1;
	list = list->next;
    };
    i = get_list(prompt,previous,converted_list,nullok);
    match = (i >= 0) ? converted_list[i] : NULL;
    errfree((VOID *)converted_list);
    return match;
}
#endif

#if GET_USER_ENABLED
/* Get a boolean value.
 *
 * Inputs:
 *	prompt		string to prompt with
 *	offString	string used if value is FALSE
 *	onString	string used if value is TRUE
 *	previous	default answer
 * Returns:
 *	TRUE if yes
 */
EXTERNDEF BOOLEAN get_boolean(char *prompt,char *offString,
			      char *onString,BOOLEAN previous)
{
    int answer;

    char *yesno[3];
    yesno[0] = (offString != NULL) ? offString : "NO";
    yesno[1] = (onString != NULL) ? onString : "YES";
    yesno[2] = NULL;
    answer = get_list(prompt,
		      ((previous) ? yesno[1] : yesno[0]),
		      yesno,FALSE);
    if (answer == 0) return FALSE;
    return TRUE;
}
#endif

#if GET_USER_ENABLED
/* Get a yes or no answer
 *
 * Inputs:
 *	prompt		string to prompt with
 *	previous	default answer
 * Returns:
 *	TRUE if yes
 */
EXTERNDEF BOOLEAN get_yesno(char *prompt,BOOLEAN previous)
{
    int answer;

    answer = get_boolean(prompt,NULL,NULL,previous);
    if (answer == 0) return FALSE;
    return TRUE;
}
#endif

/* See if date is reasonable. 0 is ok.
 *
 * Inputs:
 *	month
 *	day
 *	year
 * Returns:
 *	TRUE if date is reasonable
 */
LOCALDEF BOOLEAN ok_date(int month,int day,int year)
{
    if ((month > 12) || (month < 0)) return FALSE;
    if (year < 0) return FALSE;
    if (day < 0) return FALSE;
    switch (month) {
	case 0:
	case 1:
	case 3:
	case 5:
	case 7:
	case 8:
	case 10:
	case 12:
	    if (day > 31) return FALSE; break;
	case 2:
	    if (day > 29) return FALSE; break;
	default:
	    if (day > 30) return FALSE; break;
	};
    return TRUE;
    }

/* Parse a date converting date into 3 integers.
 * Missing fields return 0.
 *
 * Dates can be presented in one of several forms:
 *
 *	n-n-n		month-day-year
 *	n-n		month-day
 *	n		day
 *	n-ccc-n		day-month-year	(month as any number of  chars)
 *	n-ccc		day-month
 *	ccc-n-n		month-day-year
 *	ccc-n		month-day
 *	ccc		month
 *
 * The - can be either a '-' a '/' a '.' or a space.
 *
 * Inputs:
 *	string		string to parse
 *	month		address of month
 *	day		address of day
 *	year		address of year
 * Returns:
 *	TRUE if date parsed OK.
 */
EXTERNDEF BOOLEAN parse_date(char *string,int *month,int *day,int *year)
{
    char first[STRING_LENGTH+1];
    char second[STRING_LENGTH+1];
    char third[STRING_LENGTH+1];
    char *upstring;
    int fields;
    int temp;
    
    *month = 0;
    *day = 0;
    *year = 0;
    upstring = alloc_str(string);
    uc_string(upstring);
    for (temp = 0;upstring[temp];temp += 1) {
	if (upstring[temp] == '-') upstring[temp] = ' ';
	else if (upstring[temp] == '/') upstring[temp] = ' ';
	else if (upstring[temp] == ',') upstring[temp] = ' ';
	else if (upstring[temp] == '.') upstring[temp] = ' ';
	};
    if (is_blank(upstring)) {
	errfree(upstring);
	return ok_date(*month,*day,*year);
    };
    if (is_number(upstring)) {
	sscanf(upstring," %d ",day);
	return ok_date(*month,*day,*year);
	};
    fields = sscanf(upstring," %s %s %s ",first,second,third);
    errfree((VOID *)upstring);
    if (fields < 1) {
	return FALSE;
	};
    if (fields == 1) {
    	temp = find_command(first,ucmonths);
	if (temp < 0) return FALSE;
	*month = temp + 1;
	return TRUE;
	};
    if (fields > 1) {
	temp = find_command(first,ucmonths);
	if (temp >= 0) {
	    *month = temp + 1;			/* month was first */
	    if (1 != sscanf(second," %d",day)) return FALSE;
	    if (fields > 2) if (1 != sscanf(third," %d",year)) return FALSE;
	    return ok_date(*month,*day,*year);
	    };
	temp = find_command(second,ucmonths);
	if (temp >= 0) {
	    *month = temp + 1;			/* month was second */
	    if (1 != sscanf(first," %d",day)) return FALSE;
	    if (fields > 2) if (1 != sscanf(third," %d",year)) return FALSE;
	    return ok_date(*month,*day,*year);
	    };
	};
    if (1 != sscanf(first," %d",month)) return FALSE;
    if (1 != sscanf(second," %d",day)) return FALSE;
    if (fields > 2) if (1 != sscanf(third," %d",year)) return FALSE;
    return ok_date(*month,*day,*year);
    }

#if GET_USER_ENABLED
/* Get a date from the user. The date is considered a modification of
 * the previous date. There must be a previous date.
 *
 * Inputs:
 *	prompt		string to prompt with
 *	previous	previous date (the default)
 * Returns:
 *	a malloc of the new date in the form dd-mmm-yy
 */
EXTERNDEF char *get_date(char *prompt,char *previous)
{
    int prev_day,prev_month,prev_year;
    int day,month,year;
    char s_date[20];
    char *string;

    if (previous == NULL) previous = "";
    if (!parse_date(previous,&prev_month,&prev_day,&prev_year)) {
	printf("? Improper format for previous date: \"%s\"\n",previous);
    };
    while (TRUE) {
	string = getuser(prompt,previous,TRUE);
	if (0 == strcmp(string,"?")) {
	    printf("Enter a date in the form mm-dd-yy.\n");
	    errfree((VOID *)string);
	    continue;
	};
	if (parse_date(string,&month,&day,&year)) {
	    month = (month != 0) ? month : prev_month;
	    day = (day != 0) ? day : prev_day;
	    year = (year != 0) ? year : prev_year;
	    sprintf(s_date,"%d-%s-%02d",day,months[month-1],year);
	    errfree((VOID *)string);
	    break;
	} else {
	    printf("? Improper format for a date: \"%s\"\n",string);
	};
	errfree((VOID *)string);
    };
    return alloc_str(s_date);
}
#endif

/* Parse a time converting time into 2 integers.
 * Missing fields return 0.
 *
 * Times can be presented in one of several forms:
 *
 *	hh:mm
 *	hhmm
 *	hmm
 *	mm
 *	m
 *
 * The : can be either a : or a space.
 *
 * Inputs:
 *	string		string to parse
 * Outputs:
 *	hour		hour
 *	minutes		minutes
 * Returns:
 *	TRUE if time parsed OK.
 */
EXTERNDEF BOOLEAN parse_time(char *string,long int *hour,long int *minutes)
{
    long int first;
    long int second;
    int fields;

    *minutes = 0;
    *hour = 0;
    fields = sscanf(string," %ld %ld",&first,&second);
    if (fields != 2) fields = sscanf(string," %ld : %ld",&first,&second);
    if (fields < 1) {
	return FALSE;
	};
    if (fields == 1) {
	long int afirst;
	afirst = first;
	if (afirst < 0) afirst = (-afirst);
	second = afirst%100;
	first = first/100;
	}
    else if (fields != 2) return FALSE;
    *hour = first%24;
    *minutes = second%60;
    if (first > 23) return FALSE;
    if (first < 0) return FALSE;
    if (second < 0) return FALSE;
    if (second >= 60) return FALSE;
    return TRUE;
    }

#if GET_USER_ENABLED
/* Get a time from the user. A 24 hour clock is assumed. The new time
 * is a modification of the old if only minutes are supplied. If the minutes
 * are less than the old in this case the old hour is increased.
 *
 * Inputs:
 *	prompt		string to prompt with
 *	previous	previous time (the default) in minutes
 * Returns:
 *	the time of day in minutes past midnight
 */
EXTERNDEF long int get_time(char *prompt,long int previous)
{
    char *string;
    char pstring[10];
    long int prev_hours;
    long int prev_min;
    long int return_value;

    prev_hours = (previous/60)%24;
    prev_min = (previous%60);
    sprintf(pstring,"%ld:%02ld",prev_hours,prev_min);
    while (TRUE) {
	long int hours,minutes;
	string = get_commandWord(prompt,pstring,TRUE);
	if (0 == strcmp(string,"?")) {
	    printf("A time in the format hh:mm (24 hour clock).\n");
	    printf("\tIf only minutes are supplied they are treated as new\n");
	    printf("\tminutes with the same hour. If the new minutes are\n");
	    printf("\tless than the old the hour is increased by one.\n");
	    errfree((VOID *)string);
	    continue;
	};
	if (parse_time(string,&hours,&minutes)) {
	    if (strlen(string) <= 2) {
		hours = prev_hours;
		if (prev_min > minutes) hours += 1;
	    };
	    errfree((VOID *)string);
	    return_value = hours*60 + minutes;
	    break;
	} else {
	     printf("? Improper format for a time: \"%s\"\n",string);
	};
	errfree((VOID *)string);
    };
    return return_value;
}
#endif

#if GET_USER_ENABLED
/* Read number from the user.
 *
 * Inputs:
 *	prompt		ASCIZ prompt string
 *	previous	default value
 * Returns:
 *	value read
 */
EXTERNDEF long int get_number(char *prompt,long int previous)
{
    char string[STRING_LENGTH+1];
    char prevString[STRING_LENGTH+1];
    char *userString;
    long int number;

    if (prompt == NULL) prompt = "";
    while (TRUE) {
	sprintf(prevString,"%ld",previous);
	userString = get_commandWord(prompt,prevString,TRUE);
	if (userString == NULL) {
	    return previous;
	};
	if (0 == strcmp(userString,"?")) {
	    printf("Enter a number.\n");
	    errfree((VOID *)userString);
	    continue;
	};
	if (1 == sscanf(userString," %ld %s",&number,string)) {
	    break;
	} else if (1 == sscanf(userString," %s ",string)) {
	    if (string[0] == NUL) {
		number = previous;
		break;
	    };
	    printf("? Improper format for a number: \"%s\"\n",userString);
	} else {
	    number = previous;
	    break;
	};
	errfree((VOID *)userString);
    };
    errfree((VOID *)userString);
    return number;
}
#endif

#if GET_USER_ENABLED
/* Read HEX number from the user.
 *
 * Inputs:
 *	prompt		ASCIZ prompt string
 *	previous	default value
 * Returns:
 *	value read
 */
EXTERNDEF long int get_hexNumber(char *prompt,long int previous)
{
    char string[STRING_LENGTH+1];
    char prevString[STRING_LENGTH+1];
    char *userString;
    long int number;

    if (prompt == NULL) prompt = "";
    while (TRUE) {
	sprintf(prevString,"%lx",previous);
	userString = get_commandWord(prompt,prevString,TRUE);
	if (userString == NULL) {
	    return previous;
	};
	if (0 == strcmp(userString,"?")) {
	    printf("Enter a hexadecimal number.\n");
	    errfree((VOID *)userString);
	    continue;
	};
	if (1 == sscanf(userString," %lx %s",&number,string)) {
	    break;
	} else if (1 == sscanf(userString," %s ",string)) {
	    if (string[0] == NUL) {
		number = previous;
		break;
	    };
	    printf("? Improper format for a hexadecimal number: \"%s\"\n",
		   userString);
	} else {
	    number = previous;
	    break;
	};
	errfree((VOID *)userString);
    };
    errfree((VOID *)userString);
    return number;
}
#endif

#if GET_USER_ENABLED
/* Read number in a range from the user.
 *
 * Inputs:
 *	prompt		ASCIZ prompt string
 *	previous	default value
 *	min		minimum value allowed
 *	max		maximum value allowed
 * Returns:
 *	value read
 */
EXTERNDEF long int i_getNumberInRange(char *prompt,long int previous,
				       long int min,long int max)
{
    long int number;

    while (TRUE) {
	number = get_number(prompt,previous);
	if (number < min) {
	    printf("? %ld is too small, must be greater than %ld\n",
		   number,min-1);
	    continue;
	};
	if (number > max) {
	    printf("? %ld is too large, must be less than %ld\n",
		   number,max+1);
	    continue;
	};
	return number;
    };
}
#endif

#if GET_USER_ENABLED
/* Read integer number with imbedded decimal point from the user.
 *
 * Inputs:
 *	prompt		ASCIZ prompt string
 *	right		number of digits permitted right of the decimal
 *	previous	default value
 *	neg_ok		negative numbers are allowed
 * Returns:
 *	value read
 */
EXTERNDEF long int get_fixed_number(char *prompt,int right,char *previous,
    BOOLEAN neg_ok)
{
    char string[STRING_LENGTH + 1];
    char *userString;
    long int number;
    BOOLEAN is_negative;
    BOOLEAN ended;
    int i;

    if (previous == NULL) previous = "";
    while (TRUE) {
	BOOLEAN got_number;
	got_number = FALSE;
	userString = get_commandWord(prompt,previous,TRUE);
	if (userString == NULL) {
	    userString = alloc_str(previous);
	};
	if (0 == strcmp(userString,"?")) {
	    printf("Enter a number with decimal and fraction.\n");
	    errfree((VOID *)userString);
	    continue;
	};
	if (2 == sscanf(userString, " %ld.%s ", &number, string)) {
	    got_number = TRUE;		/* both halves */
	} else if (1 == sscanf(userString, " %ld. ", &number)) {
	    string[0] = NUL;/* no right half */
	    got_number = TRUE;
	} else if (1 == sscanf(userString, " .%s ", string)) {
	    number = 0;	/* no left half */
	    got_number = TRUE;
	} else if (1 == sscanf(userString, " - .%s ", string)) {
	    number = 0;			/* no left half */
	    is_negative = TRUE;
	    if (neg_ok) break;
	    number = -1;/* force a failure later */
	    break;
	} else if (1 == sscanf(userString, " %ld ", &number)) {
	    got_number = TRUE;
	} else {
	    printf("? Bad format for a number: \"%s\"\n", userString);
	};
	if (got_number) {
	    if ((number < 0) && (neg_ok)) {
		is_negative = TRUE;
		number = (-number);
		break;
	    };
	    if (number >= 0) {
		is_negative = FALSE;
		break;
	    };
	    printf("? Negative number not allowed.\n");
	};
	errfree((VOID *)userString);
    };
    errfree((VOID *)userString);
    ended = FALSE;
    for (i = 0; i < right; i++) {
	number *= 10;
	if (string[i] == NUL) ended = TRUE;
	if (!ended) {
	    long int digit;
	    sscanf(&string[i], "%1ld", &digit);
	    number += digit;
	};
    };
    if (is_negative)
	number = (-number);
    return number;
}
#endif

#if GET_USER_ENABLED
/* Read floating point number from the user.
 *
 * Inputs:
 *	prompt		ASCIZ prompt string
 *	previous	default value
 *	neg_ok		negative numbers are allowed
 * Returns:
 *	value read
 */
EXTERNDEF float get_float(char *prompt,float previous,BOOLEAN neg_ok)
{
    char prevString[STRING_LENGTH+1];
    char *userString;
    char c;
    float number;
    int matches;

    sprintf(prevString,"%f",previous);
    while (TRUE) {
	userString = get_commandWord(prompt,prevString,TRUE);
	if (userString == NULL) {
	    return previous;
	};
	if (0 == strcmp(userString,"?")) {
	    printf("Enter a floating point number.\n");
	    errfree((VOID *)userString);
	    continue;
	};
	matches = sscanf(userString," %f ",&number,&c);
	if (matches != 1) {
	    printf("? Bad format for a number: \"%s\"\n", userString);
	    errfree((VOID *)userString);
	    continue;
	};
	if (number >= 0.0) break;
	if ((number < 0) && (neg_ok)) {
	    break;
	};
	printf("? Negative number not allowed.\n");
	errfree((VOID *)userString);
    };
    errfree((VOID *)userString);
    return number;
}
#endif

/* Set the output file.
 *
 * Inputs:
 *	ofile		file handle to use
 * Returns:
 *	the old file handle
 */
EXTERNDEF FILE *set_o_file(FILE *ofile)
{
    FILE *prev;
    
    prev = cur_file;
    cur_file = ofile;
    cur_lcount = 1;
    cur_line[0] = NUL;
    return prev;
}

/* Return the current output file
 */
EXTERNDEF FILE *get_o_file(void)
{
    return cur_file;
}

#ifdef WINDOWS
/* Set the windows output file superceeding the set_o_file one for now.
 *
 * Inputs:
 *	hFile		file handle to use
 * Returns:
 *	the old file handle
 */
EXTERNDEF HFILE set_windows_file(HFILE hFile)
{
    static long int lcountAtSave;
    static char lineAtSave[STRING_LENGTH+1];
    static BOOLEAN stateWasSaved = FALSE;
    HFILE prev;
    
    prev = windowsFile;
    if (hFile == HFILE_ERROR) {
	if (stateWasSaved) {
	    stateWasSaved = FALSE;
	    cur_lcount = lcountAtSave;
	    memcpy(cur_line,lineAtSave,sizeof(cur_line));
	};
    };
    windowsFile = hFile;
    if (windowsFile != HFILE_ERROR) {
	if (!stateWasSaved) {
	    stateWasSaved = TRUE;
	    lcountAtSave = cur_lcount;
	    memcpy(lineAtSave,cur_line,sizeof(lineAtSave));
	};
	cur_lcount = 1;
	cur_line[0] = NUL;
    };
    return prev;
}
/*
 * Add WIN32 support
 */
#elif defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
/* Set the WIN32 output file superceeding the set_o_file one for now.
 *
 * Inputs:
 *	hFile		file handle to use
 * Returns:
 *	the old file handle
 */
EXTERNDEF HANDLE set_win32_file(HANDLE hFile)
{
    static long int lcountAtSave;
    static char lineAtSave[STRING_LENGTH+1];
    static BOOLEAN stateWasSaved = FALSE;
    HANDLE prev;
    
    prev = win32File;
    if (hFile == INVALID_HANDLE_VALUE) {
	if (stateWasSaved) {
	    stateWasSaved = FALSE;
	    cur_lcount = lcountAtSave;
	    memcpy(cur_line,lineAtSave,sizeof(cur_line));
	};
    };
    win32File = hFile;
    if (win32File != INVALID_HANDLE_VALUE) {
	if (!stateWasSaved) {
	    stateWasSaved = TRUE;
	    lcountAtSave = cur_lcount;
	    memcpy(lineAtSave,cur_line,sizeof(lineAtSave));
	};
	cur_lcount = 1;
	cur_line[0] = NUL;
    };
    return prev;
}
#endif

#ifdef WINDOWS
/* Return the current windows output file  
 */
EXTERNDEF HFILE get_windows_file(void)
{
    return windowsFile;
}
/*
 * Add WIN32 support
 */
#elif defined (WIN32) || defined ( _WIN32 ) || defined (__WIN32__)
EXTERNDEF HANDLE get_win32_file(void)
{
    return win32File;
}
#endif

/* Open a file for output. Only one file may be open at a time.
 *
 * Inputs:
 *	name		filename
 *	message		fragment of error messages
 */
EXTERNDEF FILE *open_o_file(char *name,char *message)
{
    if (cur_file != stdout) fclose(cur_file);
    cur_file = fopen(name,"w");
    if (cur_file == NULL) {
	log_error("? Unable to open %s file \"%s\"\n",message,name);
	return NULL;
	};
    cur_lcount = 1;
    cur_line[0] = NUL;
    tabs_allowed = TRUE;
    return cur_file;
}

/* Open a file for append. Only one file may be open at a time.
 *
 * Inputs:
 *	name		filename
 *	message		fragment of error messages
 */
EXTERNDEF FILE *open_append_file(char *name,char *message)
{
    if (cur_file != stdout) fclose(cur_file);
    cur_file = fopen(name,"a");
    if (cur_file == NULL) {
	log_error("? Unable to open %s file \"%s\"\n",message,name);
	return NULL;
	};
    cur_lcount = 1;
    cur_line[0] = NUL;
    tabs_allowed = TRUE;
    return cur_file;
}

/* Close output file.
 *
 * Inputs:
 *	ofile		FILE block of file
 *	name		filename
 *	message		fragment of error messages
 */
EXTERNDEF void close_o_file(FILE *ofile,char *name,char *message)
{
    cur_lcount = 1;
    cur_line[0] = NUL;
    if (EOF == fclose(ofile)) {
	log_error("? Unable to close %s file \"%s\"\n",message,name);
	return;
	}
    cur_file = stdout;
    return;
}

/* Extract the nth field from a string. The field is copied into a static
 * buffer that is valid untill the next call to this routine.
 * If there are insufficient fields in the string an empty string is returned.
 *
 * Inputs:
 *	seperators	characters used to seperate fields
 *	string		string to parse fields from
 *	field		field to extract from string
 *	ignoreMultiple	TRUE if multiple separators are counted as one
 * Returns:
 *	pointer to the string created
 */
EXTERNDEF char *i_getField(char *seperators,char *string,int field,
			 BOOLEAN ignoreMultiple)
{
    static char temp[STRING_LENGTH+1];
    int i;
    int start;
    int length;

    if (string == NULL) return alloc_str(NULL);
    temp[0] = NUL;
    if (seperators == NULL) seperators = " \t"; /* default is space and tab */
    start = 0;
    for (i = 1;i < field;i += 1) {	/* find beginning of field n */
	int mid;
	mid = strcspn(&string[start],seperators);
	if (mid == 0) return temp;	/* not enough fields */
	start += mid;
	if (string[start] == NUL) return temp;
	start += 1;			/* step over seperator */
	if (ignoreMultiple) {
	    int j;
	    j = strspn(&string[start],seperators);
	    start += j;
	};
    };
    length = strcspn(&string[start],seperators);
    length = MIN(STRING_LENGTH,length);
    strncpy(temp,&string[start],length);
    temp[length] = NUL;
    return temp;
}

/* Get the remainder of a string after the nth field.
 * Leading whitespace is ignored so long as it does not match the mark
 * character.
 *
 * Inputs:
 *	seperators	characters used to seperate fields
 *	string		string to parse fields from
 *	field		field to extract from string
 *	ignoreMultiple	TRUE if multiple separators are counted as one
 * Returns:
 *	malloc of the string. malloc of "" if an empty field or no field
 */
EXTERNDEF char *i_afterField(char *seperators,char *string,int field,
			 BOOLEAN ignoreMultiple)
{
    int i;
    int start;

    if (string == NULL) return alloc_str(NULL);
    if (seperators == NULL) seperators = " \t"; /* default is space and tab */
    start = 0;
    for (i = 1;i < field+1;i += 1) {	/* find beginning of field n+1 */
	int mid;
	mid = strcspn(&string[start],seperators);
	if (mid == 0) return alloc_str(NULL);	/* not enough fields */
	start += mid;
	if (string[start] == NUL) return alloc_str(NULL);
	start += 1;			/* step over seperator */
	if (ignoreMultiple) {
	    int j;
	    j = strspn(&string[start],seperators);
	    start += j;
	};
    };
    return alloc_str(&string[start]);
}

/* Open and copy a file to a second file
 *
 * Inputs:
 *	name		name of file to copy
 *	ofile		file handle for printf
 * Returns:
 *	TRUE if succesfull
 */
EXTERNDEF BOOLEAN copyfile(char *name,FILE *ofile)
{
    char line[STRING_LENGTH+1];
    FILE *infile;

    infile = fopen(name,"r");
    if (infile == NULL) {
	log_error("? Unable to open %s\n",name);
	return FALSE;
    };
    if (ofile != NULL) {
	while (NULL != fgets(line,STRING_LENGTH,infile)) {
	    fputs(line,ofile);
	};
    };
    fclose(infile);
    return TRUE;
}

/* Compare two dates.
 *
 * Inputs:
 *	date1		first date
 *	date2		second date
 * Returns:
 *	0 if either date is bad
 *	0 if either date is NULL
 *	0 if the dates are the same
 *	-1 if date1 is before date2
 *	1 if date1 is after date2
 */
EXTERNDEF int date_compare(char *date1,char *date2)
{
    int month1,day1,year1;
    int month2,day2,year2;

    if (date2 == NULL) return 0;
    if (date1 == NULL) return 0;
    if (!parse_date(date1,&month1,&day1,&year1)) return 0;
    if (!parse_date(date2,&month2,&day2,&year2)) return 0;
    if (year1 > year2) return 1;
    if (year1 < year2) return -1;
    if (month1 > month2) return 1;
    if (month1 < month2) return -1;
    if (day1 > day2) return 1;
    if (day1 < day2) return -1;
    return 0;
}

/* appends bytes to a file. This is a safe append, it opens the file
** tacks the string on the end and closes the file. The data can
** contain any values including 0.
**
** Inputs:
**	filename
**	data
**	length	    exactly this many characters are written
*/
EXTERNDEF void append_to_file(char *filename,char *data,size_t length)
{
    FILE *file;

    if (filename == NULL) return;
    if (data == NULL) return;
    file = fopen(filename,"a");
    if (file == NULL) return;
    fwrite(data,length,1,file);
    fclose(file);
    return;
}


/* compute a simple XOR checksum of all the bits in an array 8 bits at
** a time. 
** DID NOT Include a seed checksum to prevent all 0s from being a valid
** memory image. If the size is 0 return something that will not match the
** seed.
**
** Inputs:
**	ptr	pointer to array
**	size	number of elements (bytes or INT8 each)
** Returns:
**	checksum
*/
EXTERNDEF unsigned INT8 compute_checksum(VOID *arr,int size)
{
    int i;
    unsigned INT8 *ptr = (unsigned INT8 *)arr;
    unsigned INT8 check;
    if (size == 0) return 0;	/* force no match */
#if 0
    check = 0x0F;			/* this is a seed */
#else
    check = 0x00;
#endif
    for (i = 0;i < size;i += 1) {
	check ^= ptr[i];
    };
    return check;
}

/* get the current time and date as a string
**
** Returns:
**	pointer to string valid untill next call to this routine
*/
EXTERNDEF char *timestamp(void)
{
    static char time_string[30];
    struct tm *time_now;
    time_t secs_now;

#if (!AOSVS)
#ifndef __STDC__
    tzset();
#endif
#endif
    time(&secs_now);
    time_now = localtime(&secs_now);
    strftime(time_string,DIM(time_string),"%d%b%Y %H:%M:%S",time_now);
    return time_string;
}

/* Similar to strncmp. Colates numbers in an order
** more appreciated by people.  Treats upper and lower case letters
** as equivalent. Ignores trailing blanks.
**
** Inputs:
**	str1
**	str2
**	maxlen
** Returns:
**	same results as strcmp
*/
EXTERNDEF int compare_name(char *str1,char *str2,int maxlen)
{
    int i;
    int j;
    INT16 v1;	/* numeric value */
    INT16 v2;
    char c1;
    char c2;
    int l1;	/* counts number of digits in number */
    int l2;

    i = 0;
    v1 = 0;
    while (TRUE) {
	c1 = (char)toupper(str1[i]);
	c2 = (char)toupper(str2[i]);
	if (c1 != c2) break;
	if (c1 == '\0') return 0;
	if (isdigit(c1)) {
	    v1 *= (INT16)10;
	    v1 += (INT16)(c1 - '0');
	} else {
	    v1 = 0;
	};
	i += 1;
	if (i >= maxlen) return 0;
    };
    if (c1 == '\0') {
	if (is_blank(&str2[i])) return 0;
    };
    if (c2 == '\0') {
	if (is_blank(&str1[i])) return 0;
    };
    if (!isdigit(c1) && !isdigit(c2)) {
	return (c1 > c2) ? 1 : -1;
    };
    v2 = v1;
    l2 = 0;
    for (j = 0;((i+j) < maxlen) && isdigit(str2[i+j]);j += 1) {
	l2 += 1;
	v2 *= (INT16)10;
	v2 += (INT16)(str2[i+j] - '0');
    };
    l1 = 0;
    for (j = 0;((i+j) < maxlen) && isdigit(str1[i+j]);j += 1) {
	l1 += 1;
	v1 *= (INT16)10;
	v1 += (INT16)(str1[i+j] - '0');
    };
    if (v1 > v2) return 1;
    if (v1 < v2) return -1;
    return (l1 > l2) ? 1 : -1;
}

/* Report timestamped information. Takes arguments just like printf.
**
** Inputs:
**      format      pattern like printf
*/
#if ANSI_C
int log_info(BOOLEAN stamp,const char *format,...)
{
    va_list ap;
    int result;
    static long int errorCount = 0;

    va_start(ap,format);
    if (debugDetailed || (errorCount <= maxErrorPrintouts)) {
	errorCount += 1;
	if(stamp) fprintf(stdout,"%s:\n    ",timestamp());
	result = vfprintf(stdout,format,ap);
    } else {
	result = 0;
    };
    va_end(ap);
    return result;
}
#else
int log_info(stamp,format,va_alist)
int stamp;
char *format;
int va_alist;
{
    va_list va_ptr;
    int result;
    static int errorCount = 0;

    va_start(va_ptr);
    if (debugDetailed || (errorCount <= maxErrorPrintouts)) {
	errorCount += 1;
	if (stamp) fprintf(stdout,"%s:\n    ",timestamp());
	result = vfprintf(stdout,format,va_ptr);
    } else {
	result = 0;
    };
    va_end(va_ptr);
    return result;
}    
#endif


/* Report an error. Takes arguments just like printf.
**
** Inputs:
**      format      pattern like printf
*/
#if ANSI_C
int log_error(const char *format,...)
{
    va_list ap;
    int result;
    static long int errorCount = 0;

    va_start(ap,format);
    if (debugDetailed || (errorCount <= maxErrorPrintouts)) {
	errorCount += 1;
	fprintf(stdout,"Error at %s:\n    ",timestamp());
	result = vfprintf(stdout,format,ap);
    } else {
	result = 0;
    };
    va_end(ap);
    return result;
}
#else
int log_error(format,va_alist)
char *format;
int va_alist;
{
    va_list va_ptr;
    int result;
    static int errorCount = 0;

    va_start(va_ptr);
    if (debugDetailed || (errorCount <= maxErrorPrintouts)) {
	errorCount += 1;
	fprintf(stdout,"Error at %s:\n",timestamp());
	result = vfprintf(stdout,format,va_ptr);
    } else {
	result = 0;
    };
    va_end(va_ptr);
    return result;
}    
#endif
